/*
 * Program:	Mailbox Access (MAIL) routines
 *
 * Author(s):	Mark Crispin
 *		Computing And Communications
 *		University of Washington
 *		Seattle, WA  98195
 *
 * Date:	22 November 1989
 *
 * Copyright 1989, 1990, 1991 by University of Washington.
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license.
 *
 */


#include <ctype.h>
#ifdef __STDC__
#include <stdlib.h>
#else
#include "ansi.h"
#endif
#include <sys/types.h>
#include "mail.h"
#include "misc.h"
#include "osdep.h"

DRIVER *maildrivers;		/* list of mail drivers */


/* Mail routines
 *
 *  mail_xxx routines are the interface between this module and the outside
 * world.  Only these routines should be referenced by external callers.
 *
 *  Note that there is an important difference between a "sequence" and a
 * "message #" (msgno).  A sequence is a string representing a sequence in
 * {"n", "n:m", or combination separated by commas} format, whereas a msgno
 * is a single integer.
 *
 */

/* Mail link driver
 * Accepts: driver to add to list
 */

void mail_link (DRIVER *driver)
{
  DRIVER **d = &maildrivers;
  while (*d) d = &(*d)->next;	/* find end of list of drivers */
  *d = driver;			/* put driver at the end */
  driver->next = NIL;		/* this driver is the end of the list */
}

/* Mail open
 * Accepts: candidate stream for recycling
 *	    mailbox name
 *	    initial debugging flag
 * Returns: stream to use on success, NIL on failure
 */

MAILSTREAM *mail_open (MAILSTREAM *stream,char *name,short debug)
{
  int i;
  char tmp[MAILTMPLEN];
  DRIVER *factory = maildrivers ? (*maildrivers->valid) (name) : NIL;
  if (factory) {		/* must have a factory */
    if (!stream) {		/* instantiate stream if none to recycle */
      stream = (MAILSTREAM *) fs_get (sizeof (MAILSTREAM));
      stream->dtb = factory;	/* initialize stream fields */
      stream->local = NIL;
      stream->mailbox = cpystr (name);
      stream->lock = NIL;
      stream->use = stream->sequence = stream->gensym = stream->nmsgs =
	stream->recent = 0;
      stream->flagstring = NIL;
      for (i = 0; i < NUSERFLAGS; ++i) stream->user_flags[i] = NIL;
      for (i = 0; i < MAXMESSAGES; ++i) stream->messagearray[i] = NIL;
    }
    else {			/* close driver if different from factory */
      if (stream->dtb != factory) {
	if (stream->dtb) (*stream->dtb->close) (stream);
	stream->dtb = factory;	/* establish factory as our driver */
	stream->local = NIL;	/* flush old driver's local data */
      }
				/* clean up old stream for recycling */
      if (stream->mailbox) fs_give ((void **) &stream->mailbox);
      stream->mailbox = cpystr (name);
      mail_free_messagearray (stream->messagearray);
    }
    stream->debug = debug;	/* set to desired debugging state */
				/* have driver open, flush if failed */
    if (!(*factory->open) (stream)) stream = mail_close (stream);
  }
  else if (debug != 36) {	/* not from Tenex internal please */
    sprintf (tmp,"Unknown mailbox type: %s",name);
    mm_log (tmp,ERROR);		/* give an error */
  }
  return stream;		/* return the stream */
}

/* Mail close
 * Accepts: mail stream
 * Returns: NIL
 */

MAILSTREAM *mail_close (MAILSTREAM *stream)
{
  int i;
  if (stream) {			/* make sure argument given */
				/* do the driver's close action */
    if (stream->dtb) (*stream->dtb->close) (stream);
    stream->dtb = NIL;		/* disassociate from driver */
    stream->local = NIL;	/* and driver's local data */
    if (stream->mailbox) fs_give ((void **) &stream->mailbox);
    stream->mailbox = NIL;
    stream->sequence++;		/* invalidate sequence */
    mail_free_messagearray (stream->messagearray);
    if (stream->flagstring) fs_give ((void **) &stream->flagstring);
    stream->flagstring = NIL;
    for (i = 0; i < NUSERFLAGS; ++i) stream->user_flags[i] = NIL;
				/* finally free the stream's storage */
    if (!stream->use) fs_give ((void **) &stream);
  }
  return NIL;
}

/* Mail make handle
 * Accepts: Mail stream
 * Returns: handle
 *
 *  Handles provide a way to have multiple pointers to a stream yet allow the
 * stream's owner to nuke it or recycle it.
 */

MAILHANDLE *mail_makehandle (MAILSTREAM *stream)
{
  MAILHANDLE *handle = (MAILHANDLE *) fs_get (sizeof (MAILHANDLE));
  handle->stream = stream;	/* copy stream */
				/* and its sequence */
  handle->sequence = stream->sequence;
  stream->use++;		/* let stream know another handle exists */
  return handle;
}


/* Mail release handle
 * Accepts: Mail handle
 */

void mail_free_handle (MAILHANDLE **handle)
{
  MAILSTREAM *s;
  if (*handle) {		/* only free if exists */
				/* resign stream, flush unreferenced zombies */
    if ((!--(s = (*handle)->stream)->use) && !s->dtb) fs_give ((void **) &s);
    fs_give ((void **) handle);	/* now flush the handle */
  }
}


/* Mail get stream handle
 * Accepts: Mail handle
 * Returns: Mail stream or NIL if stream gone
 */

MAILSTREAM *mail_stream (MAILHANDLE *handle)
{
  MAILSTREAM *s = handle->stream;
  return (s->dtb && (handle->sequence == s->sequence)) ? s : NIL;
}

/* Mail fetch cache element
 * Accepts: Mail stream
 *	    message # to fetch
 * Returns: cache element of this message
 */

MESSAGECACHE *mail_elt (MAILSTREAM *stream,int msgno)
{
  int i = msgno-1;
				/* if no cache entry, create it */
  if (!stream->messagearray[i]) {
    stream->messagearray[i] = (MESSAGECACHE *) fs_get (sizeof (MESSAGECACHE));
				/* initialize newly-created cache entry */
    stream->messagearray[i]->lockcount = 1;
    stream->messagearray[i]->msgno = msgno;
    stream->messagearray[i]->internal_date = NIL;
    stream->messagearray[i]->seen = stream->messagearray[i]->deleted =
      stream->messagearray[i]->flagged = stream->messagearray[i]->answered =
	stream->messagearray[i]->searched = stream->messagearray[i]->recent =
	  NIL;			/* init system flags */
    stream->messagearray[i]->user_flags = NIL;
    stream->messagearray[i]->rfc822_size = 0;
    stream->messagearray[i]->env = NIL;
    stream->messagearray[i]->body = NIL;
    stream->messagearray[i]->from = NIL;
    stream->messagearray[i]->subject = NIL;
  }
				/* return the cache element */
  return stream->messagearray[i];
}

/* Mail fetch fast information
 * Accepts: Mail stream
 *	    sequence
 *
 * Generally, mail_fetchenvelope is preferred
 */

void mail_fetchfast (MAILSTREAM *stream,char *sequence)
{
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->fetchfast) (stream,sequence);
}


/* Mail fetch flags
 * Accepts: Mail stream
 *	    sequence
 */

void mail_fetchflags (MAILSTREAM *stream,char *sequence)
{
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->fetchflags) (stream,sequence);
}


/* Mail fetch envelope
 * Accepts: Mail stream
 *	    message # to fetch
 * Returns: envelope of this message
 *
 * Fetches the "fast" information as well
 */

ENVELOPE *mail_fetchenvelope (MAILSTREAM *stream,int msgno)
{
  				/* do the driver's action */
  return stream->dtb ? (*stream->dtb->fetchenvelope) (stream,msgno) : NIL;
}

/* Mail fetch message header
 * Accepts: Mail stream
 *	    message # to fetch
 * Returns: message header in RFC822 format
 */

char *mail_fetchheader (MAILSTREAM *stream,int msgno)
{
  				/* do the driver's action */
  return stream->dtb ? (*stream->dtb->fetchheader) (stream,msgno) : "";
}


/* Mail fetch message text (body only)
 * Accepts: Mail stream
 *	    message # to fetch
 * Returns: message text in RFC822 format
 */

char *mail_fetchtext (MAILSTREAM *stream,int msgno)
{
  				/* do the driver's action */
  return stream->dtb ? (*stream->dtb->fetchtext) (stream,msgno) : "";
}


/* Mail fetch message body as a structure
 * Accepts: Mail stream
 *	    message # to fetch
 *	    section #
 * Returns: pointer to section of message body
 */

char *mail_fetchbody (MAILSTREAM *stream,int msgno,int section)
{
  				/* do the driver's action */
  return stream->dtb ? (*stream->dtb->fetchbody) (stream,msgno,section) : "";
}

/* Mail fetch From string for menu
 * Accepts: Mail stream
 *	    message # to fetch
 *	    desired string length
 * Returns: string of requested length
 */

char *mail_fetchfrom (MAILSTREAM *stream,int msgno,int length)
{
  char *s;
  char tmp[MAILTMPLEN];
  ENVELOPE *env;
  MESSAGECACHE *c = mail_elt (stream,msgno);
  if (!c->from) {		/* if we don't have one yet */
    c->from = (char *) fs_get (length+1);/* create it */
    memset (c->from,' ',length);/* fill it with spaces */
    c->from[length] = '\0';	/* tie off with null */
				/* get first from address from envelope */
    if ((env = mail_fetchenvelope (stream,msgno)) && env->from) {
				/* if a personal name exists use it */
      if (s = env->from->personal) memcpy (c->from,s,min (length,strlen (s)));
      else {			/* otherwise use mailbox@host */
	sprintf (tmp,"%s@%s",env->from->mail_box,env->from->host);
	memcpy (c->from,tmp,min (length,strlen (tmp)));
      }
    }
  }
  return c->from;		/* return the resulting string */
}


/* Mail fetch Subject string for menu
 * Accepts: Mail stream
 *	    message # to fetch
 *	    desired string length
 * Returns: string of no more than requested length
 */

char *mail_fetchsubject (MAILSTREAM *stream,int msgno,int length)
{
  ENVELOPE *env;
  MESSAGECACHE *c = mail_elt (stream,msgno);
  if (!c->subject) {		/* if we don't have one yet */
				/* create it */
    c->subject = (char *) fs_get (length+1);
    memset (c->subject,'\0',length+1);
				/* copy subject from envelope */
    if ((env = mail_fetchenvelope (stream,msgno)) && env->subject)
      strncpy (c->subject,env->subject,length);
    else c->subject[0] = ' ';	/* if no subject then just a space */
  }
  return c->subject;		/* return the resulting string */
}

/* Mail set flag
 * Accepts: Mail stream
 *	    sequence
 *	    flag(s)
 */

void mail_setflag (MAILSTREAM *stream,char *sequence,char *flag)
{
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->setflag) (stream,sequence,flag);
}


/* Mail clear flag
 * Accepts: Mail stream
 *	    sequence
 *	    flag(s)
 */

void mail_clearflag (MAILSTREAM *stream,char *sequence,char *flag)
{
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->clearflag) (stream,sequence,flag);
}


/* Mail search for messages
 * Accepts: Mail stream
 *	    search criteria
 */

void mail_search (MAILSTREAM *stream,char *criteria)
{
  int i = 1;
  while (i <= stream->nmsgs) mail_elt (stream,i++)->searched = NIL;
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->search) (stream,criteria);
}


/* Mail ping mailbox
 * Accepts: Mail stream
 * Returns: stream if still open else NIL
 */

MAILSTREAM *mail_ping (MAILSTREAM *stream)
{
  				/* do the driver's action */
  return stream->dtb ? (*stream->dtb->ping) (stream) : NIL;
}

/* Mail check mailbox
 * Accepts: Mail stream
 */

void mail_check (MAILSTREAM *stream)
{
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->check) (stream);
}


/* Mail expunge mailbox
 * Accepts: Mail stream
 */

void mail_expunge (MAILSTREAM *stream)
{
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->expunge) (stream);
}


/* Mail copy message(s)
 * Accepts: Mail stream
 *	    sequence
 *	    destination mailbox
 */

void mail_copy (MAILSTREAM *stream,char *sequence,char *mailbox)
{
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->copy) (stream,sequence,mailbox);
}


/* Mail move message(s)
 * Accepts: Mail stream
 *	    sequence
 *	    destination mailbox
 */

void mail_move (MAILSTREAM *stream,char *sequence,char *mailbox)
{
  				/* do the driver's action */
  if (stream->dtb) (*stream->dtb->move) (stream,sequence,mailbox);
}

/* Mail messages have been searched out
 * Accepts: MAIL stream
 *	    message number
 *
 * Calls external "mm_searched" function to notify main program
 */

void mail_searched (MAILSTREAM *stream,int msgno)
{
				/* mark as searched */
  mail_elt (stream,msgno)->searched = T;
  mm_searched (stream,msgno);	/* notify main program */
}


/* Mail n messages exist
 * Accepts: MAIL stream
 *	    number of messages
 *
 * Calls external "mm_exists" function that notifies main program prior
 * to updating the stream
 */

void mail_exists (MAILSTREAM *stream,int nmsgs)
{
  char tmp[MAILTMPLEN];
  if (nmsgs > MAXMESSAGES) {
    sprintf (tmp,"Mailbox has too many (%d) messages.  Using max of %d",
	     nmsgs,MAXMESSAGES);
    mm_log (tmp,WARN);
    nmsgs = MAXMESSAGES;
  }
				/* notify main program of change */
  if (stream->debug != 36) mm_exists (stream,nmsgs);
  stream->nmsgs = nmsgs;	/* update stream status */
}

/* Mail n messages are recent
 * Accepts: MAIL stream
 *	    number of recent messages
 */

void mail_recent (MAILSTREAM *stream,int recent)
{
  stream->recent = recent;	/* update stream status */
}


/* Mail message n is expunged
 * Accepts: MAIL stream
 *	    message #
 *
 * Calls external "mm_expunged" function that notifies main program prior
 * to updating the stream
 */

void mail_expunged (MAILSTREAM *stream,int msgno)
{
  int i = msgno - 1;
  if (stream->messagearray[i]) {/* if an element is there */
				/* invalidate its message number */
    stream->messagearray[i]->msgno = 0;
				/* maybe free this message element */
    mail_free_elt (&stream->messagearray[i]);
  }
				/* slide down remainder of cache */
  for (i = msgno; i < stream->nmsgs; ++i)
    if (stream->messagearray[i-1] = stream->messagearray[i])
      stream->messagearray[i-1]->msgno = i;
  stream->messagearray[stream->nmsgs-1] = NIL;
  --(stream->nmsgs);		/* update stream status */
				/* notify main program of change */
  if (stream->debug != 36) mm_expunged (stream,msgno);
}

/* Mail stream status routines */


/* Mail lock stream
 * Accepts: Mail stream
 */

void mail_lock (MAILSTREAM *stream)
{
  if (stream->lock) fatal ("Lock when already locked");
  else stream->lock = T;	/* lock stream */
}


/* Mail unlock stream
 * Accepts: Mail stream
 */
 
void mail_unlock (MAILSTREAM *stream)
{
  if (!stream->lock) fatal ("Unlock when not locked");
  else stream->lock = NIL;	/* unlock stream */
}


/* Mail check if stream locked
 * Accepts: Mail stream
 * Returns: lock status
 */

short mail_checklock (MAILSTREAM *stream)
{
  return stream->lock;		/* return lock status */
}


/* Mail turn on debugging telemetry
 * Accepts: Mail stream
 */

void mail_debug (MAILSTREAM *stream)
{
  stream->debug = T;		/* turn on debugging telemetry */
}


/* Mail turn off debugging telemetry
 * Accepts: Mail stream
 */

void mail_nodebug (MAILSTREAM *stream)
{
  stream->debug = NIL;		/* turn off debugging telemetry */
}

/* Mail data structure instantiation routines */


/* Mail instantiate envelope
 * Returns: new envelope
 */

ENVELOPE *mail_newenvelope ()
{
  ENVELOPE *env = (ENVELOPE *) fs_get (sizeof (ENVELOPE));
  env->remail = NIL;		/* initialize all fields */
  env->return_path = NIL;
  env->date = NIL;
  env->subject = NIL;
  env->from = env->sender = env->reply_to = env->to = env->cc = env->bcc = NIL;
  env->in_reply_to = env->message_id = NIL;
  return env;
}


/* Mail instantiate address
 * Returns: new address
 */

ADDRESS *mail_newaddr ()
{
  ADDRESS *adr = (ADDRESS *) fs_get (sizeof (ADDRESS));
				/* initialize all fields */
  adr->personal = adr->route_list = adr->mail_box = adr->host = adr->error
    = NIL;
  adr->next = NIL;
  return adr;
}


/* Mail instantiate body
 * Returns: new body
 */

BODY *mail_newbody ()
{
  BODY *body = (BODY *) fs_get (sizeof (BODY));
  body->count = 0;		/* no lines */
  body->type = ENCTEXT;		/* default text type */
  body->options = NIL;		/* undefined options */
  body->text = NIL;		/* no text yet */
  body->next = NIL;		/* no next */
  return body;
}

/* Mail garbage collection routines */


/* Mail garbage collect address
 * Accepts: pointer to address pointer
 */

void mail_free_address (ADDRESS **address)
{
  if (*address) {		/* only free if exists */
    if ((*address)->personal) fs_give ((void **) &(*address)->personal);
    if ((*address)->route_list) fs_give ((void **) &(*address)->route_list);
    if ((*address)->mail_box) fs_give ((void **) &(*address)->mail_box);
    if ((*address)->host) fs_give ((void **) &(*address)->host);
    if ((*address)->error) fs_give ((void **) &(*address)->error);
    mail_free_address (&(*address)->next);
    fs_give ((void **) address);/* return address to free storage */
  }
}


/* Mail garbage collect body
 * Accepts: pointer to body pointer
 */

void mail_free_body (BODY **body)
{
  if (*body) {			/* only free if exists */
    if ((*body)->options) fs_give ((void **) &(*body)->options);
    if ((*body)->text) fs_give ((void **) &(*body)->text);
				/* run down the list as necessary */
    mail_free_body (&(*body)->next);
    fs_give ((void **) body);	/* return body to free storage */
  }
}


/* Mail garbage collect messagearray
 * Accepts: pointer to messagearray pointer
 *
 * The messagearray is set to all NIL when this function finishes.
 */

void mail_free_messagearray (MESSAGECACHE **messagearray)
{
  int i;
				/* free each array element */
  for (i = 0; i < MAXMESSAGES; ++i) mail_free_elt (&messagearray[i]);
}

/* Mail garbage collect cache element
 * Accepts: pointer to cache element pointer
 */

void mail_free_elt (MESSAGECACHE **elt)
{
				/* only free if exists and no sharers */
  if (*elt && !--(*elt)->lockcount) {
    if ((*elt)->internal_date) fs_give ((void **) &(*elt)->internal_date);
    mail_free_envelope (&(*elt)->env);
    mail_free_body (&(*elt)->body);
    if ((*elt)->from) fs_give ((void **) &(*elt)->from);
    if ((*elt)->subject) fs_give ((void **) &(*elt)->subject);
    fs_give ((void **) elt);	/* return cache element to free storage */
  }
  else *elt = NIL;		/* else simply drop pointer */
}


/* Mail garbage collect envelope
 * Accepts: pointer to envelope pointer
 */

void mail_free_envelope (ENVELOPE **env)
{
  if (*env) {			/* only free if exists */
    if ((*env)->remail) fs_give ((void **) &(*env)->remail);
    mail_free_address (&(*env)->return_path);
    if ((*env)->date) fs_give ((void **) &(*env)->date);
    mail_free_address (&(*env)->from);
    mail_free_address (&(*env)->sender);
    mail_free_address (&(*env)->reply_to);
    if ((*env)->subject) fs_give ((void **) &(*env)->subject);
    mail_free_address (&(*env)->to);
    mail_free_address (&(*env)->cc);
    mail_free_address (&(*env)->bcc);
    if ((*env)->in_reply_to) fs_give ((void **) &(*env)->in_reply_to);
    if ((*env)->message_id) fs_give ((void **) &(*env)->message_id);
    fs_give ((void **) env);	/* return envelope to free storage */
  }
}
