/*
 * Program:	RFC-822 (RFC822) routines
 *
 * Author(s):	Mark Crispin
 *		Symbolic Systems Resources Group
 *		Knowledge Systems Laboratory
 *		Departments of Computer Science and Medicine
 *		Stanford University
 *		Stanford, CA  94305
 *
 * Date:	27 July 1988
 *
 * Sponsorship:	This work was supported by the Biomedical Research
 *		Technology Program of the National Institutes of Health
 *		under grant number RR-00785.
 *
 * Copyright 1988, 1989, 1990, 1991 by The Leland Stanford Junior University.
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license from the Stanford Office of
 * Technology Licensing.
 *
 */


#include <ctype.h>
#ifdef __STDC__
#include <string.h>
#else
#include "ansi.h"
#endif
#include "mail.h"
#include "rfc822.h"
#include "misc.h"


/* Vector of encoding type texts */

const char *ent[] = {"TEXT", "MESSAGE", "HEX", "EVFU", "EDI", "X.400",
		       "UUENCODE", "ENCRYPTED", "UNDEF"};


/* RFC822 writing routines
 *
 * These routines are logically part of the SMTP routines, but have
 * been separated out here in case some external caller would like to
 * use them.
 *
 */

/* Write RFC822 header from message structure
 * Accepts: scratch buffer to write into
 *	    message envelope
 *	    message body
 */

void rfc822_header (char *header,ENVELOPE *env,BODY *body)
{
  char c,*s;
  if (env->remail) {		/* if remailing */
    int i = strlen (env->remail);
				/* flush extra blank line */
    if (i > 4 && env->remail[i-4] == '\r') env->remail[i-2] = '\0';
    strcpy (header,env->remail);/* start with remail header */
  }
  else *header = '\0';		/* else initialize header to null */
  rfc822_header_line (&header,"Date",env,env->date);
  rfc822_address_line (&header,"From",env,env->from);
  rfc822_address_line (&header,"Sender",env,env->sender);
  rfc822_address_line (&header,"Reply-To",env,env->reply_to);
  rfc822_header_line (&header,"Subject",env,env->subject);
  rfc822_address_line (&header,"To",env,env->to);
  rfc822_address_line (&header,"cc",env,env->cc);
/* bcc's are never written...
 * rfc822_address_line (&header,"bcc",env,env->bcc);
 */
  rfc822_header_line (&header,"In-Reply-To",env,env->in_reply_to);
  rfc822_header_line (&header,"Message-ID",env,env->message_id);
  if (body && body->next) {	/* don't need to do for simple messages */
    strcat (header,"Encoding: ");
    do {			/* for each section */
      body->count = 0;		/* recompute body line count */
      if (s = body->text) {	/* only bother if this body non-empty */
	while (c = *s++) if (c == '\n') body->count++;
	sprintf ((header += strlen (header)),
		 (body->next ? (body->options ? "%d %s %s, " : "%d %s, ") :
		  (body->options ? "%d %s %s\015\012" : "%d %s\015\012")),
		 max (body->count,1),ent[body->type],body->options);
      }
    } while (body = body->next);
  }
  strcat (header,"\015\012");	/* write terminating blank line */
}

/* Write RFC822 address from header line
 * Accepts: pointer to destination string pointer
 *	    pointer to header type
 *	    message to interpret
 *	    address to interpret
 */

void rfc822_address_line (char **header,char *type,ENVELOPE *env,ADDRESS *adr)
{
  if (adr) {
    sprintf ((*header += strlen (*header)),"%s%s: ",
	     env->remail ? "ReSent-" : "",type);
    rfc822_write_address (*header,adr);
    sprintf ((*header += strlen (*header)),"\015\012");
  }
}


/* Write RFC822 text from header line
 * Accepts: pointer to destination string pointer
 *	    pointer to header type
 *	    message to interpret
 *	    pointer to text
 */

void rfc822_header_line (char **header,char *type,ENVELOPE *env,char *text)
{
  if (text) sprintf ((*header += strlen (*header)),"%s%s: %s\015\012",
		     env->remail ? "ReSent-" : "",type,text);
}


/* Write RFC822 address
 * Accepts: pointer to destination string
 *	    address to interpret
 */

void rfc822_write_address (char *dest,ADDRESS *adr)
{
  while (adr) {
				/* simple case? */
    if (!(adr->personal || adr->route_list)) rfc822_address (dest,adr);
    else {			/* no, must use phrase <route-addr> form */
      if (adr->personal) {	/* have a personal name? */
	rfc822_cat (dest,adr->personal,"()<>@,;:\\\".[]");
	strcat (dest," <");	/* write address delimiter */
      }
      else strcat (dest,"<");	/* write address delimiter */
      rfc822_address (dest,adr);/* write address */
      strcat (dest,">");	/* closing delimiter */
    }
    adr = adr->next;		/* get next recipient */
    if (adr) strcat (dest,", ");/* delimit if there is one */
  }
}

/* Write RFC822 route-address to string
 * Accepts: pointer to destination string
 *	    address to interpret
 */

void rfc822_address (char *dest,ADDRESS *adr)
{
  if (adr) {			/* no-op if no address */
    if (adr->route_list) {	/* have an A-D-L? */
      strcat (dest,adr->route_list);
      strcat (dest,":");
    }
				/* write mailbox name */
    rfc822_cat (dest,adr->mail_box," ()<>@,;:\\\"[]");
    strcat (dest,"@");		/* host delimiter */
    strcat (dest,adr->host);	/* write host name */
  }
}


/* Concatenate RFC822 string
 * Accepts: pointer to destination string
 *	    pointer to string to concatenate
 *	    list of special characters
 */

void rfc822_cat (char *dest,char *src,char *specials)
{
  char *s;
  if (strpbrk (src,specials)) {	/* any specials present? */
    strcat (dest,"\"");		/* opening quote */
				/* truly bizarre characters in there? */
    while (s = strpbrk (src,"\\\"")) {
      strncat (dest,src,s-src);	/* yes, output leader */
      strcat (dest,"\\");	/* quoting */
      strncat (dest,s,1);	/* output the bizarre character */
      src = ++s;		/* continue after the bizarre character */
    }
    if (*src) strcat (dest,src);/* output non-bizarre string */
    strcat (dest,"\"");		/* closing quote */
  }
  else strcat (dest,src);	/* otherwise it's the easy case */
}

/* RFC822 parsing routines */


/* Parse an RFC822 message
 * Accepts: message cache element
 *	    pointer to source header
 *	    source byte count
 *	    pointer to local host name
 *	    pointer to scratch buffer
 */

void rfc822_parse_msg (MESSAGECACHE *elt,char *s,int i,char *host,char *tmp)
{
  ENVELOPE *env = (elt->env = mail_newenvelope ());
  BODY *body = NIL;
  char c,*t,*d;
  while (i > 0 && *s != '\n') {	/* until end of header */
    t = tmp;			/* initialize buffer pointer */
    c = ' ';			/* and previous character */
    while (c) {			/* collect text until logical end of line */
      switch (*s) {
      case '\n':		/* newline, possible end of logical line */
				/* tie off unless next line starts with WS */
	if (s[1] != ' ' && s[1] != '\t') *t = c = '\0';
	break;
      case '\r':		/* return, unlikely but just in case */
      case '\t':		/* tab */
      case ' ':			/* here for all forms of whitespace */
				/* insert whitespace if not already there */
	if (c != ' ') *t++ = c = ' ';
	break;
      default:			/* all other characters */
	*t++ = c = *s;		/* insert the character into the line */
	break;
      }
      if (i-- > 0) s++;		/* get next character */
      else *t++ = c = '\0';	/* end of header */
    }
    if (d = strchr (tmp,':')) {	/* find header item type */
      *d++ = '\0';		/* tie off header item, point at its data */
				/* flush whitespace */
      while (*d == ' ' || *d == '\t') d++;
      while ((t = strchr (tmp,' ')) || (t = strchr (tmp,'\t'))) *t = '\0';
      switch (*ucase (tmp)) {	/* dispatch based on first character */
      case '>':			/* possible >From: */
	if (!strcmp (tmp+1,"FROM")) rfc822_parse_adrlist (&env->from,d,host);
	break;
      case 'B':			/* possible bcc: */
	if (!strcmp (tmp+1,"CC")) rfc822_parse_adrlist (&env->bcc,d,host);
	break;
      case 'C':			/* possible cc: */
	if (!strcmp (tmp+1,"C")) rfc822_parse_adrlist (&env->cc,d,host);
	break;
      case 'D':			/* possible Date: */
	if (!env->date && !strcmp (tmp+1,"ATE")) env->date = cpystr (d);
	break;

      case 'E':			/* possible Encoding: */
	if (!strcmp (tmp+1,"NCODING") && (d = strtok (d,","))) do {
				/* create a new body section */
	  if (body) body = body->next = mail_newbody ();
	  else body = elt->body = mail_newbody ();
	  rfc822_skipws (&d);	/* skip leading whitespace */
				/* encoding count if specified */
	  if ((*ucase (d)) >= '0' && *d <= '9') body->count = strtol (d,&d,10);
	  rfc822_skipws (&d);	/* skip successive whitespace */
				/* found a delimiter? */
	  if (t = strpbrk (d," \t")) {
	    *t++ = '\0';	/* yes, tie off type */
	    rfc822_skipws (&t);	/* skip any leading whitespace */
				/* copy options */
	    if (*t) body->options = cpystr (t);
	  }
	  switch (*d++) {	/* dispatch based on first type character */
	  case 'E':		/* possible EDI or ENCRYPTED or EVFU */
	    if (!strcmp (d,"DI")) body->type = ENCEDI;
	    else if (!strcmp (d,"NCRYPTED")) body->type = ENCENCRYPTED;
	    else if (!strcmp (d,"VFU")) body->type = ENCEVFU;
	    break;
	  case 'H':		/* possible HEX */
	    if (!strcmp (d,"EX")) body->type = ENCHEX;
	    break;
	  case 'M':		/* possible MESSAGE */
	    if (!strcmp (d,"ESSAGE")) body->type = ENCMESSAGE;
	    break;
	  case 'T':		/* possible TEXT */
	    if (!strcmp (d,"EXT")) body->type = ENCTEXT;
	    break;
	  case 'U':		/* possible UUENCODE */
	    if (!strcmp (d,"UENCODE")) body->type = ENCUUENCODE;
	    break;
	  case 'X':		/* possible X.400 */
	    if (!strcmp (d,".400")) body->type = ENCX400;
	    break;
	  default:		/* unknown type */
	    body->type = ENCUNDEF;
	    break;
	  }
				/* see if any more encodings */
	} while (d = strtok (NIL,","));
	break;

      case 'F':			/* possible From: */
	if (!strcmp (tmp+1,"ROM")) rfc822_parse_adrlist (&env->from,d,host);
	break;
      case 'I':			/* possible In-Reply-To: */
	if (!env->in_reply_to && !strcmp (tmp+1,"N-REPLY-TO"))
	  env->in_reply_to = cpystr (d);
	break;
      case 'M':			/* possible Message-ID: */
	if (!env->message_id && !strcmp (tmp+1,"ESSAGE-ID"))
	  env->message_id = cpystr (d);
	break;
      case 'R':			/* possible Reply-To: */
	if (!strcmp (tmp+1,"EPLY-TO"))
	  rfc822_parse_adrlist (&env->reply_to,d,host);
	break;
      case 'S':			/* possible Subject: or Sender: */
	if (!env->subject && !strcmp (tmp+1,"UBJECT"))
	  env->subject = cpystr (d);
	else if (!strcmp (tmp+1,"ENDER"))
	  rfc822_parse_adrlist (&env->sender,d,host);
	break;
      case 'T':			/* possible To: */
	if (!strcmp (tmp+1,"O")) rfc822_parse_adrlist (&env->to,d,host);
	break;
      default:
	break;
      }
    }
  }
				/* default Sender: and Reply-To: to From: */
  if (!env->sender) env->sender = rfc822_cpy_adr (env->from);
  if (!env->reply_to) env->reply_to = rfc822_cpy_adr (env->from);
}

/* Parse RFC822 address list
 * Accepts: address list to write to
 *	    input string
 *	    default host name
 */

void rfc822_parse_adrlist (ADDRESS **lst,char *string,char *host)
{
  char tmp[MAILTMPLEN];
  char *p;
  int n = 0;
  ADDRESS *last = *lst;
  ADDRESS *adr;
				/* run to tail of list */
  if (last) while (last->next) last = last->next;
				/* loop until string exhausted */
  if (*string != '\0') while (p = string) {
				/* see if start of group */
    while ((*p == ':' || ((p = rfc822_parse_phrase (string)) && *p == ':')) &&
	   ++n) string = ++p;
    rfc822_skipws (&string);	/* skip any following whitespace */
    if (!string) return;	/* punt if unterminated group */
				/* if not empty group */
    if (*string != ';' || n <= 0) {
				/* got an address? */
      if (adr = rfc822_parse_address (&string,host)) {
	if (!*lst) *lst = adr;	/* yes, first time through? */
	else last->next = adr;	/* no, append to the list */
	last = adr;		/* set for subsequent linking */
      }
      else if (string) {	/* bad mailbox */
	sprintf (tmp,"Bad mailbox: %s",string);
	mm_log (tmp,WARN);
	return;
      }
    }
				/* handle end of group */
    if (string && *string == ';' && (--n) >= 0) {
      string++;			/* skip past the semicolon */
      rfc822_skipws (&string);	/* skip any following whitespace */
      switch (*string) {	/* see what follows */
      case ',':			/* another address? */
	++string;		/* yes, skip past the comma */
	break;
      case '\0':		/* end of string */
	return;
      default:
	sprintf (tmp,"Junk at end of group: %s",string);
	mm_log (tmp,WARN);
	return;
      }
    }
  }
}

/* Parse RFC822 address
 * Accepts: pointer to string pointer
 *	    default host
 * Returns: address
 *
 * Updates string pointer
 */

ADDRESS *rfc822_parse_address (char **string,char *defaulthost)
{
  char tmp[MAILTMPLEN];
  ADDRESS *adr;
  char *st;
  char *phrase;
  if (!string) return NIL;
  rfc822_skipws (string);	/* flush leading whitespace */

  /* This is much more complicated than it should be because users like
   * to write local addrspecs without "@localhost".  This makes it very
   * difficult to tell a phrase from an addrspec!
   * The other problem we must cope with is a route-addr without a leading
   * phrase.  Yuck!
   */

  if (*(st = *string) == '<') 	/* note start, handle case of phraseless RA */
    adr = rfc822_parse_routeaddr (st,string,defaulthost);
  else {			/* get phrase if any */
    if ((phrase = rfc822_parse_phrase (st)) &&
	(adr = rfc822_parse_routeaddr (phrase,string,defaulthost))) {
      *phrase = '\0';		/* tie off phrase */
				/* phrase is a personal name */
      adr->personal = rfc822_cpy (st);
    }
    else adr = rfc822_parse_addrspec (st,string,defaulthost);
  }
				/* analyze what follows */
  if (*string) switch (**string) {
  case ',':			/* comma? */
    ++*string;			/* then another address follows */
    break;
  case ';':			/* possible end of group? */
    break;			/* let upper level deal with it */
  default:
    sprintf (tmp,"Junk at end of address: %s",*string);
    mm_log (tmp,WARN);
    *string = NIL;		/* punt remainder of parse */
  }
  return adr;			/* return the address */
}

/* Parse RFC822 route-address
 * Accepts: string pointer
 *	    pointer to string pointer to update
 * Returns: address
 *
 * Updates string pointer
 */

ADDRESS *rfc822_parse_routeaddr (char *string,char **ret,char *defaulthost)
{
  char tmp[MAILTMPLEN];
  ADDRESS *adr;
  char *route_list = NIL;
  char *routeend = NIL;
  if (!string) return NIL;
  rfc822_skipws (&string);	/* flush leading whitespace */
				/* must start with open broket */
  if (*string != '<') return NIL;
  if (string[1] == '@') {	/* have an A-D-L? */
    route_list = ++string;	/* yes, remember that fact */
    while (*string != ':') {	/* search for end of A-D-L */
				/* punt if never found */
      if (*string == '\0') return NIL;
      ++string;			/* try next character */
    }
    *string = '\0';		/* tie off A-D-L */
    routeend = string;		/* remember in case need to put back */
  }
				/* parse address spec */
  if (!(adr = rfc822_parse_addrspec (++string,ret,defaulthost))) {
				/* put colon back since parse barfed */
    if (route_list) *routeend = ':';
    return NIL;
  }
				/* have an A-D-L? */
  if (route_list) adr->route_list = cpystr (route_list);
				/* make sure terminated OK */
    if (*ret) if (**ret == '>') {
    ++*ret;			/* skip past the broket */
				/* wipe pointer if at end of string */
    if (**ret == '\0') *ret = NIL;
    return adr;			/* return the address */
  }
  sprintf (tmp,"Unterminated mailbox: %s@%s",adr->mail_box,adr->host);
  mm_log (tmp,WARN);
  return adr;			/* return the address */
}

/* Parse RFC822 address-spec
 * Accepts: string pointer
 *	    pointer to string pointer to update
 *	    default host
 * Returns: address
 *
 * Updates string pointer
 */

ADDRESS *rfc822_parse_addrspec (char *string,char **ret,char *defaulthost)
{
  ADDRESS *adr;
  char *end;
  char c,*s;
  if (!string) return NIL;
  rfc822_skipws (&string);	/* flush leading whitespace */
				/* find end of mailbox */
  if (!(end = rfc822_parse_word (string,NIL))) return NIL;
  adr = mail_newaddr ();	/* create address block */
  c = *end;			/* remember delimiter */
  *end = '\0';			/* tie off mailbox */
				/* copy mailbox */
  adr->mail_box = rfc822_cpy (string);
  *end = c;			/* restore delimiter */
  rfc822_skipws (&end);		/* skip whitespace */
  if (*end == '@') {		/* have host name? */
    ++end;			/* skip delimiter */
    rfc822_skipws (&end);	/* skip whitespace */
    *ret = end;			/* update return pointer */

    				/* search for end of host */
    end = rfc822_parse_word ((string = end)," ()<>@,;:\"\\");
    if (end) {			/* got host name, copy it */
      c = *end;			/* remember delimiter */
      *end = '\0';		/* tie off host */
				/* copy host */
      adr->host = rfc822_cpy (string);
      *end = c;			/* restore delimiter */
      if (!adr->personal) {	/* see if we can find a person name here */
	while ((*end == ' ') || (*end == '\t')) ++end;
				/* found something that may be a name */
	if ((*end == '(') && (s = rfc822_parse_phrase (++end))) {
	  rfc822_skipws (&s);	/* heinous kludge for trailing WS in comment */
	  if (*s == ')') {	/* good end of comment? */
	    *s++ = '\0';	/* yes, tie off the likely name and copy */
	    adr->personal = rfc822_cpy (end);
	    end = s;		/* skip after it */
	  }
	}
	rfc822_skipws (&end);	/* skip any other WS in the normal way */
      }
      *ret = end;		/* set new end pointer */
    }
				/* bogus input, shove in default */
    else adr->host = cpystr (defaulthost);
  }
  else {			/* no host name */
    *ret = end;			/* update return pointer */
				/* and default host to current host */
    adr->host = cpystr (defaulthost);
  }
  if (**ret == '\0') *ret = NIL;/* wipe pointer if at end of string */
  return adr;			/* return the address we got */
}

/* Parse RFC822 phrase
 * Accepts: string pointer
 * Returns: pointer to end of phrase
 */

char *rfc822_parse_phrase (char *s)
{
  char *curpos;
  if (!s) return NIL;		/* no-op if no string */
				/* find first word of phrase */
  curpos = rfc822_parse_word (s,NIL);
  if (!curpos) return NIL;	/* no words means no phrase */
				/* check if string ends with word */
  if (*curpos == '\0') return curpos;
  s = curpos;			/* sniff past the end of this word and WS */
  rfc822_skipws (&s);		/* skip whitespace */
				/* recurse to see if any more */
  return (s = rfc822_parse_phrase (s)) ? s : curpos;
}

/* Parse RFC822 word
 * Accepts: string pointer
 * Returns: pointer to end of word
 */

char *rfc822_parse_word (char *s,char *delimiters)
{
  char *st,*str;
  if (!s) return NIL;		/* no-op if no string */
  rfc822_skipws (&s);		/* flush leading whitespace */
  if (*s == '\0') return NIL;	/* end of string */
				/* default delimiters to standard */
  if (!delimiters) delimiters = " ()<>@,;:\"[\\]";
  str = s;			/* hunt pointer for strpbrk */
  while (T) {			/* look for delimiter */
    if (!(st = strpbrk (str,delimiters))) {
      while (*s != '\0') ++s;	/* no delimiter, hunt for end */
      return s;			/* return it */
    }
    switch (*st) {		/* dispatch based on delimiter */
    case '"':			/* quoted string */
				/* look for close quote */
      while (*++st != '"') switch (*st) {
      case '\0':		/* unbalanced quoted string */
	return NIL;		/* sick sick sick */
      case '\\':		/* quoted character */
	++st;			/* skip the next character */
      default:			/* ordinary character */
	break;			/* no special action */
      }
      str = ++st;		/* continue parse */
      break;
    case '\\':			/* quoted character */
      str = st + 2;		/* skip quoted character and go on */
      break;
    default:			/* found a word delimiter */
      return (st == s) ? NIL : st;
    }
  }
}

/* Copy an RFC822 format string
 * Accepts: string
 * Returns: copy of string
 */

char *rfc822_cpy (char *src)
{
  char *ret = cpystr (src);
  if (strpbrk (src,"\\\"")) {	/* any quoting in string? */
    char *dst = ret;
    while (*src) {		/* copy string */
      if (*src == '\"') src++;	/* skip double quote entirely */
      else {
	if (*src == '\\') src++;/* skip over single quote, copy next always */
	*dst++ = *src++;	/* copy character */
      }
    }
    *dst = '\0';		/* tie off string */
  }
  return ret;			/* return our string */
}


/* Copy address list
 * Accepts: address list
 * Returns: address list
 */

ADDRESS *rfc822_cpy_adr (ADDRESS *adr)
{
  ADDRESS *dadr;
  ADDRESS *ret = NIL;
  ADDRESS *prev = NIL;
  while (adr) {			/* loop while there's still an MAP adr */
    dadr = mail_newaddr ();	/* instantiate a new address */
    if (!ret) ret = dadr;	/* note return */
    if (prev) prev->next = dadr;/* tie on to the end of any previous */
    dadr->personal = cpystr (adr->personal);
    dadr->route_list = cpystr (adr->route_list);
    dadr->mail_box = cpystr (adr->mail_box);
    dadr->host = cpystr (adr->host);
    prev = dadr;		/* this is now the previous */
    adr = adr->next;		/* go to next address in list */
  }
  return (ret);			/* return the MTP address list */
}

/* Skips RFC822 whitespace
 * Accepts: pointer to string pointer
 */

void rfc822_skipws (char **s)
{
  int n = 0;
				/* while whitespace or start of comment */
  while ((**s == ' ') || (**s == '\t') || (n = (**s == '('))) {
    if (**s == '(') while (n) {	/* while comment in effect */
      switch (*++*s) {		/* get character of comment */
      case '(':			/* nested comment? */
	n++;			/* increment nest count */
	break;
      case ')':			/* end of comment? */
	n--;			/* decrement nest count */
	break;
      case '\0':		/* end of string */
	return;			/* this is wierd if it happens */
      case '"':			/* quoted string? */
				/* skip to close quote */
	while (*++*s != '\"') if (**s == '\\' && *++*s == '\0') return;
	break;
      case '\\':		/* quote next character? */
	if (*++*s == '\0') return;
	break;
      default:			/* random character */
	break;
      }
    }
    ++*s;			/* skip past whitespace character */
  }
}
