/*
 * Program:	Distributed Electronic Mail Manager (ReadWindow object)
 *
 * Author(s):	Mark Crispin
 *		Computing And Communications
 *		University of Washington
 *		Seattle, WA  98195
 *
 * Date:	24 February 1989
 *
 * Copyright (c) 1989, 1990, 1991 University of Washington
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license.
 *
 */


#import "MailManager.h"

@implementation ReadWindow

// Create a new ReadWindow

+ new
{
  self = [super new];		// instantiate ourselves
  [NXApp loadNibSection:"ReadWindow.nib" owner:self withNames:NIL];
  handle = NIL;			// no handle initially
  sequence = NIL;		// no sequence initially
  return self;
}


// Note our messageView

- setMessageView:anObject
{
  NXRect frame;			// note our view and get its frame
  [(messageView = [anObject docView]) getFrame:&frame];
				// reset its font
  [messageView renewFont:defaultfont text:"" frame:&frame tag:0];
  return self;
}


// Note our window

- setWindow:anObject
{
  window = anObject;		// note our window
  [window setDelegate:self];	// we're the window's delegate
  return self;
}

// Non-interface methods


// Note calling window message items view

- setMessageItems:anObject
{
  messageItems = anObject;	// note caller's message items view
  return self;
}


// Note our MAIL handle

- setHandle:(MAILHANDLE *) h
{
  mail_free_handle (&handle);	// flush any old handle
  handle = h;			// note MAIL handle
  return self;
}

// Set message sequence

- setSequence:(SEQUENCE *) seq
{
  SEQUENCE *s;
  if (s = sequence) {		// find head of old sequence list
    while (s->previous) s = s->previous;
    do {
      mail_free_elt (&s->elt);	// flush the cache entry in case nuked
      sequence = s->next;	// get the next in the list
      fs_give ((void **) &s);	// now free that sequence
    } while (s = sequence);	// run down the sequence list
  }
  return [self moveSequence:seq];
}


// Move message sequence

- moveSequence:(SEQUENCE *) seq
{
  int m;
  id brw;
  if (sequence = seq) {		// set new sequence, update display if non-NIL
    if (lockedread &&		// highlight browser message if locked read
	(seq->elt->msgno == [[(brw = [messageItems docView])
			     cellAt:(m = seq->elt->msgno - 1) :0] msgno]))
      [[brw selectCellAt:m :0] scrollCellToVisible: m :0];
    [self updateTitle];		// show the first message's title
    [self displayMessage:self];	// now display the message
    [self updateTitle];		// update the window's title now that read
    [window orderFront:self];	// reopen window in case taken out
  }
  return self;
}

// Update window title after action taken to change status

- updateTitle
{
  char text[TMPLEN];
  MAILSTREAM *stream = mail_stream (handle);
				// stream and message must be valid
  if (stream && sequence->elt->msgno) {
				// get the window's title
    headerline (text,stream,sequence->elt);
    [window setTitle:text];	// set the title
				// update primary window
    [getstreamprop (stream)->window updateTitle];
  }
				// update Flagged switch
  [flagged setIntValue:sequence->elt->flagged];
				// update Deleted switch
  [deleted setIntValue:sequence->elt->deleted];
				// update Seen switch
  [seen setIntValue:sequence->elt->seen];
				// update Answered switch
  [answered setIntValue:sequence->elt->answered];
  return self;
}

// Message display operations


// Note our display

- setDisplay:anObject
{
				// set according to the default
  [(display = anObject) selectCellAt:literaldisplay :0];
  return self;
}


// Note attachment panel

- setAttachmentPanel:anObject
{
  attachmentPanel = anObject;	// note attachment panel
  return self;
}


// Note attachment browser

- setAttachmentView:anObject
{
  NXSize size;
  NXRect frame;
  [anObject getFrame:&frame];	// get frame of attachments view
				// create a scroll view
  [(attachmentView = [ScrollView newFrame:&frame]) setBorderType:NX_BEZEL];
				// only vertical scrolling is required
  [attachmentView setVertScrollerRequired:T];
				// allow its size to change
  [attachmentView setAutosizing:NX_HEIGHTSIZABLE|NX_WIDTHSIZABLE];
				// put up the view
  [[anObject superview] replaceSubview:anObject with:attachmentView];
				// get the size minus the scroll bars
  [attachmentView getContentSize:&size];
				// set it as such in the browser frames
  NX_WIDTH (&frame) = size.width = 1300.0; NX_HEIGHT (&frame) = size.height;
				// create the browser
  browser = [Matrix newFrame:&frame mode:NX_RADIOMODE
	     cellClass:[ActionCell class] numRows:0 numCols:0];
  size.height = 13.0;		// large enough to hold the text
  [browser setCellSize:&size];	// set size of each cell
  size.width = size.height = 0;	// no intercell spacing
  [browser setIntercell:&size];
  [browser setAutoscroll:T];	// make it autoscrolling
  [[browser setAction:@selector(doAttachment:)] setTarget:self];
				// put the browser in the scroller and display
  [[attachmentView setDocView:browser] display];
  return self;
}

// Here is where we actually display a message

- displayMessage:sender
{
  char *s;
  int i = 0;
  int m = sequence->elt->msgno;
  int lit = [display selectedRow];
  const char *att[] = {"Text data","Encapsulated Message","Hex binary data",
		       "Electronic VFU (FORTRAN carriage control",
		       "Electronic Document Interchange","X.400 (unconverted)",
		       "Data file","Encrypted (RFC-1115)","Unknown"};
  NXRect frame;
  MAILSTREAM *stream = mail_stream (handle);
  ENVELOPE *env = sequence->elt->env;
  BODY *body = sequence->elt->body;
  if (stream && m) {		// must have open stream and not be expunged
				// get frame, don't update display until ready
    [[messageView getFrame:&frame] setAutodisplay:NIL];
				// initialize view
    [messageView renewFont:defaultfont text:"" frame:&frame tag:0];
				// install header
    [[messageView setSel:0 :0] replaceSel:(s = (lit || !env) ?
     fixnl (cpystr (mail_fetchheader (stream,m))) : filtered_header (env))];
    fs_give ((void **) &s);	// done with header string
    if ((!lit) && body && ((i = (body->type != ENCTEXT)) || body->next)) {
				// get initial text segment
      if (!i) s = fixnl (cpystr (mail_fetchbody (stream,m,0)));
				// resize to number of attachments
      while (body = body->next) i++;
      [browser renewRows:i cols:1];
      for (i = 0,body = s ? sequence->elt->body->next : sequence->elt->body;
	   body; i++,body = body->next)
	[[[browser cellAt:i :0] setStringValueNoCopy:att[body->type]]
	 setTag:s ? i + 1 : i];	// set text and tag for each attachment
      [[browser sizeToCells] display];
				// make attachments visible
      [[attachmentPanel attachWindow:window] orderFront:self];
    }
    else {			// literal or no attachments, fetch entire text
      s = fixnl (cpystr (mail_fetchtext (stream,m)));
      [attachmentPanel orderOut:self];
    }
    if (s) {			// insert body text after header
      [messageView replaceSel:s];
      fs_give ((void **) &s);
    }
				// put cursor at start, display text
    [[[messageView setSel:0 :0] display] setAutodisplay:T];
  }
  return self;
}

// Process an attachment

- doAttachment:sender
{
  char tmp[TMPLEN];
  char *s,*t,*t1;
  char *file = NIL;
  short f;
  int a = [[browser selectedCell] tag];
  int m = sequence->elt->msgno;
  BODY *en = sequence->elt->body;
  MAILSTREAM *stream = mail_stream (handle);
  SavePanel *dest = [SavePanel new];
  FILE *pipe;
				// must have stream, valid message, attachment
  if (stream && m && (s = fixnl (cpystr (mail_fetchbody (stream,m,a))))) {
    while (a--) en = en->next;	// find encoding type
    switch (en->type) {		// dispatch based on type
    case ENCTEXT:		// text (no particular encoded format)
    case ENCMESSAGE:		// encapsulated message
    case ENCEVFU:		// electronic VFU (FORTRAN carriage control)
      file = [self writeText:s using:dest];
      break;
    case ENCEDI:		// electronic document interchange
     if (NXRunAlertPanel ("Electronic Document Interchange",
			   "Write EDI data as text?",NIL,"Cancel",NIL))
	file = [self writeText:s using:dest];
      break;
    case ENCX400:		// CCITT X.400 unconverted section
     if (NXRunAlertPanel ("X.400","Write unconverted X.400 data as text?",
			  NIL,"Cancel",NIL))
	file = [self writeText:s using:dest];
      break;
    case ENCHEX:		// binary data in hexadecimal
     if (NXRunAlertPanel ("Hexadecimal binary",
			  "Write hex binary as text?",NIL,"Cancel",NIL))
	file = [self writeText:s using:dest];
      break;
    case ENCENCRYPTED:		// RFC 1115 encryption
     if (NXRunAlertPanel ("RFC-1115 Encryption",
			  "Write encrypted data as text?",NIL,"Cancel",NIL))
	file = [self writeText:s using:dest];
      break;

    case ENCUUENCODE:		// uuencoded data
      if (s[0] == 'b' && s[1] == 'e' && s[2] == 'g'&& s[3] == 'i' &&
	  s[4] == 'n' && s[5] == ' ' && (t1 = strchr (s + 6,' ')) &&
	  (t = strchr (s,'\n')) && t > t1) {
	*t++ = *t1++ = '\0';	// tie off header prefix & file strings
				// get file mode
	a = strtol (s + 6,NIL,8);
				// note if compressed
	if (f = (t[-2] == 'Z' && t[-3] == '.')) t[-3] = '\0';
				// get file name from user
	if (![dest runModalForDirectory:[dest directory]
	      file:strchr (t1,'/') ? strrchr (t1,'/') + 1 : t1]) break;
				// start pipe on uudecode
	if (!(pipe = popen ("uudecode","w")))
	  NXRunAlertPanel ("Decode Failed","Can't run uudecode",NIL,NIL,NIL);
	else {			// make file name for uudecode of attachment
	  fprintf (pipe,"begin 600 %s%s\n",(file = (char *) [dest filename]),
		   f ? ".Z" : "");
	  fputs (t,pipe);	// write the uuencode data
	  fflush (pipe);	// flush the buffer
	  pclose (pipe);	// close the pipe
	  if (f) {		// data compressed?
	    sprintf (tmp,"uncompress %s",file);
	    if (system (tmp))	// do the uncompression
	      NXRunAlertPanel ("Uncompress Failed",tmp,NIL,NIL,NIL);
	  }
	  if (a) chmod (file,a);// set file protection if specified
	}
      }
      else if (NXRunAlertPanel ("Bad uuencode",
				"Write bad format uuencode data as text?",
				NIL,"Cancel",NIL))
	file = [self writeText:s using:dest];
      break;
    case ENCUNDEF:		// undefined type
    default:			// supposedly impossible
      if (NXRunAlertPanel ("Unknown","Write unknown format data as text?",
			   NIL,"Cancel",NIL))
	file = [self writeText:s using:dest];
      break;
    }
				// if user wants to open the attachment
    if (file && NXRunAlertPanel ("Open","Open attachment?",NIL,"No",NIL)) {
      sprintf (tmp,"open %s",file);
      system (tmp);		// there's gotta be a better way to do this
    }
    fs_give ((void **) &s);	// flush the attachment
  }
  return self;
}

// Write text into a file, return file name

- (char *) writeText:(char *) text using:(SavePanel *) dest
{
  char *s = NIL;
  NXStream *file;
  if ([dest runModal]) {	// user really want to do this?
				// get memory stream for write
    if (!(file = NXOpenMemory (NIL,0,NX_WRITEONLY)))
      mm_log ("Can't open memory stream to write attachment",ERROR);
    else {			// write the data
      NXWrite (file,text,strlen (text));
      NXFlush (file);		// dump out the file
      if (NXSaveToFile (file,(text = (char *) [dest filename])))
	mm_log ("Can't write attachment",ERROR);
      else s = text;		// success, set up returned file name value
				// flush the memory stream
      NXCloseMemory (file,NX_FREEBUFFER);
    }
  }
  return s;			// return file name to caller
}

// We are ordered to close

- close
{
  [self windowWillClose:self];	// do cleanup
  [window close];		// close our window
  window = NIL;
  return self;
}


// Set FreeWhenClosed state for our window

- setFreeWhenClosed:(BOOL) flag
{
				// set the state
  [window setFreeWhenClosed:flag];
  return self;
}


// Window is closing

- windowWillClose:sender
{
  [attachmentPanel close];	// nuke attachment panel
  attachmentPanel = NIL;
  [self setSequence:NIL];	// flush the sequence
  [self setHandle:NIL];		// flush the handle
  return self;
}


// Window moved

- windowDidMove:sender
{
				// make attachment panel track this window
  [attachmentPanel attachWindow:window];
  return self;
}

// Miscellaneous and message motion buttons


// Open the help panel

- help:sender
{
  [NXApp readHelp:sender];
  return self;
}


// Print message

- print:sender
{
				// print what's in the message view
  [messageView printPSCode:self];
  return self;
}


// Kill button - delete this message, move to next

- kill:sender
{
  [deleted setIntValue:T];	// mark that we want deletion
  [self delete:self];		// now delete the message
				// move to next message, quit if at end
  if ((![self next:self]) && closefinalkill) [self close];
  return self;
}

// Next button - move to next message

- next:sender
{
  int i;
  MAILSTREAM *stream = mail_stream (handle);
  SEQUENCE *seq = sequence;
  if (stream) {
    while (seq->next) {		// if there's a next
				// if not expunged we can use it
      if (seq->next->elt->msgno) break;
      else seq = seq->next;	// otherwise skip over it
    }
				// if have a useful message use it
    if (seq->next) [self moveSequence:seq->next];
    else {
				// get this message number
      i = sequence->elt->msgno;
				// if single message sequence, not expunged,
				//  and not at end of mail file
      if (sequence->previous || (i == 0) || (i >= stream->nmsgs)) {
	mm_log ("No next message",WARN);
	return NIL;
      }
      else {			// we can do motion within the mail file
				// don't need this elt any more
	mail_free_elt (&sequence->elt);
				// move to next message
	sequence->elt = mail_elt (stream,++i);
				// lock down this elt
	sequence->elt->lockcount++;
				// now move to that message
	[self moveSequence:sequence];
      }
    }
  }
  return self;
}

// Previous button - move to previous message

- previous:sender
{
  int i;
  MAILSTREAM *stream = mail_stream (handle);
  SEQUENCE *seq = sequence;
  if (stream) {
    while (seq->previous) {	// if there's a previous
				// if not expunged we can use it
      if (seq->previous->elt->msgno) break;
      else seq = seq->previous;	// otherwise skip over it
    }
				// if have a useful message use it
    if (seq->previous) [self moveSequence:seq->previous];
    else {
				// get this message number
      i = sequence->elt->msgno;
				// if single message sequence, not expunged,
				//  and not at top of mail file
      if (sequence->next || (i <= 1)) mm_log ("No previous message",WARN);
      else {			// we can do motion within the mail file
				// don't need this elt any more
	mail_free_elt (&sequence->elt);
				// move to previous message
	sequence->elt = mail_elt (stream,--i);
				// lock down this elt
	sequence->elt->lockcount++;
				// now move to that message
	[self moveSequence:sequence];
      }
    }
  }
  return self;
}

// System flag switches


// Note Answered switch

- setAnswered:anObject;
{
  answered = anObject;		// note Answered
  return self;
}


// Note Flagged switch

- setFlagged:anObject;
{
  flagged = anObject;		// note Flagged
  return self;
}


// Note Deleted switch

- setDeleted:anObject;
{
  deleted = anObject;		// note Deleted
  return self;
}


// Note Seen switch

- setSeen:anObject;
{
  seen = anObject;		// note Seen
  return self;
}

// Answered switch - set the answered status

- answer:sender
{
  char seq[10];
  MAILSTREAM *stream = mail_stream (handle);
				// stream and message must be valid
  if (stream && sequence->elt->msgno) {
				// make string form
    sprintf (seq,"%d",sequence->elt->msgno);
				// only allow clearing this
    if (![answered intValue]) mail_clearflag (stream,seq,"\\Answered");
  }
  return [self updateTitle];
}


// Deleted switch - set the deleted status

- delete:sender
{
  char seq[10];
  MAILSTREAM *stream = mail_stream (handle);
				// stream and message must be valid
  if (stream && sequence->elt->msgno) {
				// make string form
    sprintf (seq,"%d",sequence->elt->msgno);
				// set state per the switch
    if ([deleted intValue]) mail_setflag (stream,seq,"\\Deleted");
    else mail_clearflag (stream,seq,"\\Deleted");
  }
  return [self updateTitle];
}

// Flag switch - set the flagged status

- flag:sender
{
  char seq[10];
  MAILSTREAM *stream = mail_stream (handle);
				// stream and message must be valid
  if (stream && sequence->elt->msgno) {
				// make string form
    sprintf (seq,"%d",sequence->elt->msgno);
				// set state per the switch
    if ([flagged intValue]) mail_setflag (stream,seq,"\\Flagged");
    else mail_clearflag (stream,seq,"\\Flagged");
  }
  return [self updateTitle];
}


// Seen switch - set the seen status

- mark:sender
{
  char seq[10];
  MAILSTREAM *stream = mail_stream (handle);
				// stream and message must be valid
  if (stream && sequence->elt->msgno) {
				// make string form
    sprintf (seq,"%d",sequence->elt->msgno);
				// set state per the switch
    if ([seen intValue]) mail_setflag (stream,seq,"\\Seen");
    else mail_clearflag (stream,seq,"\\Seen");
  }
  return [self updateTitle];
}

// Keyword operations


// Note our Keyword button

- setKeyword:anObject
{
  id l = [PullOutMenu new:(keyword = anObject)];
  [l addItem:"Keyword..." action:@selector(keywordSet:)];
  [l addItem:"Clear Keyword..." action:@selector(keywordClear:)];
  return self;
}


// Set the selected keywords in selected messages

- keywordSet:sender
{
  return [self keyword:sender set:T];
}


// Clear the selected keywords in selected messages

- keywordClear:sender
{
  return [self keyword:sender set:NIL];
}

// Change keywords

- keyword:sender set:(BOOL) set
{
  char seq[10];
  char *keys;
  MAILSTREAM *stream = mail_stream (handle);
				// stream and message must be valid
  if (stream && sequence->elt->msgno &&
       (keys = [(getstreamprop (stream))->window getKeywords])) {
				// make string form of this sequence
    sprintf (seq,"%d",sequence->elt->msgno);
				// set the flag, if we can
    if (set) mail_setflag (stream,seq,keys);
    else mail_clearflag (stream,seq,keys);
    fs_give ((void **) &keys);
  }
  return [self updateTitle];	// update title with new status
}

// File operations


// Note our Copy button

- setCopy:anObject
{
  id l = [PullOutMenu new:(copy = anObject)];
  [l addItem:"Copy..." action:@selector(fileCopy:)];
  [l addItem:"Move..." action:@selector(fileMove:)];
  return self;
}


// Copy message to another mailbox

- fileCopy:sender
{
  return [self copy:sender delete:NIL];
}


// Move (copy + delete) message to another mailbox

- fileMove:sender
{
  [self copy:sender delete:T];
  if ((![self next:self]) && closefinalkill) [self close];
  return self;
}

// Do the copy or move

- copy:sender delete:(BOOL) del
{
  char seq[10];
  char *dest;
  FILE *file;
  BODY body;
  MESSAGECACHE *elt = sequence->elt;
  MAILSTREAM *stream = mail_stream (handle);
				// get file
  if (stream && elt->msgno && (dest = [NXApp getFile])) {
    if (*dest == '*') {		// want local copy
      if (!(file = fopen (dest+1,"a")))
	NXRunAlertPanel ("Open failed","Can't open local file",NIL,NIL,NIL);
      else {			// blat it to the file
	body.type = ENCTEXT;	// build fake body
	body.options = NIL;
	body.text = cpystr (mail_fetchtext (stream,elt->msgno));
	body.next = NIL;
	append_msg (file,elt->env ? elt->env->sender : NIL,elt->internal_date,
		    mail_fetchheader (stream,elt->msgno),&body);
	fs_give ((void **) &body.text);
	fclose (file);		// close off the file
	if (del) {		// if called by move:
				// mark for deletion
	  [deleted setIntValue:T];
	  [self delete:self];	// now delete the message
	}
      }
    }
    else {			// want remote copy, make string form of msgno
      sprintf (seq,"%d",elt->msgno);
				// now do the copy
      if (del) mail_move (stream,seq,dest+1);
      else mail_copy (stream,seq,dest+1);
    }
    fs_give ((void **) &dest);	// flush the destination file string
  }
  return [self updateTitle];
}

// Message forwarding


// Note our Forward button

- setForward:anObject;
{
  id l = [PullOutMenu new:(forward = anObject)];
  [l addItem:"Forward..." action:@selector(forward:)];
  [l addItem:"Remail..." action:@selector(remail:)];
  return self;
}


// Forward button - forward a message

- forward:sender
{
  char tmp[TMPLEN],t[TMPLEN];
  char *s;
  id compose;
  id body;
  int m = sequence->elt->msgno;
  ENVELOPE *msg;
  ENVELOPE *env = sequence->elt->env;
  MAILSTREAM *stream = mail_stream (handle);
  if (stream && m) {		// have to have a stream to do this
    strcpy (t,sequence->elt->from);
				// tie off trailing spaces
    for (s = t+FROMLEN-2; *s == ' '; --s) *s = '\0';
    sprintf (tmp,"[%s: %s]",t,(env && env->subject) ?
	     env->subject : "(forwarded message)");
    compose = [NXApp compose:self];
    msg = [compose msg];
    msg->subject = cpystr (tmp);// set up Subject
    [compose updateEnvelopeView];
    body = [compose bodyView];	// get the message text body view
    [body selectAll:sender];	// prepare to initialize text
    [body replaceSel:"\n ** Begin Forwarded Message **\n\n"];
				// insert message
    [body replaceSel:(s = ([display selectedRow] || !env) ?
     fixnl (cpystr (mail_fetchheader (stream,m))) : filtered_header (env))];
    fs_give ((void **) &s);
    [body replaceSel:(s = fixnl (cpystr (mail_fetchtext (stream,m))))];
    fs_give ((void **) &s);	// don't need space any more
    [body setSel:0:0];		// put cursor at start of message
    [body display];		// update the display
    [compose selectBody];	// now select the appropriate field
  }
  return self;
}

// Remail button - remail a message

- remail:sender
{
  char *s;
  id compose;
  id body;
  int msgno = sequence->elt->msgno;
  ENVELOPE *msg;
  MAILSTREAM *stream = mail_stream (handle);
  if (stream && msgno) {	// have to have a stream and msgno to do this
    compose = [NXApp compose:self];
    msg = [compose msg];
				// get original header
    msg->remail = cpystr (mail_fetchheader (stream,msgno));
    body = [compose bodyView];	// get the message text body view
    [body selectAll:sender];	// prepare to initialize text
    [body replaceSel:(s = fixnl (cpystr (mail_fetchtext (stream,msgno))))];
    fs_give ((void **) &s);
    [body setSel:0:0];		// put cursor at start of message
    [body setEditable:NIL];	// can't change remail text
    [body display];		// update the display
    [compose selectBody];	// now select the appropriate field
  }
  return self;
}

// Message reply operations



// Note our Reply button

- setReply:anObject;
{
  id l = [PullOutMenu new:(reply = anObject)];
  [l addItem:"Reply..." action:@selector(replySender:)];
  [l addItem:"Reply to All..." action:@selector(replyAll:)];
  return self;
}


// Reply to All button - reply to all recipients of message

- replyAll:sender
{
  return [self reply:sender all:T];
}


// Reply to Sender button - reply to sender of message

- replySender:sender
{
  return [self reply:sender all:NIL];
}

// Reply to message

- reply:sender all:(BOOL) all
{
  char *s,*d;
  int i = [messageView textLength];
  ENVELOPE *env = sequence->elt->env;
  MAILSTREAM *stream = mail_stream (handle);
  d = (s = fs_get (1+i));	// get message we're replying to
  [messageView getSubstring:s start:0 length:i];
  s[i] = '\0';			// tie off string
				// search for double newline
  while (*d) if (*d++ == '\n' && *d == '\n') break;
  if (*d) d++;			// skip over the second one if found it
  else d = s;			// else use original string
				// create a compose window under it
  [[ReplyWindow new:(stream ? mail_makehandle (stream) : NIL) elt:sequence->elt
    text:d all:all] attachWindow:window];
  fs_give ((void **) &s);	// flush the text
}


@end
