/*
 * Program:	Berkeley mail (BEZERK) routines
 *
 * Author(s):	Mark Crispin
 *		Computing And Communications
 *		University of Washington
 *		Seattle, WA  98195
 *
 * Date:	20 December 1989
 *
 * Copyright 1989, 1990, 1991 by University of Washington.
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license.
 *
 */


#include <stdio.h>
#include <ctype.h>
#ifdef __STDC__
#include <string.h>
#include <stdlib.h>
#else
#include "ansi.h"
#endif
#include <pwd.h>
#include <netdb.h>
#include <errno.h>
extern int errno;		/* just in case */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/uio.h>
#include "mail.h"
#include "bezerk.h"
#include "rfc822.h"
#include "misc.h"
#include "osdep.h"

/* Berkeley mail routines */


/* Driver dispatch used by MAIL */

DRIVER bezerkdriver = {
  (DRIVER *) NIL,		/* next driver */
  bezerk_valid,			/* mailbox is valid for us */
  bezerk_open,			/* open mailbox */
  bezerk_close,			/* close mailbox */
  bezerk_fetchfast,		/* fetch message "fast" attributes */
  bezerk_fetchflags,		/* fetch message flags */
  bezerk_fetchenvelope,		/* fetch message envelopes */
  bezerk_fetchheader,		/* fetch message header only */
  bezerk_fetchtext,		/* fetch message body only */
  bezerk_fetchbody,		/* fetch message body section */
  bezerk_setflag,		/* set message flag */
  bezerk_clearflag,		/* clear message flag */
  bezerk_search,		/* search for message based on criteria */
  bezerk_ping,			/* ping mailbox to see if still alive */
  bezerk_check,			/* check for new messages */
  bezerk_expunge,		/* expunge deleted messages */
  bezerk_copy,			/* copy messages to another mailbox */
  bezerk_move			/* move messages to another mailbox */
};

/* Berkeley mail validate mailbox
 * Accepts: mailbox name
 * Returns: our driver if name is valid, otherwise calls valid in next driver
 */

/* Test for valid header */

#define VALID (s[1] == 'r') && (s[2] == 'o') && (s[3] == 'm') && \
  (s[4] == ' ') && (t = strchr (s,'\n')) && (t-s >= 27) && (t[-5] == ' ') && \
  (t[(rn = -4 * (t[-9] == ' '))-8] == ':') && \
  ((sysv = 3 * (t[rn-11] == ' ')) || t[rn-11] == ':') && \
  (t[rn+sysv-14] == ' ') && (t[rn+sysv-17] == ' ') && \
  (t[rn+sysv-21] == ' ')

DRIVER *bezerk_valid (name)
	char *name;
{
  int fd;
  int rn,sysv;
  char *t;
  char s[MAILTMPLEN];
  struct stat sbuf;
				/* if file, get its status */
  if (*name != '{' && (stat (bezerk_file (s,name),&sbuf) == 0)) {
    if (sbuf.st_size != 0) {	/* if non-empty file */
      if ((fd = open (s,O_RDONLY,NIL)) >= 0 && read (fd,s,MAILTMPLEN-1) >= 0) {
	close (fd);		/* close the file */
				/* must be valid header */
	if ((s[0] == 'F') && VALID) return &bezerkdriver;
      }
    }
				/* allow if not a ".txt" file */
    else if (!strstr (name,".txt")) return &bezerkdriver;
  }
				/* else let next driver have a try */
  return bezerkdriver.next ? (*bezerkdriver.next->valid) (name) : NIL;
}

/* Berkeley mail open
 * Accepts: stream to open
 * Returns: stream on success, NIL on failure
 */

MAILSTREAM *bezerk_open (stream)
	MAILSTREAM *stream;
{
  int i;
  int fd;
  char tmp[MAILTMPLEN];
  struct hostent *host_name;
  if (LOCAL) {			/* close old file if stream being recycled */
    bezerk_close (stream);	/* dump and save the changes */
    stream->dtb = &bezerkdriver;/* reattach this driver */
				/* clean up messagearray (again) */
    mail_free_messagearray (stream->messagearray);
  }
  stream->local = fs_get (sizeof (BEZERKLOCAL));
				/* canonicalize the stream mailbox name */
  bezerk_file (tmp,stream->mailbox);
  fs_give ((void **) &stream->mailbox);
  stream->mailbox = cpystr (tmp);
  LOCAL->fname = cpystr (stream->mailbox);
  gethostname(tmp,MAILTMPLEN);	/* get local host name */
  LOCAL->host = cpystr ((host_name = gethostbyname (tmp)) ?
			host_name->h_name : tmp);
				/* build name of our lock file */
  sprintf (tmp,"/tmp/.%s",LOCAL->fname);
  for (i = 6; i < strlen (tmp); ++i) if (tmp[i] == '/') tmp[i] = '\\';
  LOCAL->lname = cpystr (tmp);	/* note name for later */
  LOCAL->ld = NIL;		/* no state locking yet */
  LOCAL->filesize = 0;		/* initialize file information */
  LOCAL->filetime = 0;
  for (i = 0; i < MAXMESSAGES; ++i) LOCAL->msgs[i] = NIL;
  LOCAL->text = NIL;		/* no text yet */
  LOCAL->textsize = LOCAL->textend = 0;
  LOCAL->lsjumb = "BEAT CAL";	/* Go Big Red !!! */
  LOCAL->buf = (char *) fs_get ((LOCAL->buflen = MAXMESSAGESIZE) + 1);
  stream->sequence++;		/* bump sequence number */

				/* parse mailbox */
  stream->nmsgs = stream->recent = 0;
  if ((fd = bezerk_parse (stream,tmp)) >= 0) {
				/* will we be able to get write access? */
    if (access (LOCAL->fname,W_OK))
      mm_log ("Can't get write access to mailbox, access is readonly",WARN);
				/* yes, make lock for state */
    else if ((i = open (LOCAL->lname,O_RDONLY|O_CREAT,0666)) < 0)
      mm_log ("Can't open mailbox lock, access is readonly",WARN);
				/* try lock it */
    else if (flock (i,LOCK_EX|LOCK_NB)) {
      close (i);		/* couldn't lock, give up on it then */
      mm_log ("Other processes using this mailbox, access is readonly",WARN);
    }
				/* got the lock, nobody else can alter state */
    else fchmod ((LOCAL->ld = i),0666);
    bezerk_unlock (fd,tmp);
    mail_unlock (stream);
    if (!(stream->nmsgs || stream->debug == 36))
      mm_log ("Mailbox is empty",WARN);
  }
  return LOCAL ? stream : NIL;	/* if stream is alive, return to caller */
}

/* Berkeley mail close
 * Accepts: MAIL stream
 */

void bezerk_close (stream)
	MAILSTREAM *stream;
{
  int fd;
  char lock[MAILTMPLEN];
				/* is stream alive? */
  if (stream && LOCAL && LOCAL->ld) {
    LOCAL->lsjumb = NIL;	/* note this stream is dying */
				/* lock mailbox and parse new messages */
    if ((fd = bezerk_parse (stream,lock)) >= 0) {
      flock (fd,LOCK_EX);	/* upgrade the lock to exclusive access */
      bezerk_save (stream,fd);	/* dump any changes not saved yet */
      bezerk_unlock (fd,lock);	/* flush the lock */
      mail_unlock (stream);
    }
  }
  bezerk_abort (stream);	/* now punt the file and local data */
}


/* Berkeley mail fetch fast information
 * Accepts: MAIL stream
 *	    sequence
 */

void bezerk_fetchfast (stream,sequence)
	MAILSTREAM *stream;
	char *sequence;
{
  return;			/* no-op for local mail */
}


/* Berkeley mail fetch flags
 * Accepts: MAIL stream
 *	    sequence
 */

void bezerk_fetchflags (stream,sequence)
	MAILSTREAM *stream;
	char *sequence;
{
  return;			/* no-op for local mail */
}

/* Berkeley mail fetch envelope
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: envelope of this message
 *
 * Fetches the "fast" information as well
 */

ENVELOPE *bezerk_fetchenvelope (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  int i = LOCAL->msgs[msgno-1]->headersize;
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  if (i > LOCAL->buflen)	/* make sure buffer can hold the text */
    fs_resize ((void **) &LOCAL->buf,(LOCAL->buflen = i) + 1);
  if (!elt->env)		/* make envelope now if don't have one */
    rfc822_parse_msg (elt,LOCAL->text + LOCAL->msgs[msgno-1]->header,i,
		      LOCAL->host,LOCAL->buf);
  return elt->env;		/* return the envelope */
}


/* Berkeley mail snarf message, only for Tenex driver
 * Accepts: MAIL stream
 *	    message # to snarf
 *	    pointer to size to return
 * Returns: message text in RFC822 format
 */

char *bezerk_snarf (stream,msgno,size)
	MAILSTREAM *stream;
	int msgno;
	int *size;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  FILECACHE *m = LOCAL->msgs[msgno-1];
  				/* make sure stream can hold the text */
  if ((*size = m->headersize + m->bodysize) > LOCAL->buflen)
    fs_resize ((void **) &LOCAL->buf,(LOCAL->buflen = *size) + 1);
				/* copy the text */
  if (m->headersize) memcpy (LOCAL->buf,LOCAL->text + m->header,m->headersize);
  if (m->bodysize) memcpy (LOCAL->buf + m->headersize,
			   LOCAL->text + m->body,m->bodysize);
  LOCAL->buf[*size] = '\0';	/* tie off string */
  return LOCAL->buf;
}

/* Berkeley mail fetch message header
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message header in RFC822 format
 */

char *bezerk_fetchheader (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  FILECACHE *m = LOCAL->msgs[msgno-1];
				/* copy the string */
  return strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,
		     LOCAL->text + m->header,m->headersize);
}


/* Berkeley mail fetch message text (only)
	body only;
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message text in RFC822 format
 */

char *bezerk_fetchtext (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  FILECACHE *m = LOCAL->msgs[msgno-1];
  elt->seen = T;		/* mark message as seen */
				/* recalculate Status/X-Status lines */
  bezerk_update_status (m->status,elt);
  return strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,
		     LOCAL->text + m->body,m->bodysize);
}

/* Berkeley fetch message body as a structure
 * Accepts: Mail stream
 *	    message # to fetch
 *	    section #
 * Returns: pointer to section of message body
 */

char *bezerk_fetchbody (stream,msgno,sec)
	MAILSTREAM *stream;
	int msgno;
	int sec;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  FILECACHE *m = LOCAL->msgs[msgno-1];
  BODY *body;
  int i;
  char *s,*t,*v;
  if (sec < 0) {		/* validate section */
    sprintf (LOCAL->buf,"Bogus section number: %d",sec);
    fatal (LOCAL->buf);
  }
				/* make sure header parsed */
  bezerk_fetchenvelope (stream,msgno);
				/* treat as text if no structured body */
  if (!(body = elt->body)) return sec ? "" : bezerk_fetchtext (stream,msgno);
  elt->seen = T;		/* yes, mark message as seen */
				/* recalculate Status/X-Status lines */
  bezerk_update_status (m->status,elt);
  t = LOCAL->text + m->body;	/* start at body origin */
  v = t + m->bodysize;		/* absolute end of message */
  do {				/* count lines */
    s = t;			/* start of section */
				/* zero count means rest of message */
    if (!(i = body->count)) t = v;
				/* otherwise count lines */
    else while (i && (t < v)) if (*t++ == '\n') i--;
				/* until at the appropriate section */
  } while (sec-- && (!i) && (body = body->next) && (*t++ == '\n'));
				/* return the section requested if found */
  return (sec >= 0) ? "": strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,s,t - s);
}

/* Berkeley mail set flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void bezerk_setflag (stream,sequence,flag)
	MAILSTREAM *stream;
	char *sequence;
	char *flag;
{
  MESSAGECACHE *elt;
  unsigned char s[MAXMESSAGES];
  int i = stream->nmsgs;
  short f = bezerk_getflags (stream,flag);
  if (!f) return;		/* no-op if no flags to modify */
				/* get sequence and loop on it */
  if (bezerk_getsequence (stream,s,sequence)) while (i) if (s[--i]) {
    elt = mail_elt (stream,i+1);/* get cache entry */
    if (f&fSEEN) elt->seen = T;	/* set all requested flags */
    if (f&fDELETED) elt->deleted = T;
    if (f&fFLAGGED) elt->flagged = T;
    if (f&fANSWERED) elt->answered = T;
				/* recalculate Status/X-Status lines */
    bezerk_update_status (LOCAL->msgs[i]->status,elt);
  }
}


/* Berkeley mail clear flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void bezerk_clearflag (stream,sequence,flag)
	MAILSTREAM *stream;
	char *sequence;
	char *flag;
{
  MESSAGECACHE *elt;
  unsigned char s[MAXMESSAGES];
  int i = stream->nmsgs;
  short f = bezerk_getflags (stream,flag);
  if (!f) return;		/* no-op if no flags to modify */
				/* get sequence and loop on it */
  if (bezerk_getsequence (stream,s,sequence)) while (i) if (s[--i]) {
    elt = mail_elt (stream,i+1);/* get cache entry */
				/* clear all requested flags */
    if (f&fSEEN) elt->seen = NIL;
    if (f&fDELETED) elt->deleted = NIL;
    if (f&fFLAGGED) elt->flagged = NIL;
    if (f&fANSWERED) elt->answered = NIL;
				/* recalculate Status/X-Status lines */
    bezerk_update_status (LOCAL->msgs[i]->status,elt);
  }
}

/* Berkeley mail search for messages
 * Accepts: MAIL stream
 *	    search criteria
 */

void bezerk_search (stream,criteria)
	MAILSTREAM *stream;
	char *criteria;
{
  int i;
  int n;
  char *d;
  unsigned char s[MAXMESSAGES];
  char (*f) ();
  memset (s,T,MAXMESSAGES);
				/* get first criterion */
  if (criteria && (criteria = strtok (criteria," "))) {
				/* for each criterion */
    for (; criteria; (criteria = strtok (NIL," "))) {
      f = NIL; d = NIL; n = 0;	/* init then scan the criterion */
      switch (*ucase (criteria)) {
      case 'A':			/* possible ALL, ANSWERED */
	if (!strcmp (criteria+1,"LL")) f = bezerk_search_all;
	else if (!strcmp (criteria+1,"NSWERED")) f = bezerk_search_answered;
	break;
      case 'B':			/* possible BCC, BEFORE, BODY */
	if (!strcmp (criteria+1,"CC"))
	  f = bezerk_search_string (bezerk_search_bcc,&d,&n);
	else if (!strcmp (criteria+1,"EFORE"))
	  f = bezerk_search_date (bezerk_search_before,&n);
	else if (!strcmp (criteria+1,"ODY"))
	  f = bezerk_search_string (bezerk_search_body,&d,&n);
	break;
      case 'C':			/* possible CC */
	if (!strcmp (criteria+1,"C"))
	  f = bezerk_search_string (bezerk_search_cc,&d,&n);
	break;
      case 'D':			/* possible DELETED */
	if (!strcmp (criteria+1,"ELETED")) f = bezerk_search_deleted;
	break;
      case 'F':			/* possible FLAGGED, FROM */
	if (!strcmp (criteria+1,"LAGGED")) f = bezerk_search_flagged;
	else if (!strcmp (criteria+1,"ROM"))
	  f = bezerk_search_string (bezerk_search_from,&d,&n);
	break;
      case 'K':			/* possible KEYWORD */
	if (!strcmp (criteria+1,"EYWORD"))
	  f = bezerk_search_flag (bezerk_search_keyword,&d);
	break;
      case 'N':			/* possible NEW */
	if (!strcmp (criteria+1,"EW")) f = bezerk_search_new;
	break;

      case 'O':			/* possible OLD, ON */
	if (!strcmp (criteria+1,"LD")) f = bezerk_search_old;
	else if (!strcmp (criteria+1,"N"))
	  f = bezerk_search_date (bezerk_search_on,&n);
	break;
      case 'R':			/* possible RECENT */
	if (!strcmp (criteria+1,"ECENT")) f = bezerk_search_recent;
	break;
      case 'S':			/* possible SEEN, SINCE, SUBJECT */
	if (!strcmp (criteria+1,"EEN")) f = bezerk_search_seen;
	else if (!strcmp (criteria+1,"INCE"))
	  f = bezerk_search_date (bezerk_search_since,&n);
	else if (!strcmp (criteria+1,"UBJECT"))
	  f = bezerk_search_string (bezerk_search_subject,&d,&n);
	break;
      case 'T':			/* possible TEXT, TO */
	if (!strcmp (criteria+1,"EXT"))
	  f = bezerk_search_string (bezerk_search_text,&d,&n);
	else if (!strcmp (criteria+1,"O"))
	  f = bezerk_search_string (bezerk_search_to,&d,&n);
	break;
      case 'U':			/* possible UN* */
	if (criteria[1] == 'N') {
	  if (!strcmp (criteria+2,"ANSWERED")) f = bezerk_search_unanswered;
	  else if (!strcmp (criteria+2,"DELETED")) f = bezerk_search_undeleted;
	  else if (!strcmp (criteria+2,"FLAGGED")) f = bezerk_search_unflagged;
	  else if (!strcmp (criteria+2,"KEYWORD"))
	    f = bezerk_search_flag (bezerk_search_unkeyword,&d);
	  else if (!strcmp (criteria+2,"SEEN")) f = bezerk_search_unseen;
	}
	break;
      default:			/* we will barf below */
	break;
      }
      if (!f) {			/* if can't determine any criteria */
	sprintf (LOCAL->buf,"Unknown search criterion: %s",criteria);
	mm_log (LOCAL->buf,ERROR);
	return;
      }
				/* run the search criterion */
      for (i = 0; i<stream->nmsgs; ++i) s[i] = (s[i] && (*f) (stream,i+1,d,n));
    }
				/* report search results to main program */
    for (i = 0; i < stream->nmsgs; ++i) if (s[i]) mail_searched (stream,i+1);
  }
}

/* Berkeley mail ping mailbox
 * Accepts: MAIL stream
 * Returns: stream if still alive
 * No-op for readonly files, since read/writer can expunge it from under us!
 */

MAILSTREAM *bezerk_ping (stream)
	MAILSTREAM *stream;
{
  char lock[MAILTMPLEN];
  struct stat sbuf;
  int fd;
				/* make it is alright to do this at all */
  if (LOCAL && LOCAL->ld && !mail_checklock (stream)) {
				/* get current mailbox size */
    stat (LOCAL->fname,&sbuf);
				/* parse if mailbox changed */
    if ((sbuf.st_size != LOCAL->filesize) &&
	((fd = bezerk_parse (stream,lock)) >= 0)) {
      bezerk_unlock (fd,lock);	/* unlock mailbox */
      mail_unlock (stream);	/* and stream */
    }
  }
  return LOCAL ? stream : NIL;	/* return if still alive */
}


/* Berkeley mail check mailbox
 * Accepts: MAIL stream
 * No-op for readonly files, since read/writer can expunge it from under us!
 */

void bezerk_check (stream)
	MAILSTREAM *stream;
{
  char lock[MAILTMPLEN];
  int fd;
				/* parse and lock mailbox */
  if (stream && LOCAL && LOCAL->ld &&
      ((fd = bezerk_parse (stream,lock)) >= 0)) {
    flock (fd,LOCK_EX);		/* upgrade the lock to exclusive access */
    bezerk_save (stream,fd);	/* dump out the checkpointed version */
    bezerk_unlock (fd,lock);
    mail_unlock (stream);
    mm_log ("Check completed",NIL);
  }
}

/* Berkeley mail expunge mailbox
 * Accepts: MAIL stream
 */

void bezerk_expunge (stream)
	MAILSTREAM *stream;
{
  int j;
  int i = 1;
  int n = 0;
  MESSAGECACHE *elt;
  char lock[MAILTMPLEN];
  int fd;
				/* parse and lock mailbox first */
  if (stream && LOCAL && LOCAL->ld &&
      ((fd = bezerk_parse (stream,lock)) >= 0)) {
    flock (fd,LOCK_EX);		/* upgrade the lock to exclusive access */
    while (i <= stream->nmsgs) {/* flush deleted messages */
      if ((elt = mail_elt (stream,i))->deleted) {
				/* if recent, note one less recent message */
	if (elt->recent) --(stream->recent);
				/* flush local cache entry */
	fs_give ((void **) &LOCAL->msgs[i-1]);
	for (j = i; j<stream->nmsgs; j++) LOCAL->msgs[j-1] = LOCAL->msgs[j];
	LOCAL->msgs[stream->nmsgs-1] = NIL;
				/* notify upper levels */
	mail_expunged (stream,i);
	n++;			/* count another expunged message */
      }
      else i++;			/* otherwise try next message */
    }
    if (n) {			/* did we flush anything? */
      bezerk_save (stream,fd);	/* yes, dump a checkpoint */
      if (stream->debug != 36) {/* not if called from Tenex */
	sprintf (LOCAL->buf,"Expunged %d messages",n);
	mm_log (LOCAL->buf,NIL);/* tell user what we have done */
      }
    }
    else mm_log ("No messages deleted, so no update needed",NIL);
    if (stream->debug != 36) {	/* notify upper level of new mailbox size */
      mail_exists (stream,stream->nmsgs);
      mail_recent (stream,stream->recent);
    }
    bezerk_unlock (fd,lock);	/* free the lock */
    mail_unlock (stream);
  }
}

/* Berkeley mail copy message(s)
	s;
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 */

void bezerk_copy (stream,sequence,mailbox)
	MAILSTREAM *stream;
	char *sequence;
	char *mailbox;
{
  unsigned char s[MAXMESSAGES];
				/* copy the messages */
  if (bezerk_getsequence (stream,s,sequence))
    bezerk_copy_messages (stream,s,mailbox);
}


/* Berkeley mail move message(s)
	s;
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 */

void bezerk_move (stream,sequence,mailbox)
	MAILSTREAM *stream;
	char *sequence;
	char *mailbox;
{
  int i = 0;
  unsigned char s[MAXMESSAGES];
  MESSAGECACHE *elt;
  if (bezerk_getsequence (stream,s,sequence)) {
				/* copy the messages */
    bezerk_copy_messages (stream,s,mailbox);
				/* delete all requested messages */
    while (i < stream->nmsgs) if (s[i++]) {
      (elt = mail_elt (stream,i))->deleted = T;
				/* recalculate Status/X-Status lines */
      bezerk_update_status (LOCAL->msgs[i-1]->status,elt);
    }
  }
}

/* Internal routines */


/* Berkeley mail abort stream
 * Accepts: MAIL stream
 */

void bezerk_abort (stream)
	MAILSTREAM *stream;
{
  int i;
  if (stream && LOCAL) {	/* only if a file is open */
    if (LOCAL->host) fs_give ((void **) &LOCAL->host);
    if (LOCAL->ld) {		/* have a mailbox lock? */
      flock (LOCAL->ld,LOCK_UN);/* yes, release the lock */
      close (LOCAL->ld);	/* close the lock file */
      unlink (LOCAL->lname);	/* and delete it */
    }
    fs_give ((void **) &LOCAL->lname);
				/* free local cache */
    for (i = 0; i < MAXMESSAGES; ++i)
      if (LOCAL->msgs[i]) fs_give ((void **) &LOCAL->msgs[i]);
				/* free local text buffers */
    if (LOCAL->text) fs_give ((void **) &LOCAL->text);
    if (LOCAL->buf) fs_give ((void **) &LOCAL->buf);
    if (LOCAL->fname) fs_give ((void **) &LOCAL->fname);
				/* nuke the local data */
    fs_give ((void **) &stream->local);
    stream->dtb = NIL;		/* log out the DTB */
  }
}


/* Berkeley mail generate file string
 * Accepts: temporary buffer to write into
 *	    mailbox name string
 * Returns: local file string
 */


char *bezerk_file (dst,name)
	char *dst;
	char *name;
{
  struct passwd *pwd;
  char *uid;
  strcpy (dst,name);		/* copy the mailbox name */
  if (dst[0] != '/') {		/* pass absolute file paths */
				/* get user ID and directory */
    uid = (pwd = getpwuid (geteuid ()))->pw_name;
    if (strcmp (ucase (dst),"INBOX")) sprintf (dst,"%s/%s",pwd->pw_dir,name);
				/* INBOX becomes /usr/spool/mail/USER file */
    else sprintf (dst,"/usr/spool/mail/%s",uid);
  }
  return dst;
}

/* Berkeley open and lock mailbox
 * Accepts: file name to open/lock
 *	    file open mode
 *	    destination buffer for lock file name
 *	    type of locking operation (LOCK_SH or LOCK_EX)
 */

int bezerk_lock (file,flags,mode,lock,op)
	char *file;
	int flags;
	int mode;
	char *lock;
	int op;
{
  int fd;
  int ld = -1;
  int i = LOCKTIMEOUT * 60 - 1;
  char tmp[MAILTMPLEN];
  struct timeval tp;
  struct timezone tzp;
  struct stat sb;
				/* build lock filename */
  strcat (bezerk_file (lock,file),".lock");
  do {				/* until OK or out of tries */
    gettimeofday (&tp,&tzp);	/* get the time now */
				/* try to get the lock */
    if ((ld = open (lock,O_WRONLY|O_CREAT|O_EXCL,0666)) < 0) switch (errno) {
    case EEXIST:		/* lock file already exists, wait if young */
      if (((!stat (lock,&sb)) && tp.tv_sec >= sb.st_ctime + LOCKTIMEOUT * 60)||
	  (ld = open (lock,O_WRONLY|O_CREAT,0666)) < 0) sleep (1);
      break;
    case EACCES:		/* this can happen; assume flock() is OK */
      if ((!stat (lock,&sb)) && tp.tv_sec >= sb.st_ctime + LOCKTIMEOUT * 60)
	sleep (1);		/* wait if young file */
      else *lock = '\0';	/* probably no world write /usr/spool/mail */
      break;
    default:			/* some other failure */
      sprintf (tmp,"Error creating %s: %s",lock,strerror (errno));
      mm_log (tmp,WARN);	/* this is probably not good */
      *lock = '\0';		/* no lock file */
      break;
    }
    if (i%15 && ld < 0 && *lock) mm_log ("File is locked, waiting...",NIL);
  } while (i-- && ld < 0 && *lock);
  if (ld >= 0) {		/* if made a lock file */
    fchmod (ld,0666);		/* make sure others can break the lock */
    close (ld);			/* close the lock file */
  }
				/* open file */
  if (fd = open (file,flags,mode)) {
    flock (fd,op);		/* lock out flock'ers */
    return fd;
  }
  if (*lock) unlink (lock);	/* open failed, flush the lock file if any */
  return NIL;
}

/* Berkeley unlock and close mailbox
 * Accepts: file descriptor
 *	    buffer for lock file name
 */

void bezerk_unlock (fd,file)
	int fd;
	char *file;
{
  flock (fd,LOCK_UN);		/* release flock'ers */
  close (fd);			/* close the file */
  if (*file) unlink (file);	/* flush the lock file if any */
}

/* Berkeley mail parse and lock mailbox
 * Accepts: MAIL stream
 *	    space to write lock file name
 * Returns: file descriptor if parse OK, mailbox is locked shared
 *	    -1 if failure, stream aborted
 */

#define PAD(s) LOCAL->textsize = (s | 0x7fff) + 0x8001

int bezerk_parse (stream,lock)
	MAILSTREAM *stream;
	char *lock;
{
  MESSAGECACHE *elt = NIL;
  FILECACHE *m = NIL;
  struct stat sbuf;
  int fd;
  int i,rn,sysv,delta;
  char *s,*t,*k1;
  char c = '\n';
  char *k = NIL;
  int n = 0;
  int nmsgs = stream->nmsgs;
  int recent = stream->recent;
  mail_lock (stream);		/* guard against recursion or pingers */
				/* open and lock mailbox (shared OK) */
  if ((fd = bezerk_lock (LOCAL->fname,LOCAL->ld ? O_RDWR : O_RDONLY,NIL,
			  lock,LOCK_SH)) < 0) {
    mm_log ("Mailbox open failed, aborted",ERROR);
    bezerk_abort (stream);
    mail_unlock (stream);
    return -1;
  }
  fstat (fd,&sbuf);		/* get status */
				/* calculate change in size */
  if ((delta = sbuf.st_size - LOCAL->filesize) < 0) {
    sprintf (LOCAL->buf,"Mailbox shrank from %d to %d bytes, aborted",
	     LOCAL->filesize,sbuf.st_size);
    mm_log (LOCAL->buf,ERROR);	/* this is pretty bad */
    bezerk_unlock (fd,lock);
    bezerk_abort (stream);
    mail_unlock (stream);
    return -1;
  }

  else if (delta == 0) {	/* size unchanged? */
				/* don't babble if never got a time */
    if (LOCAL->filetime && LOCAL->filetime != sbuf.st_mtime)
      mm_log ("New mailbox mtime but apparently no changes, ignored",WARN);
  }
  else {			/* create buffer if first time through */
    if (!LOCAL->text) LOCAL->text = (char *) fs_get (PAD (sbuf.st_size));
				/* else resize if need more space */
    else if (LOCAL->textsize < (i = delta + LOCAL->textend))
      fs_resize ((void **) &LOCAL->text,PAD (i));
				/* get to that position in the file */
    lseek (fd,LOCAL->filesize,L_SET);
				/* read the new text */
    if (read (fd,(s = LOCAL->text+LOCAL->textend),delta) < 0) {
      sprintf (LOCAL->buf,"I/O error reading mail file: %s",strerror (errno));
      mm_log (LOCAL->buf,ERROR);
      bezerk_unlock (fd,lock);
      bezerk_abort (stream);
      mail_unlock (stream);
      return -1;
    }
    if ((s[0] =='F') && VALID)	/* calculate and tie off end of text */
      *(LOCAL->text + (LOCAL->textend += delta)) = '\0';
    else {			/* bogon */
      mm_log ("Mailbox format invalidated (consult an expert), aborted",ERROR);
      bezerk_unlock (fd,lock);
      bezerk_abort (stream);
      mail_unlock (stream);
      return -1;
    }

    while (s < (LOCAL->text + LOCAL->textend)) {
				/* examine each line */
      if (c == '\n') switch (*(t = s)) {
      case 'F' :		/* possible From header */
	if (VALID) {		/* valid From header? */
	  if (elt) {		/* if a previous message */
	    i = (s[-1] == '\n');/* non-zero if flush extra newline or CRLF */
				/* compute body size and RFC822 size */
	    if (m->body) m->bodysize = ((s - LOCAL->text) - m->body) - i;
	    elt->rfc822_size = m->headersize + m->bodysize + n - i;
	  }
	  if (k) {		/* must nuke kill region if one here */
	    memmove (k,k1,LOCAL->text + LOCAL->textend - k1);
	    s -= k1 - k;	/* adjust for kill region */
	    t -= k1 - k;
	    LOCAL->textend -= k1 - k;
	    k = k1 = NIL;	/* no more kill region */
	  }
				/* create new cache entries for this message */
	  m = LOCAL->msgs[nmsgs] = (FILECACHE *) fs_get (sizeof (FILECACHE));
				/* pointer to internal header */
	  m->internal = s - LOCAL->text;
	  s = t;		/* now at end of this line */
				/* pointer to RFC 822 header */
	  m->header = ++t - LOCAL->text;
				/* nothing else yet */
	  n = m->headersize = m->body = m->bodysize = 0;
	  recent++;		/* assume recent by default */
	  (elt = mail_elt (stream,++nmsgs))->recent = T;
				/* calculate initial Status/X-Status lines */
	  bezerk_update_status (m->status,elt);
	  t += rn + sysv - 21;	/* calculate start of date */
				/* generate plausable IMAPish date string */
	  sprintf (LOCAL->buf,"%2.2s-%3.3s-%2.2s %5.5s:%2.2s-%3.3s",
		   t+4,t,s-2,t+7,(sysv ? "00" : t+13),
		   (rn ? t+16-sysv : "LCL"));
				/* set internal date */
	  elt->internal_date = cpystr (LOCAL->buf);
				/* as well as parsed value */
	  m->date = bezerk_date (LOCAL->buf);
	}
	else n++;		/* count another newline */
	break;

      case 'X':			/* possible X-Status: line */
				/* if "X-" bump past it, drop into 'S' case */
	if (t[1] == '-' && t[2] == 'S') t += 2;
	else {
	  n++;			/* count another newline */
	  break;
	}
      case 'S':			/* see if Status: line */
	if (m && !m->headersize && t[1] == 't' && t[2] == 'a' && t[3] == 't' &&
	    t[4] == 'u' && t[5] == 's' && t[6] == ':' && strchr (t,'\n')) {
	  if (!k) k = s;	/* set start of kill region if none yet */
	  else if (k1 != s) {	/* still in our kill region? */
				/* no, blat up the rest of the mailbox */
	    memmove (k,k1,LOCAL->text + LOCAL->textend - k1);
	    LOCAL->textend -= k1 - k;
	    k = t -= k1 - k;	/* adjust for kill region, create new one */
	  }
	  s = t + 6;		/* advance to status flags and parse */
	  while (*++s != '\n') switch (*s) {
	  case 'R':		/* message read */
	    elt->seen = T;
	    break;
	  case 'O':		/* message old */
	    if (elt->recent) {	/* don't do this more than once! */
	      elt->recent = NIL;
	      recent--;		/* definitely not recent now! */
	    }
	    break;
	  case 'D':		/* message deleted */
	    elt->deleted = T;
	    break;
	  case 'F':		/* message flagged */
	    elt->flagged = T;
	    break;
	  case 'A':		/* message answered */
	    elt->answered = T;
	    break;
	  default:		/* some other crap */
	    break;
	  }
				/* recalculate Status/X-Status lines */
	  bezerk_update_status (m->status,elt);
	  k1 = s + 1;		/* end of kill region starts at next char */
	}
	else n++;		/* count another newline */
	break;

      case '\n':		/* double newline, looking for header? */
	if (m && !m->headersize) {
	  if (k && (k1 != s)) {	/* must nuke embedded kill region */
	    memmove (k,k1,LOCAL->text + LOCAL->textend - k1);
	    s -= k1 - k;	/* adjust for kill region */
	    LOCAL->textend -= k1 - k;
	    k = k1 = NIL;	/* no more kill region */
	  }
				/* compute body position */
	  m->body = (s + 1) - LOCAL->text;
	  if (k) *k++ = '\n';	/* patch in trailing newline */
				/* compute header size */
	  m->headersize = (k ? k - LOCAL->text : m->body) - m->header;
	  k = k1 = NIL;		/* no more kill region */
	}
	n++;			/* count another newline */
	break;
      default:			/* other characters are uninteresting */
	n++;			/* count another newline */
	break;
      }
      c = *s++;			/* remember last character and advance */
    }

    if (elt) {			/* compute size of final message body */
      i = (s[-1] == '\n');	/* non-zero if flush extra newline or CRLF */
				/* compute body size and RFC822 size */
      if (m->body) m->bodysize = ((s - LOCAL->text) - m->body) - i;
      elt->rfc822_size = m->headersize + m->bodysize + n - i;
    }
  }
				/* update parsed file size and time */
  LOCAL->filesize = sbuf.st_size;
  LOCAL->filetime = sbuf.st_mtime;
  if (LOCAL->lsjumb) {		/* All Right Now ??? */
    mail_exists (stream,nmsgs);	/* notify upper level of new mailbox size */
    mail_recent (stream,recent);/* and of change in recent messages */
  }
  return fd;			/* return the winnage */
}

/* Berkeley save mailbox
 * Accepts: MAIL stream
 *	    mailbox file descriptor
 *
 * Mailbox must be readwrite and locked for exclusive access.
 */

void bezerk_save (stream,fd)
	MAILSTREAM *stream;
	int fd;
{
  struct stat sbuf;
  struct iovec iov[4*MAXMESSAGES];
  int iovc = 0;
  int i = 0;
  lseek (fd,0,L_SET);		/* start at beginning of file */
  while (i < stream->nmsgs)	/* write each message */
    bezerk_write_message (iov,&iovc,LOCAL->text,LOCAL->msgs[i++]);
				/* now do the write */
  if ((LOCAL->filesize = bezerk_writev (fd,iov,iovc)) < 0) {
    sprintf (LOCAL->buf,"Unable to rewrite mailbox: %s",strerror (errno));
    fatal (LOCAL->buf);
  }
  fsync (fd);			/* make sure the disk has the update */
				/* nuke any cruft after that */
  ftruncate (fd,LOCAL->filesize);
  fsync (fd);			/* is this necessary? */
  fstat (fd,&sbuf);		/* now get updated file time */
  LOCAL->filetime = sbuf.st_mtime;
}


/* Berkeley fixed version of writev (requests)
	handles 0 and > 16 requests;
 * Accepts: file descriptor
 *	    iovec array
 *	    iovec count
 * Returns: number of bytes written or -1 if failure
 */

int bezerk_writev (fd,iov,iovc)
	int fd;
	struct iovec *iov;
	int iovc;
{
  int i,s;
  int size = 0;
  while (iovc > 0) {		/* while there are requests to do */
				/* write 16 at a time, return if error */
    if ((s = writev (fd,iov,(i = min (iovc,16)))) < 0) return s;
    size += s;			/* count up size of last batch */
    iov += i;			/* move to next batch */
    iovc -= i;			/* count down this batch */
  }
  return size;			/* return total number of bytes written */
}

/* Berkeley copy messages
 * Accepts: MAIL stream
 *	    mailbox copy vector
 *	    mailbox name
 */

void bezerk_copy_messages (stream,s,mailbox)
	MAILSTREAM *stream;
	unsigned char s[];
	char *mailbox;
{
  char file[MAILTMPLEN];
  char lock[MAILTMPLEN];
  int fd;
  struct iovec iv[4*MAXMESSAGES];
  int ivc = 0;
  int i = 0;
				/* open file for create/append and lock */
  if ((fd = bezerk_lock (bezerk_file (file,mailbox),O_WRONLY|O_APPEND|O_CREAT,
		  S_IREAD|S_IWRITE,lock,LOCK_EX)) >= 0) {
    while (i < stream->nmsgs)	/* write all requested messages to mailbox */
      if (s[i++]) bezerk_write_message (iv,&ivc,LOCAL->text,LOCAL->msgs[i-1]);
				/* now do the write */
    if (bezerk_writev (fd,iv,ivc) < 0) {
      sprintf (LOCAL->buf,"Unable to write copy: %s",strerror (errno));
      mm_log (LOCAL->buf,ERROR);
    }
    fsync (fd);			/* force out the update */
    bezerk_unlock (fd,lock);	/* unlock and close the file */
  }
  else {
    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
    mm_log (LOCAL->buf,ERROR);
  }
}

/* Berkeley write message to mailbox
 * Accepts: MAIL stream
 *	    local cache for this message
 *
 * This routine is the reason why the local cache has a copy of the status.
 * We can be called to dump out the mailbox as part of a stream recycle, since
 * we don't write out the mailbox when flags change and hence an update may be
 * needed.  However, at this point the elt has already become history, so we
 * can't use any information other than what is local to us.
 */

void bezerk_write_message (iov,i,s,m)
	struct iovec iov[];
	int *i;
	char *s;
	FILECACHE *m;
{
				/* pointer/counter to headers */
  iov[*i].iov_base = s + m->internal;
				/* length of internal + message headers */
  iov[*i].iov_len = (m->header + m->headersize) - m->internal;
				/* suppress extra newline if present */
  if ((iov[*i].iov_base)[iov[*i].iov_len - 2] == '\n') iov[(*i)++].iov_len--;
  else (*i)++;			/* unlikely but... */
  iov[*i].iov_base = m->status;	/* pointer/counter to status */
  iov[(*i)++].iov_len = strlen (m->status);
				/* pointer/counter to text body */
  iov[*i].iov_base = s + m->body;
  iov[(*i)++].iov_len = m->bodysize;
  iov[*i].iov_base = "\n";	/* pointer/counter to extra newline */
  iov[(*i)++].iov_len = 1;
}

/* Berkeley update status string
 * Accepts: destination string to write
 *	    message cache entry
 */

void bezerk_update_status (status,elt)
	char *status;
	MESSAGECACHE *elt;
{
  sprintf (status,STATUS,elt->seen ? "R" : "",elt->deleted ? "D" : "",
	   elt->flagged ? "F" : "",elt->answered ? "A" : "");
}

/* Berkeley parse sequence
 * Accepts: MAIL stream
 *	    sequence vector
 *	    sequence to parse
 * Returns: T if parse successful, else NIL
 */

short bezerk_getsequence (stream,s,sequence)
	MAILSTREAM *stream;
	unsigned char s[];
	char *sequence;
{
  int i,j,x;
  memset (s,NIL,MAXMESSAGES);	/* initialize sequence */
  while (*sequence) {		/* while there is something to parse */
				/* parse and validate message number */
    if (((i = (int) strtol ((const char *) sequence,&sequence,10)) < 1) ||
	(i > stream->nmsgs)) {
      mm_log ("Sequence invalid",ERROR);
      return NIL;
    }
    switch (*sequence) {	/* see what the delimiter is */
    case ':':			/* sequence range */
				/* parse end of range */
      if (((j = (int) strtol ((const char *) ++sequence,&sequence,10)) < 1) ||
	  (j > stream->nmsgs) || (*sequence && *sequence++ != ',')) {
	mm_log ("Sequence range invalid",ERROR);
	return NIL;
      }
      if (i > j) {		/* swap the range if backwards */
	x = i; i = j; j = x;
      }
      while (i <= j) s[--j] = T;/* mark each item in the sequence */
      break;
    case ',':			/* single message */
      ++sequence;		/* skip the delimiter, fall into end case */
    case '\0':			/* end of sequence */
      s[i-1] = T;		/* mark this message */
      break;
    default:			/* anything else is a syntax error! */
      mm_log ("Syntax error in sequence",ERROR);
      return NIL;
    }
  }
  return T;			/* successfully parsed sequence */
}

/* Parse flag list
 * Accepts: MAIL stream
 *	    flag list as a character string
 * Returns: flag command list
 */


short bezerk_getflags (stream,flag)
	MAILSTREAM *stream;
	char *flag;
{
  char *t;
  short f = 0;
  short i,j;
  if (flag && *flag) {		/* no-op if no flag string */
				/* check if a list and make sure valid */
    if ((i = (*flag == '(')) ^ (flag[strlen (flag)-1] == ')')) {
      mm_log ("Bad flag list",ERROR);
      return NIL;
    }
				/* copy the flag string w/o list construct */
    strncpy (LOCAL->buf,flag+i,(j = strlen (flag) - (2*i)));
    LOCAL->buf[j] = '\0';
    t = ucase (LOCAL->buf);	/* uppercase only from now on */

    while (*t) {		/* parse the flags */
      if (*t == '\\') {		/* system flag? */
	switch (*++t) {		/* dispatch based on first character */
	case 'S':		/* possible \Seen flag */
	  if (t[1] == 'E' && t[2] == 'E' && t[3] == 'N') i = fSEEN;
	  t += 4;		/* skip past flag name */
	  break;
	case 'D':		/* possible \Deleted flag */
	  if (t[1] == 'E' && t[2] == 'L' && t[3] == 'E' && t[4] == 'T' &&
	      t[5] == 'E' && t[6] == 'D') i = fDELETED;
	  t += 7;		/* skip past flag name */
	  break;
	case 'F':		/* possible \Flagged flag */
	  if (t[1] == 'L' && t[2] == 'A' && t[3] == 'G' && t[4] == 'G' &&
	      t[5] == 'E' && t[6] == 'D') i = fFLAGGED;
	  t += 7;		/* skip past flag name */
	  break;
	case 'A':		/* possible \Answered flag */
	  if (t[1] == 'N' && t[2] == 'S' && t[3] == 'W' && t[4] == 'E' &&
	      t[5] == 'R' && t[6] == 'E' && t[7] == 'D') i = fANSWERED;
	  t += 8;		/* skip past flag name */
	  break;
	default:		/* unknown */
	  i = 0;
	  break;
	}
				/* add flag to flags list */
	if (i && ((*t == '\0') || (*t++ == ' '))) f |= i;
	else {			/* bitch about bogus flag */
	  mm_log ("Unknown system flag",ERROR);
	  return NIL;
	}
      }
      else {			/* no user flags yet */
	mm_log ("Unknown flag",ERROR);
	return NIL;
      }
    }
  }
  return f;
}

/* Search support routines
 * Accepts: MAIL stream
 *	    message number
 *	    pointer to additional data
 * Returns: T if search matches, else NIL
 */


char bezerk_search_all (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return T;			/* ALL always succeeds */
}


char bezerk_search_answered (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->answered ? T : NIL;
}


char bezerk_search_deleted (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->deleted ? T : NIL;
}


char bezerk_search_flagged (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->flagged ? T : NIL;
}


char bezerk_search_keyword (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return NIL;			/* keywords not supported yet */
}


char bezerk_search_new (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (elt->recent && !elt->seen) ? T : NIL;
}

char bezerk_search_old (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->recent ? NIL : T;
}


char bezerk_search_recent (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->recent ? T : NIL;
}


char bezerk_search_seen (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->seen ? T : NIL;
}


char bezerk_search_unanswered (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->answered ? NIL : T;
}


char bezerk_search_undeleted (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->deleted ? NIL : T;
}


char bezerk_search_unflagged (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->flagged ? NIL : T;
}


char bezerk_search_unkeyword (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return T;			/* keywords not supported yet */
}


char bezerk_search_unseen (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return mail_elt (stream,msgno)->seen ? NIL : T;
}

char bezerk_search_before (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return (char) (LOCAL->msgs[msgno-1]->date < n);
}


char bezerk_search_on (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  return (char) (LOCAL->msgs[msgno-1]->date == n);
}


char bezerk_search_since (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
				/* everybody interprets "since" as .GE. */
  return (char) (LOCAL->msgs[msgno-1]->date >= n);
}


char bezerk_search_body (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  FILECACHE *m = LOCAL->msgs[msgno-1];
  return search (LOCAL->text + m->body,m->bodysize,d,n);
}


char bezerk_search_subject (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  char *s = bezerk_fetchenvelope (stream,msgno)->subject;
  return s ? search (s,strlen (s),d,n) : NIL;
}


char bezerk_search_text (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  FILECACHE *m = LOCAL->msgs[msgno-1];
  return search (LOCAL->text + m->header,m->headersize,d,n) ||
    bezerk_search_body (stream,msgno,d,n);
}

char bezerk_search_bcc (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,bezerk_fetchenvelope (stream,msgno)->bcc);
  return search (LOCAL->buf,strlen (LOCAL->buf),d,n);
}


char bezerk_search_cc (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,bezerk_fetchenvelope (stream,msgno)->cc);
  return search (LOCAL->buf,strlen (LOCAL->buf),d,n);
}


char bezerk_search_from (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,bezerk_fetchenvelope (stream,msgno)->from);
  return search (LOCAL->buf,strlen (LOCAL->buf),d,n);
}


char bezerk_search_to (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	int n;
{
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,bezerk_fetchenvelope (stream,msgno)->to);
  return search (LOCAL->buf,strlen (LOCAL->buf),d,n);
}

/* Search parsers */


/* Parse a date
 * Accepts: function to return
 *	    pointer to date integer to return
 * Returns: function to return
 */

char (*bezerk_search_date (f,n)) ()
      char (*f) ();
      int *n;
{
  int i;
  char *s;
				/* parse the date and return fn if OK */
  return (bezerk_search_string (f,&s,&i) && (*n = bezerk_date (s))) ? f : NIL;
}


/* Actual date parser (routine)
	minimal and stupid routine;
 * Accepts: date to parse
 * Returns: date integer
 */

int bezerk_date (s)
	char *s;
{
  long ms;
  unsigned int d,m,y;
				/* parse first number (probable month) */
  if (!(s && (m = (int) strtol ((const char *) s,&s,10)))) return NIL;
  switch (*s) {			/* different parse based on delimiter */
  case '/':			/* mm/dd/yy format */
				/* parse remainder of date */
    if (!((d = (int) strtol ((const char *) ++s,&s,10)) && *s == '/' &&
	  (y = (int) strtol ((const char *) ++s,&s,10)) && *s == '\0'))
      return NIL;
    break;

  case '-':			/* dd-mmm-yy format */
    d = m;			/* so the number we got is a day */
				/* make sure string is UC and long enough! */
    if (strlen (ucase (s)) < 5) return NIL;
				/* slurp up the month string */
    ms = (s[1] << 16) + (s[2] << 8) + s[3];
    switch (ms) {		/* determine the month */
    case ('J' << 16) + ('A' << 8) + 'N':
      m = 1; break;
    case ('F' << 16) + ('E' << 8) + 'B':
      m = 2; break;
    case ('M' << 16) + ('A' << 8) + 'R':
      m = 3; break;
    case ('A' << 16) + ('P' << 8) + 'R':
      m = 4; break;
    case ('M' << 16) + ('A' << 8) + 'Y':
      m = 5; break;
    case ('J' << 16) + ('U' << 8) + 'N':
      m = 6; break;
    case ('J' << 16) + ('U' << 8) + 'L':
      m = 7; break;
    case ('A' << 16) + ('U' << 8) + 'G':
      m = 8; break;
    case ('S' << 16) + ('E' << 8) + 'P':
      m = 9; break;
    case ('O' << 16) + ('C' << 8) + 'T':
      m = 10; break;
    case ('N' << 16) + ('O' << 8) + 'V':
      m = 11; break;
    case ('D' << 16) + ('E' << 8) + 'C':
      m = 12; break;
    default:
      return NIL;
    }
				/* parse the year */
    if (s[4] == '-' && (y = (int) strtol ((const char *) s+5,&s,10)) &&
	(*s == '\0' || *s == ' ')) break;
  default:			/* unknown format */
    return NIL;
  }
  y -= (y >= 1900) ? 1970 : 70;	/* the world began for Unix in 1970 */
				/* minimal validity check of date */
  if (d < 1 || d > 31 || m < 1 || m > 12 || y < 0 || y >= 100) return NIL;
  return (y*12+m-1)*31+d-1;	/* calculate and return date value */
}

/* Parse a flag
 * Accepts: function to return
 *	    pointer to string to return
 * Returns: function to return
 */

char (*bezerk_search_flag (f,d)) ()
      char (*f) ();
      char **d;
{
				/* get a keyword, return if OK */
  return (*d = strtok (NIL," ")) ? f : NIL;
}


/* Parse a string
 * Accepts: function to return
 *	    pointer to string to return
 *	    pointer to string length to return
 * Returns: function to return
 */


char (*bezerk_search_string (f,d,n)) ()
      char (*f) ();
      char **d;
      int *n;
{
  char *c = strtok (NIL,"");	/* remainder of criteria */
  if (c) {			/* better be an argument */
    switch (*c) {		/* see what the argument is */
    case '\0':			/* catch bogons */
    case ' ':
      return NIL;
    case '"':			/* quoted string */
      if (!(strchr (c+1,'"') && (*d = strtok (c,"\"")) && (*n = strlen (*d))))
	return NIL;
      break;
    case '{':			/* literal string */
      *n = strtol (c+1,&c,10);	/* get its length */
      if (*c++ != '}' || *c++ != '\015' || *c++ != '\012' ||
	  *n > strlen (*d = c)) return NIL;
      c[*n] = '\255';		/* write new delimiter */
      strtok (c,"\255");	/* reset the strtok mechanism */
      break;
    default:			/* atomic string */
      *n = strlen (*d = strtok (c," "));
      break;
    }
    return f;
  }
  else return NIL;
}
