/*
 * Program:	Interactive Mail Access Protocol 2 (IMAP2) routines
 *
 * Author(s):	Mark Crispin
 *		Symbolic Systems Resources Group
 *		Knowledge Systems Laboratory
 *		Departments of Computer Science and Medicine
 *		Stanford University
 *		Stanford, CA  94305
 *
 * Date:	15 June 1988
 *
 * Sponsorship:	This work was supported by the Biomedical Research
 *		Technology Program of the National Institutes of Health
 *		under grant number RR-00785.
 *
 * Copyright 1988, 1989, 1990, 1991 by The Leland Stanford Junior University.
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license from the Stanford Office of
 * Technology Licensing.
 *
 */


#include <ctype.h>
#ifdef __STDC__
#include <string.h>
#else
#include "ansi.h"
#endif
#include <sys/types.h>
#include "mail.h"
#include "imap2.h"
#include "misc.h"
#include "osdep.h"
#include "tcpsio.h"

/* Mail Access Protocol routines
 *
 *  Note that there is an important difference between a "sequence" and a
 * "message #" (msgno).  A sequence is a string representing a sequence in
 * IMAP2 format ("n", "n:m", or combination separated by commas), whereas
 * a msgno is a single integer.
 *
 */


/* Driver dispatch used by MAIL */

DRIVER imapdriver = {
  (DRIVER *) NIL,		/* next driver */
  map_valid,			/* mailbox is valid for us */
  map_open,			/* open mailbox */
  map_close,			/* close mailbox */
  map_fetchfast,		/* fetch message "fast" attributes */
  map_fetchflags,		/* fetch message flags */
  map_fetchenvelope,		/* fetch message envelopes */
  map_fetchheader,		/* fetch message header only */
  map_fetchtext,		/* fetch message body only */
  map_fetchbody,		/* fetch message body section */
  map_setflag,			/* set message flag */
  map_clearflag,		/* clear message flag */
  map_search,			/* search for message based on criteria */
  map_ping,			/* ping mailbox to see if still alive */
  map_check,			/* check for new messages */
  map_expunge,			/* expunge deleted messages */
  map_copy,			/* copy messages to another mailbox */
  map_move			/* move messages to another mailbox */
};


/* Mail Access Protocol validate mailbox
 * Accepts: mailbox name
 * Returns: our driver if name is valid, otherwise calls valid in next driver
 */

DRIVER *map_valid (name)
	char *name;
{				/* assume IMAP if starts with host delimiter */
  if (*name == '{') return &imapdriver;
				/* else let next driver have a try */
  else return imapdriver.next ? (*imapdriver.next->valid) (name) : NIL;
}

/* Mail Access Protocol open
 * Accepts: stream to open
 *	    mailbox name in "{host}mailbox" form
 *	    initial debugging flag
 * Returns: stream to use on success, NIL on failure
 */

MAILSTREAM *map_open (stream)
	MAILSTREAM *stream;
{
  int i;
  char username[MAILTMPLEN];
  char pwd[MAILTMPLEN];
  char tmp[MAILTMPLEN];
  char *host;
  char *mailbox;
  void *tcpstream;
  IMAPPARSEDREPLY *reply;
  if (!(host = imap_hostfield (stream->mailbox)))
    mm_log ("Host not specified",ERROR);
  else {
    if (!(mailbox = imap_mailboxfield (stream->mailbox))) {
      fs_give ((void **) &host);/* prevent storage leaks */
      mm_log ("Mailbox not specified",ERROR);
    }
    else {
      if (LOCAL) {		/* if stream opened earlier by us */
	if (strcmp (ucase (strcpy (tmp,host)),
		    ucase (strcpy (pwd,imap_host (stream))))) {
				/* if hosts are different punt it */
	  sprintf (tmp,"Closing connection to %s",imap_host (stream));
	  mm_log (tmp,NIL);
	  map_close (stream);
	}			/* else recycle if still alive */
	else if (map_ping (stream)) {
	  sprintf (tmp,"Reusing connection to %s",host);
	  mm_log (tmp,NIL);
	}
      }

      if (!LOCAL) {		/* open new connection if no recycle */
	if (tcpstream = tcp_open (host,IMAPTCPPORT)) {
	  stream->local = fs_get (sizeof (IMAPLOCAL));
				/* bind our stream */
	  LOCAL->tcpstream = tcpstream;
	  LOCAL->reply.line = LOCAL->reply.tag = LOCAL->reply.key =
	    LOCAL->reply.text = NIL;
				/* assume maximal server */
	  LOCAL->use_encoding = LOCAL->use_find = LOCAL->use_bboard = T;
				/* no cached texts */
	  for (i = 0; i < MAXMESSAGES; ++i)
	    LOCAL->header[i] = LOCAL->text[i] = NIL;
				/* get greeting */
	  if (!imap_OK (reply = imap_reply (stream,NIL))) {
	    mm_log (reply->text,ERROR);
	    map_close (stream);	/* clean up */
	  }
	  else {		/* only so many tries to login */
	    for (i = 0; i < MAXLOGINTRIALS; ++i) {
	      *pwd = 0;		/* initialize the password */
				/* get username and password from user */
	      mm_login (tcp_host (tcpstream),username,pwd,i);
				/* abort if he refuses to give a password */
	      if (*pwd == '\0') i = MAXLOGINTRIALS;
	      else {		/* send "LOGIN username pwd" */
		sprintf (tmp,"%s {%d}\015\012%s",username,strlen (pwd),pwd);
		if (imap_OK (reply = imap_send (stream,"LOGIN",tmp))) break;
				/* output failure and try again */
		sprintf (tmp,"Login failed: %s",reply->text);
		mm_log (tmp,WARN);
				/* give up now if connection died */
		if (!strcmp (reply->key,"BYE")) i = MAXLOGINTRIALS;
	      }
	    }
				/* give up if too many failures */
	    if (i >=  MAXLOGINTRIALS) {
	      mm_log (*pwd ? "Too many login failures" :
		      "Login aborted",ERROR);
	      map_close (stream);
	    }
	  }
	}
      }

      if (LOCAL) {		/* have a connection now??? */
	stream->sequence++;	/* bump sequence number */
				/* prepare to update mailbox name */
	fs_give ((void **) &stream->mailbox);
				/* send "SELECT mailbox" */
	if (!imap_OK (reply = imap_send (stream,"SELECT",mailbox))) {
	  sprintf (tmp,"{%s}<no mailbox>",imap_host (stream));
	  stream->mailbox = cpystr (tmp);
	  sprintf (tmp,"Can't select mailbox: %s",reply->text);
	  mm_log (tmp,ERROR);
				/* make sure dummy message counts */
	  mail_exists (stream,0);
	  mail_recent (stream,0);
	}
	else {			/* update mailbox name */
	  sprintf (tmp,"{%s}%s",imap_host (stream),mailbox);
	  stream->mailbox = cpystr (tmp);
	}
	if (!stream->nmsgs) mm_log ("Mailbox is empty",WARN);
      }
      else {			/* failed to open, prevent storage leaks */
	fs_give ((void **) &host);
	fs_give ((void **) &mailbox);
      }
    }
  }
  return LOCAL ? stream : NIL;	/* if stream is alive, return to caller */
}

/* Mail Access Protocol close
 * Accepts: MAIL stream
 */

void map_close (stream)
	MAILSTREAM *stream;
{
  int i;
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
  if (stream && LOCAL) {	/* send "LOGOUT" */
    if (!imap_OK (reply = imap_send (stream,"LOGOUT",NIL))) {
      sprintf (tmp,"Error in logout: %s",reply->text);
      mm_log (tmp,WARN);
    }
				/* close TCP connection */
    if (LOCAL->tcpstream) tcp_close (LOCAL->tcpstream);
				/* free up memory */
    if (LOCAL->reply.line) fs_give ((void **) &LOCAL->reply.line);
    for (i = 0; i < MAXMESSAGES; ++i) {
      if (LOCAL->header[i]) fs_give ((void **) &LOCAL->header[i]);
      if (LOCAL->text[i]) fs_give ((void **) &LOCAL->text[i]);
    }
				/* nuke the local data */
    fs_give ((void **) &stream->local);
  }
}


/* Mail Access Protocol fetch fast information
 * Accepts: MAIL stream
 *	    sequence
 *
 * Generally, map_fetchenvelope is preferred
 */

void map_fetchfast (stream,sequence)
	MAILSTREAM *stream;
	char *sequence;
{				/* send "FETCH sequence FAST" */
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
  sprintf (tmp,"%s FAST",sequence);
  if (!imap_OK (reply = imap_send (stream,"FETCH",tmp))) {
    sprintf (tmp,"Fetch fast rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
}

/* Mail Access Protocol fetch flags
 * Accepts: MAIL stream
 *	    sequence
 */

void map_fetchflags (stream,sequence)
	MAILSTREAM *stream;
	char *sequence;
{				/* send "FETCH sequence FLAGS" */
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
  sprintf (tmp,"%s FLAGS",sequence);
  if (!imap_OK (reply = imap_send (stream,"FETCH",tmp))) {
    sprintf (tmp,"Fetch flags rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
}


/* Mail Access Protocol fetch envelope
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: envelope of this message
 *
 * Fetches the "fast" information as well
 */

ENVELOPE *map_fetchenvelope (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  int i = msgno;
  int j = min (msgno+MAPLOOKAHEAD-1,stream->nmsgs);
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  if (!elt->env) {		/* if we don't have an envelope */
    if (msgno != stream->nmsgs)	/* never lookahead if last message */
      while (i < j && !(mail_elt (stream,i+1)->env)) i++;
				/* build "FETCH msgno:i (ALL ENCODING)" */
    sprintf (tmp,"%d:%d (ALL ENCODING)",msgno,i);
    if (!(LOCAL->use_encoding &&/* use encoding if we can */
	  strcmp ((reply = imap_send (stream,"FETCH",tmp))->key,"BAD"))) {
      LOCAL->use_encoding = NIL;/* can't, send "FETCH msgno:i ALL" instead */
      sprintf (tmp,"%d:%d ALL",msgno,i);
      reply = imap_send (stream,"FETCH",tmp);
    }
    if (!imap_OK (reply)) {	/* did it win? */
      sprintf (tmp,"Fetch envelope rejected: %s",reply->text);
      mm_log (tmp,ERROR);
      return NIL;
    }
  }
  return elt->env;		/* return the envelope */
}

/* Mail Access Protocol fetch message header
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message header in RFC822 format
 */

char *map_fetchheader (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  char tmp[MAILTMPLEN];
  int i = msgno - 1;
  IMAPPARSEDREPLY *reply;
  if (!LOCAL->header[i]) {	/* send "FETCH msgno RFC822.HEADER" */
    sprintf (tmp,LOCAL->text[i] ? "%d RFC822.HEADER" :
	     "%d (RFC822.HEADER RFC822.TEXT)",msgno);
    if (!imap_OK (reply = imap_send (stream,"FETCH",tmp))) {
      sprintf (tmp,"Fetch header rejected: %s",reply->text);
      mm_log (tmp,ERROR);
    }
  }
  return LOCAL->header[i] ? LOCAL->header[i] : "";
}


/* Mail Access Protocol fetch message text (only)
	body only;
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message text in RFC822 format
 */

char *map_fetchtext (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  char tmp[MAILTMPLEN];
  int i = msgno - 1;
  IMAPPARSEDREPLY *reply;
  if (!LOCAL->text[i]) {	/* send "FETCH msgno RFC822.TEXT" */
    sprintf (tmp,"%d RFC822.TEXT",msgno);
    if (!imap_OK (reply = imap_send (stream,"FETCH",tmp))) {
      sprintf (tmp,"Fetch text rejected: %s",reply->text);
      mm_log (tmp,ERROR);
    }
  }
  return LOCAL->text[i] ? LOCAL->text[i] : "";
}

/* Mail Access Protocol fetch message body as a structure
 * Accepts: Mail stream
 *	    message # to fetch
 *	    section #
 * Returns: pointer to section of message body
 */

char *map_fetchbody (stream,msgno,section)
	MAILSTREAM *stream;
	int msgno;
	int section;
{
  char tmp[MAILTMPLEN];
  int i = section;
  IMAPPARSEDREPLY *reply;
  BODY *body;
				/* if not structured & sec 0, return text */
  if (!(LOCAL->use_encoding && (map_fetchenvelope (stream,msgno) || T) &&
	(body = mail_elt (stream,msgno)->body)))
    return section ? "" : map_fetchtext (stream,msgno);
				/* see if the section exists */
  while (i-- && body) body = body->next;
  if (!body) return "";		/* section doesn't exist */
  if (!body->text) {		/* send "FETCH msgno BODY[section]" */
    sprintf (tmp,"%d BODY[%d]",msgno,section);
    if (!imap_OK (reply = imap_send (stream,"FETCH",tmp))) {
      sprintf (tmp,"Fetch body section rejected: %s",reply->text);
      mm_log (tmp,ERROR);
    }
				/* look up body again */
    for (i = section,body = mail_elt (stream,msgno)->body; i-- && body;
	 body = body->next);
    if (!body) return "";	/* section doesn't exist */
  }
				/* return the text */
  return (body->text) ? body->text : "";
}

/* Mail Access Protocol set flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void map_setflag (stream,sequence,flag)
	MAILSTREAM *stream;
	char *sequence;
	char *flag;
{
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
				/* "STORE sequence +Flags flag" */
  sprintf (tmp,"%s +Flags %s",sequence,flag);
  if (!imap_OK (reply = imap_send (stream,"STORE",tmp))) {
    sprintf (tmp,"Set flag rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
}


/* Mail Access Protocol clear flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void map_clearflag (stream,sequence,flag)
	MAILSTREAM *stream;
	char *sequence;
	char *flag;
{
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
				/* "STORE sequence -Flags flag" */
  sprintf (tmp,"%s -Flags %s",sequence,flag);
  if (!imap_OK (reply = imap_send (stream,"STORE",tmp))) {
    sprintf (tmp,"Clear flag rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
}

/* Mail Access Protocol search for messages
 * Accepts: MAIL stream
 *	    search criteria
 */

void map_search (stream,criteria)
	MAILSTREAM *stream;
	char *criteria;
{
  int i,j;
  char *s;
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
  MESSAGECACHE *elt;
				/* do implicit CHECK, then SEARCH */
  if (imap_OK (reply = imap_send (stream,"CHECK",NIL)) &&
      imap_OK (reply = imap_send (stream,"SEARCH",criteria))) {
    s = tmp;			/* build sequence in temporary buffer */
    *s = '\0';			/* initially nothing */
				/* search through mailbox */
    for (i = 1; i <= stream->nmsgs; ++i)
				/* for searched messages with no envelope */
      if ((elt = stream->messagearray[i-1]) && elt->searched && !elt->env) {
	if (tmp[0]) *s++ = ','; /* prepend with comma if not first time */
	sprintf (s,"%d",j = i);	/* output message number */
	s += strlen (s);	/* point at end of string */
				/* search for possible end of range */
	while (i < stream->nmsgs && (elt = stream->messagearray[i]) &&
	       elt->searched && !elt->env) i++;
	if (i != j) {		/* if a range */
	  sprintf (s,":%d",i);	/* output delimiter and end of range */
	  s += strlen (s);	/* point at end of string */
	}
      }
    if (tmp[0]) {		/* anything to pre-fetch? */
      sprintf (s," ALL");	/* envelope and fast poop */
				/* pre-fetch the searched messages */
      if (!imap_OK (reply = imap_send (stream,"FETCH",tmp))) {
	sprintf (tmp,"Search pre-fetch rejected: %s",reply->text);
	mm_log (tmp,ERROR);
      }
    }
  }
  else {
    sprintf (tmp,"Search rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
}

/* Mail Access Protocol ping mailbox
 * Accepts: MAIL stream
 */

MAILSTREAM *map_ping (stream)
	MAILSTREAM *stream;
{				/* send "NOOP" */
  return imap_OK (imap_send (stream,"NOOP",NIL)) ? stream : NIL;
}


/* Mail Access Protocol check mailbox
 * Accepts: MAIL stream
 */

void map_check (stream)
	MAILSTREAM *stream;
{
  char tmp[MAILTMPLEN];
				/* send "CHECK" */
  IMAPPARSEDREPLY *reply = imap_send (stream,"CHECK",NIL);
  if (!imap_OK (reply)) {
    sprintf (tmp,"Check rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
  else mm_log (reply->text,NIL);
}


/* Mail Access Protocol expunge mailbox
 * Accepts: MAIL stream
 */

void map_expunge (stream)
	MAILSTREAM *stream;
{
  char tmp[MAILTMPLEN];
				/* send "EXPUNGE" */
  IMAPPARSEDREPLY *reply = imap_send (stream,"EXPUNGE",NIL);
  if (!imap_OK (reply)) {
    sprintf (tmp,"Expunge rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
  else mm_log (reply->text,NIL);
}

/* Mail Access Protocol copy message(s)
	s;
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 */

void map_copy (stream,sequence,mailbox)
	MAILSTREAM *stream;
	char *sequence;
	char *mailbox;
{
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
				/* send "COPY sequence mailbox" */
  sprintf (tmp,"%s %s",sequence,mailbox);
  if (!imap_OK (reply = imap_send (stream,"COPY",tmp))) {
    sprintf (tmp,"Copy rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
  else map_setflag (stream,sequence,"\\Seen");
}


/* Mail Access Protocol move message(s)
	s;
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 */

void map_move (stream,sequence,mailbox)
	MAILSTREAM *stream;
	char *sequence;
	char *mailbox;
{
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
				/* send "COPY sequence mailbox" */
  sprintf (tmp,"%s %s",sequence,mailbox);
  if (!imap_OK (reply = imap_send (stream,"COPY",tmp))) {
    sprintf (tmp,"Copy rejected: %s",reply->text);
    mm_log (tmp,ERROR);
  }
  else map_setflag (stream,sequence,"\\Deleted \\Seen");
}

/* Internal routines */


/* Mail Access Protocol return host name field of "{host}mailbox"
 * Accepts: "{host}mailbox" format name
 * Returns: host name or NIL if error
 */

char *imap_hostfield (string)
	char *string;
{
  register char *start;
  char c;
  int i = 0;
				/* string must begin with "{" */
  if (*string != '{') return NIL;
  start = ++string;		/* save pointer to first host char */
				/* search for closing "}" */
  while ((c = *string++) != '}') if (!c) return NIL;
				/* must have at least one char */
  if (!(i = string - start - 1)) return NIL;
  string = (char *) fs_get (i + 1);
  strncpy (string,start,i);	/* copy the string */
  string[i] = '\0';		/* tie off with null */
  return string;
}


/* Mail Access Protocol return mailbox field of "{host}mailbox"
 * Accepts: "{host}mailbox" format name
 * Returns: mailbox name or NIL if error
 */

char *imap_mailboxfield (string)
	char *string;
{
  char c;
				/* string must begin with "{" */
  if (*string != '{') return NIL;
				/* search for closing "}" */
  while ((c = *string++) != '}') if (!c) return NIL;
  return cpystr (strlen (string) ? string : "INBOX");
}


/* Mail Access Protocol return host name
 * Accepts: MAIL stream
 * Returns: host name
 */

char *imap_host (stream)
	MAILSTREAM *stream;
{
				/* return host name on stream if open */
  return (LOCAL && LOCAL->tcpstream) ? tcp_host (LOCAL->tcpstream) :
    "<closed stream>";
}

/* Mail Access Protocol send command
 * Accepts: MAIL stream
 *	    command
 *	    command argument
 * Returns: parsed reply
 */

IMAPPARSEDREPLY *imap_send (stream,command,args)
	MAILSTREAM *stream;
	char *command;
	char *args;
{
  char tmp[MAILTMPLEN];
  char tag[7];
  char *text;
  IMAPPARSEDREPLY *reply;
  				/* gensym a new tag */
  sprintf (tag,"A%05d",stream->gensym++);
  if (!LOCAL->tcpstream) return imap_fake (stream,tag,"OK No-op dead stream");
  mail_lock (stream);		/* lock up the stream */
  if (args) sprintf (tmp,"%s %s %s",tag,command,args);
  else sprintf (tmp,"%s %s",tag,command);
  if (stream->debug) mm_dlog (tmp);
  strcat (tmp,"\015\012");
				/* send the command */
  if (!(tcp_soutr (LOCAL->tcpstream,tmp))) {
				/* close TCP connection */
    tcp_close (LOCAL->tcpstream);
    LOCAL->tcpstream = NIL;
    return imap_fake (stream,tag,"BYE IMAP connection broken in send");
  }
				/* get reply from server */
  reply = imap_reply (stream,tag);
  mail_unlock (stream);		/* unlock stream */
  return reply;			/* return reply to caller */
}

/* Mail Access Protocol get reply
 * Accepts: MAIL stream
 *	    tag to search or NIL if want a greeting
 * Returns: parsed reply, never NIL
 */

IMAPPARSEDREPLY *imap_reply (stream,tag)
	MAILSTREAM *stream;
	char *tag;
{
  char tmp[MAILTMPLEN];
  IMAPPARSEDREPLY *reply;
  while (LOCAL->tcpstream) {	/* parse reply from server */
    if ((reply = imap_parse_reply (stream,tcp_getline (LOCAL->tcpstream))) &&
	strcmp (reply->tag,"+")) {
				/* untagged response means unsolicited data */
      if (!strcmp (reply->tag,"*")) {
	imap_parse_unsolicited (stream,reply);
	if (!tag) return reply;	/* return greeting */
      }
      else {			/* tagged reponse, return if desired reply */
	if (tag && !strcmp (tag,reply->tag)) return reply;
				/* report bogon */
	sprintf (tmp,"Unexpected tagged response: %s %s %s",
		 reply->tag,reply->key,reply->text);
	mm_log (tmp,WARN);
      }
    }
  }
  return imap_fake (stream,tag,"BYE IMAP connection broken in reply");
}

/* Mail Access Protocol parse reply
 * Accepts: MAIL stream
 *	    text of reply
 * Returns: parsed reply, or NIL if can't parse at least a tag and key
 */


IMAPPARSEDREPLY *imap_parse_reply (stream,text)
	MAILSTREAM *stream;
	char *text;
{
  char tmp[MAILTMPLEN];
  if (LOCAL->reply.line) fs_give ((void **) &LOCAL->reply.line);
  if (!(LOCAL->reply.line = text)) {
				/* NIL text means the stream died */
    tcp_close (LOCAL->tcpstream);
    LOCAL->tcpstream = NIL;
    return NIL;
  }
  if (stream->debug) mm_dlog (LOCAL->reply.line);
  LOCAL->reply.key = NIL;	/* init fields in case error */
  LOCAL->reply.text = NIL;
				/* parse separate tag, key, text */
  if (!((LOCAL->reply.tag = (char *) strtok (LOCAL->reply.line," ")) &&
	(LOCAL->reply.key = (char *) strtok (NIL," ")))) {
				/* determine what is missing */
    if (!LOCAL->reply.tag) strcpy (tmp,"IMAP2 server sent a blank line");
    else sprintf (tmp,"Missing IMAP2 reply key: %s",LOCAL->reply.tag);
    mm_log (tmp,WARN);		/* pass up the barfage */
    return NIL;			/* can't parse this text */
  }
  ucase (LOCAL->reply.key);	/* make sure key is upper case */
				/* get text as well, allow empty text */
  if (!(LOCAL->reply.text = (char *) strtok (NIL,"\n")))
    LOCAL->reply.text = LOCAL->reply.key + strlen (LOCAL->reply.key);
  return &LOCAL->reply;		/* return parsed reply */
}

/* Mail Access Protocol fake reply
 * Accepts: MAIL stream
 *	    tag
 *	    text of fake reply
 * Returns: parsed reply
 */

IMAPPARSEDREPLY *imap_fake (stream,tag,text)
	MAILSTREAM *stream;
	char *tag;
	char *text;
{
  char tmp[MAILTMPLEN];
				/* build fake reply string */
  sprintf (tmp,"%s %s",tag,text);
				/* parse and return it */
  return imap_parse_reply (stream,cpystr (tmp));
}


/* Mail Access Protocol check for OK response in tagged reply
 * Accepts: parsed IMAP reply
 * Returns: T if OK else NIL
 */

int imap_OK (reply)
	IMAPPARSEDREPLY *reply;
{
  char tmp[MAILTMPLEN];
				/* OK - operation succeeded */
  if (!strcmp (reply->key,"OK")) return T;
				/* NO - operation failed */
  if (!strcmp (reply->key,"NO")) return NIL;
				/* BAD - operation rejected */
  if (!strcmp (reply->key,"BAD")) sprintf (tmp,"IMAP error: %s",reply->text);
				/* BYE - server is going away */
  else if (!strcmp (reply->key,"BYE")) strcpy (tmp,reply->text);
  else sprintf (tmp,"Unexpected IMAP response: %s %s",reply->key,reply->text);
  mm_log (tmp,WARN);		/* log the sucker */
  return NIL;
}

/* Unsolicited reply handling
 *
 *  This is perhaps the most important part of all of the IMAP2 code; the
 * code that actually handles received data from the mail server.
 *
 */


/* Mail Access Protocol parse and act upon unsolicited reply
 * Accepts: MAIL stream
 *	    parsed reply
 */

void imap_parse_unsolicited (stream,reply)
	MAILSTREAM *stream;
	IMAPPARSEDREPLY *reply;
{
  char tmp[MAILTMPLEN];
  long msgno;
  char *endptr;
  char *keyptr,*txtptr;
				/* see if key is a number */
  msgno = strtol (reply->key,&endptr,10);
  if (*endptr) {		/* if non-numeric */
    IFEQUAL (reply->key,"FLAGS",imap_parse_flaglst (stream,reply))
    IFEQUAL (reply->key,"SEARCH",imap_searched (stream,reply->text))
    IFEQUAL (reply->key,"BYE",mm_log (reply->text,NIL))
    IFEQUAL (reply->key,"OK",mm_notify (stream,reply->text,NIL))
    IFEQUAL (reply->key,"NO",mm_notify (stream,reply->text,WARN))
    IFEQUAL (reply->key,"BAD",mm_notify (stream,reply->text,ERROR))
    sprintf (tmp,"Unexpected unsolicited message: %s",reply->key);
    mm_log (tmp,WARN);
  }
  else {			/* if numeric, a keyword follows */
				/* deposit null at end of keyword */
    keyptr = ucase ((char *) strtok (reply->text," "));
				/* and locate the text after it */
    txtptr = (char *) strtok (NIL,"\n");
				/* now take the action */
    IFEQUAL (keyptr,"EXISTS",mail_exists (stream,msgno))
    IFEQUAL (keyptr,"RECENT",mail_recent (stream,msgno))
    IFEQUAL (keyptr,"EXPUNGE",imap_expunged (stream,msgno))
    IFEQUAL (keyptr,"FETCH",imap_parse_data (stream,msgno,txtptr,reply))
				/* obsolete alias for FETCH */
    IFEQUAL (keyptr,"STORE",imap_parse_data (stream,msgno,txtptr,reply))
    IFEQUAL (keyptr,"COPY",)
    sprintf (tmp,"Unknown message data: %d %s",msgno,keyptr);
    mm_log (tmp,WARN);
  }
}

/* Mail Access Protocol parse flag list
 * Accepts: MAIL stream
 *	    parsed reply
 *
 *  The reply->line is yanked out of the parsed reply and stored on
 * stream->flagstring.  This is the original fs_get'd reply string, and
 * has all the flagstrings embedded in it.
 */

void imap_parse_flaglst (stream,reply)
	MAILSTREAM *stream;
	IMAPPARSEDREPLY *reply;
{
  char *text = reply->text;
  char *flag;
  int i;
				/* flush old flagstring and flags if any */
  if (stream->flagstring) fs_give ((void **) &stream->flagstring);
  for (i = 0; i < NUSERFLAGS; ++i) stream->user_flags[i] = NIL;
				/* remember this new one */
  stream->flagstring = reply->line;
  reply->line = NIL;
  ++text;			/* skip past open parenthesis */
				/* get first flag if any */
  if (flag = (char *) strtok (text," )")) {
    i = 0;			/* init flag index */
				/* add all user flags */
    do if (*flag != '\\') stream->user_flags[i++] = flag;
      while (flag = (char *) strtok (NIL," )"));
  }
}


/* Mail Access Protocol messages have been searched out
 * Accepts: MAIL stream
 *	    list of message numbers
 *
 * Calls external "mail_searched" function to notify main program
 */

void imap_searched (stream,text)
	MAILSTREAM *stream;
	char *text;
{
				/* only do something if have text */
  if (text && (text = (char *) strtok (text," ")))
    for (; text; text = (char *) strtok (NIL," "))
      mail_searched (stream,atoi (text));
}

/* Mail Access Protocol message has been expunged
 * Accepts: MAIL stream
 *	    message number
 *
 * Calls external "mail_searched" function to notify main program
 */

void imap_expunged (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  int i = msgno - 1;
  if (LOCAL->header[i]) fs_give ((void **) &LOCAL->header[i]);
  if (LOCAL->text[i]) fs_give ((void **) &LOCAL->text[i]);
				/* slide down remainder of cache */
  for (i = msgno; i <= stream->nmsgs; ++i) {
    LOCAL->header[i - 1] = LOCAL->header[i];
    LOCAL->text[i - 1] = LOCAL->text[i];
  }
  LOCAL->header[i] = LOCAL->text[i] = NIL;
  mail_expunged (stream,msgno);	/* notify upper level */
}


/* Mail Access Protocol parse data
 * Accepts: MAIL stream
 *	    message #
 *	    text to parse
 *	    parsed reply
 *
 *  This code should probably be made a bit more paranoid about malformed
 * S-expressions.
 */

void imap_parse_data (stream,msgno,text,reply)
	MAILSTREAM *stream;
	int msgno;
	char *text;
	 		      IMAPPARSEDREPLY *reply;
{
  char *prop;
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  ++text;			/* skip past open parenthesis */
				/* parse Lisp-form property list */
  while (prop = (char *) strtok (text," )")) {
				/* point at value */
    text = (char *) strtok (NIL,"\n");
				/* parse the property and its value */
    imap_parse_prop (stream,elt,ucase (prop),&text,reply);
  }
}

/* Mail Access Protocol parse property
 * Accepts: MAIL stream
 *	    cache item
 *	    property name
 *	    property value text pointer
 *	    parsed reply
 */

void imap_parse_prop (stream,elt,prop,txtptr,reply)
	MAILSTREAM *stream;
	MESSAGECACHE *elt;
	char *prop;
	 		      char **txtptr;
	IMAPPARSEDREPLY *reply;
{
  char tmp[MAILTMPLEN];
  int i = elt->msgno - 1;
  IFEQUAL (prop,"ENVELOPE",imap_parse_envelope (stream,elt,txtptr,reply))
  IFEQUAL (prop,"FLAGS",imap_parse_flags (stream,elt,txtptr))
  IFEQUAL (prop,"INTERNALDATE",{
    if (elt->internal_date) fs_give ((void **) &elt->internal_date);
    elt->internal_date = imap_parse_string (stream,txtptr,reply);
  })
  IFEQUAL (prop,"RFC822",{	/* this shouldn't happen with our client */
    if (LOCAL->text[i]) fs_give ((void **) &LOCAL->text[i]);
    LOCAL->text[i] = imap_parse_string (stream,txtptr,reply);
  })
  IFEQUAL (prop,"RFC822.HEADER",{
    if (LOCAL->header[i]) fs_give ((void **) &LOCAL->header[i]);
    LOCAL->header[i] = imap_parse_string (stream,txtptr,reply);
  })
  IFEQUAL (prop,"RFC822.SIZE",elt->rfc822_size = imap_parse_number (txtptr))
  IFEQUAL (prop,"RFC822.TEXT",{
    if (LOCAL->text[i]) fs_give ((void **) &LOCAL->text[i]);
    LOCAL->text[i] = imap_parse_string (stream,txtptr,reply);
  })
  IFEQUAL (prop,"ENCODING",imap_parse_enclist (elt,txtptr))
  if (prop[0] == 'B' && prop[1] == 'O' && prop[2] == 'D' && prop[3] == 'Y') {
    imap_parse_body (stream,elt,prop+4,txtptr,reply);
    return;
  }
  sprintf (tmp,"Unknown message property: %s",prop);
  mm_log (tmp,WARN);
}

/* Mail Access Protocol parse envelope
 * Accepts: MAIL stream
 *	    current text pointer
 *	    parsed reply
 * Returns:  envelope, NIL on failure
 *
 * Updates text pointer
 */

void imap_parse_envelope (stream,elt,txtptr,reply)
	MAILSTREAM *stream;
	MESSAGECACHE *elt;
	char **txtptr;
	 			  IMAPPARSEDREPLY *reply;
{
  char tmp[MAILTMPLEN];
  switch (*((*txtptr)++)) {	/* dispatch on first character */
  case '(':			/* if envelope S-expression */
				/* free any old envelope */
    if (elt->env) mail_free_envelope (&elt->env);
				/* parse the new envelope */
    elt->env = mail_newenvelope ();
    elt->env->date = imap_parse_string (stream,txtptr,reply);
    elt->env->subject = imap_parse_string (stream,txtptr,reply);
    elt->env->from = imap_parse_adrlist (stream,txtptr,reply);
    elt->env->sender = imap_parse_adrlist (stream,txtptr,reply);
    elt->env->reply_to = imap_parse_adrlist (stream,txtptr,reply);
    elt->env->to = imap_parse_adrlist (stream,txtptr,reply);
    elt->env->cc = imap_parse_adrlist (stream,txtptr,reply);
    elt->env->bcc = imap_parse_adrlist (stream,txtptr,reply);
    elt->env->in_reply_to = imap_parse_string (stream,txtptr,reply);
    elt->env->message_id = imap_parse_string (stream,txtptr,reply);
    if (**txtptr != ')') {
      sprintf (tmp,"Junk at end of envelope: %s",*txtptr);
      mm_log (tmp,WARN);
    }
    else ++*txtptr;		/* skip past delimiter */
    break;
  case 'N':			/* if NIL */
  case 'n':
    ++*txtptr;			/* bump past "I" */
    ++*txtptr;			/* bump past "L" */
    break;
  default:
    sprintf (tmp,"Not an envelope: %s",*txtptr);
    mm_log (tmp,WARN);
    break;
  }
}

/* Mail Access Protocol parse address list
 * Accepts: MAIL stream
 *	    current text pointer
 *	    parsed reply
 * Returns: address list, NIL on failure
 *
 * Updates text pointer
 */

ADDRESS *imap_parse_adrlist (stream,txtptr,reply)
	MAILSTREAM *stream;
	char **txtptr;
	 			     IMAPPARSEDREPLY *reply;
{
  char tmp[MAILTMPLEN];
  ADDRESS *adr = NIL;
  char c = **txtptr;		/* sniff at first character */
				/* ignore leading spaces */
  while (c == ' ') c = *++*txtptr;
  ++*txtptr;			/* skip past open paren */
  switch (c) {
  case '(':			/* if envelope S-expression */
    adr = imap_parse_address (stream,txtptr,reply);
    if (**txtptr != ')') {
      sprintf (tmp,"Junk at end of address list: %s",*txtptr);
      mm_log (tmp,WARN);
    }
    else ++*txtptr;		/* skip past delimiter */
    break;
  case 'N':			/* if NIL */
  case 'n':
    ++*txtptr;			/* bump past "I" */
    ++*txtptr;			/* bump past "L" */
    break;
  default:
    sprintf (tmp,"Not an address: %s",*txtptr);
    mm_log (tmp,WARN);
    break;
  }
  return adr;
}

/* Mail Access Protocol parse address
 * Accepts: MAIL stream
 *	    current text pointer
 *	    parsed reply
 * Returns: address, NIL on failure
 *
 * Updates text pointer
 */

ADDRESS *imap_parse_address (stream,txtptr,reply)
	MAILSTREAM *stream;
	char **txtptr;
	 			     IMAPPARSEDREPLY *reply;
{
  char tmp[MAILTMPLEN];
  ADDRESS *adr = NIL;
  ADDRESS *ret = NIL;
  ADDRESS *prev = NIL;
  char c = **txtptr;		/* sniff at first address character */
  switch (c) {
  case '(':			/* if envelope S-expression */
    while (c == '(') {		/* recursion dies on small stack machines */
      ++*txtptr;		/* skip past open paren */
      if (adr) prev = adr;	/* note previous if any */
      adr = mail_newaddr ();	/* instantiate address and parse its fields */
      adr->personal = imap_parse_string (stream,txtptr,reply);
      adr->route_list = imap_parse_string (stream,txtptr,reply);
      adr->mail_box = imap_parse_string (stream,txtptr,reply);
      adr->host = imap_parse_string (stream,txtptr,reply);
      if (**txtptr != ')') {	/* handle trailing paren */
	sprintf (tmp,"Junk at end of address: %s",*txtptr);
	mm_log (tmp,WARN);
      }
      else ++*txtptr;		/* skip past close paren */
      c = **txtptr;		/* set up for while test */
				/* ignore leading spaces in front of next */
      while (c == ' ') c = *++*txtptr;
      if (adr->mail_box) {	/* make sure address has a mailbox */
				/* if no host default to server */
	if (!adr->host) adr->host = cpystr (imap_host (stream));
	if (!ret) ret = adr;	/* if first time note first adr */
				/* if previous link new block to it */
	if (prev) prev->next = adr;
      }
      else {			/* server return sick address, flush it */
	mail_free_address (&adr);
	adr = NIL;		/* get a new one next time through the loop */
      }
    }
    break;

  case 'N':			/* if NIL */
  case 'n':
    *txtptr += 3;		/* bump past NIL */
    break;
  default:
    sprintf (tmp,"Not an address: %s",*txtptr);
    mm_log (tmp,WARN);
    break;
  }
  return ret;
}

/* Mail Access Protocol parse flags
 * Accepts: current message cache
 *	    current text pointer
 *
 * Updates text pointer
 */

void imap_parse_flags (stream,elt,txtptr)
	MAILSTREAM *stream;
	MESSAGECACHE *elt;
	char **txtptr;
{
  char *flag;
  char c;
  elt->user_flags = NIL;	/* zap old flag values */
  elt->seen = elt->deleted = elt->flagged = elt->answered = elt->searched =
    elt->recent = NIL;
  while (T) {			/* parse list of flags */
    flag = ++*txtptr;		/* point at a flag */
				/* scan for end of flag */
    while (**txtptr != ' ' && **txtptr != ')') ++*txtptr;
    c = **txtptr;		/* save delimiter */
    **txtptr = '\0';		/* tie off flag */
    if (*flag != '\0') {	/* if flag is non-null */
				/* if starts with \ must be sys flag */
      if (*flag == '\\') imap_parse_sys_flag (elt,ucase (flag));
				/* otherwise user flag */
      else imap_parse_user_flag (stream,elt,flag);
    }
    if (c == ')') break;	/* quit if end of list */
  }
  ++*txtptr;			/* bump past delimiter */
}

/* Mail Access Protocol parse system flag
 * Accepts: message cache element
 *	    flag name
 */

void imap_parse_sys_flag (elt,flag)
	MESSAGECACHE *elt;
	char *flag;
{
  char *tmp;
				/* real flags */
  IFEQUAL (flag,"\\SEEN",elt->seen = T);
  IFEQUAL (flag,"\\DELETED",elt->deleted = T);
  IFEQUAL (flag,"\\FLAGGED",elt->flagged = T);
  IFEQUAL (flag,"\\ANSWERED",elt->answered = T);
  IFEQUAL (flag,"\\XXXX",);	/* coddle TOPS-20 server */
  IFEQUAL (flag,"\\YYYY",);
  if (tmp = (char *) strpbrk (flag,"#")) *tmp = '\0';
  IFEQUAL (flag,"\\UNDEFINEDFLAG",);
				/* pseudo-flag */
  IFEQUAL (flag,"\\RECENT",elt->recent = T);
  sprintf (tmp,"Unknown system flag: %s",flag);
  mm_log (tmp,WARN);
}


/* Mail Access Protocol parse user flag
 * Accepts: message cache element
 *	    flag name
 */

void imap_parse_user_flag (stream,elt,flag)
	MAILSTREAM *stream;
	MESSAGECACHE *elt;
	char *flag;
{
  char tmp[MAILTMPLEN];
  int i;
  				/* sniff through all user flags */
  for (i = 0; i < NUSERFLAGS; ++i)
  				/* match this one? */
    if (!strcmp (flag,stream->user_flags[i])) {
      elt->user_flags |= 1 << i;/* yes, set the bit for that flag */
      return;			/* and quit */
    }
  sprintf (tmp,"Unknown user flag: %s",flag);
  mm_log (tmp,WARN);
}

/* Mail Access Protocol parse string
 * Accepts: MAIL stream
 *	    current text pointer
 *	    parsed reply
 * Returns: string
 *
 * Updates text pointer
 */

char *imap_parse_string (stream,txtptr,reply)
	MAILSTREAM *stream;
	char **txtptr;
	 			 IMAPPARSEDREPLY *reply;
{
  char tmp[MAILTMPLEN+1];
  char *st;
  char *string = NIL;
  unsigned long i,j;
  char c = **txtptr;		/* sniff at first character */
				/* ignore leading spaces */
  while (c == ' ') c = *++*txtptr;
  st = ++*txtptr;		/* remember start of string */
  switch (c) {
  case '"':			/* if quoted string */
    i = 1;			/* initial byte count */
    while (**txtptr != '"') {	/* search for end of string */
      ++i;			/* bump count */
      ++*txtptr;		/* bump pointer */
    }
    **txtptr = '\0';		/* tie off string */
    string = (char *) fs_get (i);
    strncpy (string,st,i);	/* copy the string */
    ++*txtptr;			/* bump past delimiter */
    break;
  case 'N':			/* if NIL */
  case 'n':
    ++*txtptr;			/* bump past "I" */
    ++*txtptr;			/* bump past "L" */
    break;

  case '{':			/* if literal string */
				/* get size of string */
    i = imap_parse_number (txtptr);
#ifdef __MSDOS__
				/* truncate? */
    if (j = (i > MAXMESSAGESIZE)) {
      sprintf (tmp,"%d character literal truncated to %d characters",
	       i,MAXMESSAGESIZE);
      mm_log (tmp,WARN);	/* warn user */
      j = i - MAXMESSAGESIZE;	/* number of bytes of slop */
      i = MAXMESSAGESIZE;	/* maximum length string we can read */
    }
#endif
    string = (char *) fs_get (i + 1);
    *string = '\0';		/* init in case getbuffer fails */
				/* get the literal */
    tcp_getbuffer (LOCAL->tcpstream,i,string);
#ifdef __MSDOS__
				/* toss out everything after that */
    for (; j > 0; j -= MAILTMPLEN)
      tcp_getbuffer (LOCAL->tcpstream,min (MAILTMPLEN,j),tmp);
#endif
    fs_give ((void **) &reply->line);
				/* get new reply text line */
    reply->line = tcp_getline (LOCAL->tcpstream);
    if (stream->debug) mm_dlog (reply->line);
    *txtptr = reply->line;	/* set text pointer to point at it */
    break;
  default:
    sprintf (tmp,"Not a string: %c%s",c,*txtptr);
    mm_log (tmp,WARN);
    break;
  }
  return string;
}


/* Mail Access Protocol parse number
 * Accepts: current text pointer
 * Returns: number parsed
 *
 * Updates text pointer
 */

unsigned long imap_parse_number (txtptr)
	char **txtptr;
{				/* parse number */
  char tmp[MAILTMPLEN];
  long i = strtol (*txtptr,txtptr,10);
  if (i < 0) {			/* number valid? */
    sprintf (tmp,"Bad number: %d",i);
    mm_log (tmp,WARN);
    i = 0;			/* make sure valid */
  }
  return (unsigned long) i;
}

/* Mail Access Protocol parse encoding list
 * Accepts: message cache element
 *	    current text pointer
 *
 * Updates text pointer, stores encoding list in the body
 */

void imap_parse_enclist (elt,txtptr)
	MESSAGECACHE *elt;
	char **txtptr;
{
  char tmp[MAILTMPLEN];
  char *s;
  BODY *enc = NIL;
  BODY *new,*old = NIL;
  switch (*((*txtptr)++)) {	/* dispatch on first character */
  case '(':			/* parse encoding list */
    while (new = imap_parse_encoding (txtptr)) {
				/* add new encoding to the list */
      if (enc) old = old->next = new;
      else enc = old = new;	/* instantiate new list */
    }
    if (**txtptr != ')') {	/* validate ending */
      sprintf (tmp,"Junk at end of encoding: %s",*txtptr);
      mm_log (tmp,WARN);
    }
    else ++*txtptr;		/* skip past delimiter */
    break;
  case 'N':			/* if NIL */
  case 'n':
    ++*txtptr;			/* bump past "I" */
    ++*txtptr;			/* bump past "L" */
    break;
  default:
    sprintf (tmp,"Not an encoding list: %s",*txtptr);
    mm_log (tmp,WARN);
    break;
  }
  for (old = elt->body,new = enc; old && new; old = old->next,new = new->next){
    new->text = old->text;	/* move text from old body */
    old->text = NIL;		/* make sure text isn't flushed */
  }
  mail_free_body (&elt->body);	/* flush the old body */
  elt->body = enc;		/* set new body from list */
}

/* Mail Access Protocol parse encoding
 * Accepts: current text pointer
 *
 * Updates text pointer, returns single encoding
 */

BODY *imap_parse_encoding (txtptr)
	char **txtptr;
{
  char tmp[MAILTMPLEN];
  char *s;
  BODY *body;
				/* return immediately if no encoding here */
  if (**txtptr != '(') return NIL;
  body = mail_newbody ();	/* instantiate a new body encoding */
  ++*txtptr;			/* skip past opening parenthesis */
				/* get encoding count */
  body->count = imap_parse_number (txtptr);
				/* type into s, txtptr to end of encoding */
  if ((s = strtok (*txtptr,")")) && (*txtptr = strtok (NIL,"\n")) &&
      (s = strtok (s," "))) {
    switch (*s++) {
    case 'E':			/* possible EDI or ENCRYPTED or EVFU */
      if (!strcmp (s,"DI")) body->type = ENCEDI;
      else if (!strcmp (s,"NCRYPTED")) body->type = ENCENCRYPTED;
      else if (!strcmp (s,"VFU")) body->type = ENCEVFU;
      break;
    case 'H':			/* possible HEX */
      if (!strcmp (s,"EX")) body->type = ENCHEX;
      break;
    case 'M':			/* possible MESSAGE */
      if (!strcmp (s,"ESSAGE")) body->type = ENCMESSAGE;
      break;
    case 'T':			/* possible TEXT */
      if (!strcmp (s,"EXT")) body->type = ENCTEXT;
      break;
    case 'U':			/* possible UUENCODE */
      if (!strcmp (s,"UENCODE")) body->type = ENCUUENCODE;
      break;
    case 'X':			/* possible X.400 */
      if (!strcmp (s,".400")) body->type = ENCX400;
      break;
    default:			/* unknown type */
      body->type = ENCUNDEF;
      break;
    }
				/* parse options */
    if (s = strtok (NIL,"\n")) body->options = cpystr (s);
  }
  else mm_log ("Unparsable encoding",WARN);
  return body;
}

/* Mail Access Protocol parse body
 * Accepts: MAIL stream
 *	    message number
 *	    pointer to segment
 *	    current text pointer
 *	    parsed reply
 *
 * Updates text pointer, stores body section
 */

void imap_parse_body (stream,elt,seg,txtptr,reply)
	MAILSTREAM *stream;
	MESSAGECACHE *elt;
	char *seg;
	 		      char **txtptr;
	IMAPPARSEDREPLY *reply;
{
  long i = -1;
  char tmp[MAILTMPLEN];
  char *s;
  BODY *prev = NIL,*new;
  BODY *body = NIL;
  BODY **old = &elt->body;
  switch (*seg++) {		/* dispatch based on type of fetch */
  case '\0':			/* full body structure */
    break;
  case '[':			/* body section index */
				/* get index of body section */
    if ((s = strchr (seg,']')) && (!(*s = s[1])) &&
	((i = imap_parse_number (&seg)) >= 0)) break;
    sprintf (tmp,"Bad body section %d: %s",i,*txtptr);
    mm_log (tmp,WARN);
    return;
  default:
    sprintf (tmp,"Bad body fetch: %s",*txtptr);
    mm_log (tmp,WARN);
    return;
  }

  switch (*((*txtptr)++)) {	/* parse body list as argument */
  case '(':			/* body list */
    while (new = imap_parse_encoding (txtptr)) {
				/* add new body section to the list */
      if (body) prev = prev->next = new;
      else prev = body = new;
      prev->text = imap_parse_string (stream,txtptr,reply);
    }
    if (**txtptr != ')') {	/* make sure end is valid */
      sprintf (tmp,"Junk at end of body: %s",*txtptr);
      mm_log (tmp,WARN);
    }
    else ++*txtptr;		/* skip past delimiter */
    break;
  case 'N':			/* if NIL */
  case 'n':
    ++*txtptr;			/* bump past "I" */
    ++*txtptr;			/* bump past "L" */
    break;
  default:
    sprintf (tmp,"Bad body: %s",*txtptr);
    mm_log (tmp,WARN);
    return;
  }

  if (i >= 0) {			/* single body section */
    if (body->next) {		/* better be only one element */
      sprintf (tmp,"Multiple sections in single section %d",i);
      mm_log (tmp,WARN);
      mail_free_body (&body);	/* flush this body */
      return;			/* gruesome lossage */
    }
				/* locate body section */
    while (i--) old = &(*old ? *old : (*old = mail_newbody ()))->next;
    if (*old) {			/* copy poop if something already there */
      (*old)->count = body->count;
      (*old)->type = body->type;
      (*old)->options = body->options;
      (*old)->text = body->text;
      body->text = NIL;		/* retain text */
      mail_free_body (&body);	/* flush this */
    }
    else *old = body;		/* set body */
  }
  else {
    new = body;			/* prepare to copy the body list */
    if (*old) do {		/* copy bodies */
      (*old)->count = new->count;
      (*old)->type = new->type;
      (*old)->options = new->options;
      (*old)->text = new->text;
      new->text = NIL;		/* retain text */
      old = &(*old ? *old : (*old = mail_newbody ()))->next;
    } while (new = new->next);
    else *old = body;		/* set body */
  }
}
