/*
 * Program:	Mail Access (MAIL) definitions
 *
 * Author(s):	Mark Crispin
 *		Computing And Communications
 *		University of Washington
 *		Seattle, WA  98195
 *
 * Date:	22 November 1989
 *
 * Copyright 1989, 1990, 1991 by University of Washington.
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license.
 *
 */


/* Build parameters */

#define MAXMESSAGES 5000	/* size of messagearray */
#define MAILTMPLEN 1024		/* size of a temporary buffer */
#define MAXMESSAGESIZE 65000	/* MS-DOS: maximum text buffer size
				 * other:  initial text buffer size */
#define NUSERFLAGS 30		/* # of user flags (current servers 30 max) */


/* Constants */

#define NIL 0			/* convenient name */
#define T 1			/* opposite of NIL */

#define WARN 1			/* mm_log warning type */
#define ERROR 2			/* mm_log error type */

/* Message structures */


/* Item in an address list */

#define ADDRESS struct mail_address
ADDRESS {
  char *personal;		/* personal name phrase */
  char *route_list;		/* at-domain-list source route */
  char *mail_box;		/* mailbox name */
  char *host;			/* domain name of mailbox's host */
  char *error;			/* error in address */
  ADDRESS *next;		/* pointer to next address in list */
};


/* Message envelope */

typedef struct mail_envelope {
  char *remail;			/* remail header if any */
  ADDRESS *return_path;		/* error return address */
  char *date;			/* message composition date string */
  ADDRESS *from;		/* originator address list */
  ADDRESS *sender;		/* sender address list */
  ADDRESS *reply_to;		/* reply address list */
  char *subject;		/* message subject string */
  ADDRESS *to;			/* primary recipient list */
  ADDRESS *cc;			/* secondary recipient list */
  ADDRESS *bcc;			/* blind secondary recipient list */
  char *in_reply_to;		/* replied message ID */
  char *message_id;		/* message ID */
} ENVELOPE;

/* Message body (encoded according to RFC 1154) */

#define BODY struct mail_body
BODY {
  int count;			/* line count */
  short type;			/* encoding type */
  char *options;		/* type-specific options */
  char *text;			/* text of this section */
  BODY *next;			/* pointer to next body section in list */
};


/* Body encoding types - must match ent vector in rfc822.c */

#define ENCTEXT 0		/* text (no particular encoded format) */
#define ENCMESSAGE 1		/* encapsulated message */
#define ENCHEX 2		/* binary data in hexadecimal */
#define ENCEVFU 3		/* electronic VFU (FORTRAN carriage control) */
#define ENCEDI 4		/* electronic document interchange */
#define ENCX400 5		/* CCITT X.400 unconverted section */
#define ENCUUENCODE 6		/* uuencode section */
#define ENCENCRYPTED 7		/* RFC 1115 encryption */
#define ENCUNDEF 8		/* undefined type or option */

extern char *ent[];		/* body encoding types in rfc822.c */


/* Entry in the message cache array (messagearray) */

typedef struct message_cache {
  short lockcount;		/* non-zero if multiple references */
  unsigned int msgno;		/* message number (= array index) */
  char *internal_date;		/* date message placed in mailbox */
  unsigned int seen : 1;	/* system Seen flag */
  unsigned int deleted : 1;	/* system Deleted flag */
  unsigned int flagged : 1; 	/* system Flagged flag */
  unsigned int answered : 1;	/* system Answered flag */
  unsigned int searched : 1;	/* message was searched */
  unsigned int recent : 1;	/* message is new as of this mailbox open */
  unsigned long user_flags;	/* user-assignable flags */
  unsigned long rfc822_size;	/* # of bytes of message in RFC822 */
  ENVELOPE *env;		/* pointer to message envelope */
  BODY *body;			/* pointer to message body */
  char *from;			/* from text for menu display */
  char *subject;		/* subject text for menu display */
} MESSAGECACHE;

/* Mail Access I/O stream */


/* Structure for mail driver dispatch */

#define DRIVER struct driver	


/* Mail I/O stream */

typedef struct mail_stream {
  DRIVER *dtb;			/* dispatch table for this driver */
  void *local;			/* pointer to driver's other data */
  char *mailbox;		/* mailbox name */
  short lock;			/* stream lock */
  short debug;			/* stream debug flag */
  short use;			/* stream use count */
  short sequence;		/* stream sequence */
  int gensym;			/* generated tag */
  unsigned int nmsgs;		/* # of associated msgs */
  unsigned int recent;		/* # of recent msgs */
  char *flagstring;		/* string where flags are located */
  char *user_flags[NUSERFLAGS];	/* list of user flags in bit order */
				/* message cache array */
  MESSAGECACHE *messagearray[MAXMESSAGES];
} MAILSTREAM;


/* Mail I/O stream handle */

typedef struct mail_stream_handle {
  MAILSTREAM *stream;		/* pointer to mail stream */
  short sequence;		/* sequence of what we expect stream to be */
} MAILHANDLE;

/* Mail driver dispatch */

DRIVER {
  DRIVER *next;			/* next driver */
				/* mailbox is valid for us */
  DRIVER *(*valid) ();
				/* open mailbox */
  MAILSTREAM *(*open) ();
				/* close mailbox */
  void (*close) ();
				/* fetch message "fast" attributes */
  void (*fetchfast) ();
				/* fetch message flags */
  void (*fetchflags) ();
				/* fetch message envelopes */
  ENVELOPE *(*fetchenvelope) ();
				/* fetch message header only */
  char *(*fetchheader) ();
				/* fetch message body only */
  char *(*fetchtext) ();
				/* fetch message body section */
  char *(*fetchbody) ();
				/* set message flag */
  void (*setflag) ();
				/* clear message flag */
  void (*clearflag) ();
				/* search for message based on criteria */
  void (*search) ();
				/* ping mailbox to see if still alive */
  MAILSTREAM *(*ping) ();
				/* check for new messages */
  void (*check) ();
				/* expunge deleted messages */
  void (*expunge) ();
				/* copy messages to another mailbox */
  void (*copy) ();
				/* move messages to another mailbox */
  void (*move) ();
};

/* Coddle certain compilers' 6-character symbol limitation */

#ifdef __COMPILER_KCC__
#define mail_link mlink
#define mail_open mopen
#define mail_close mclose
#define mail_makehandle mmkhdl
#define mail_free_handle mfrhdl
#define mail_stream mstrm
#define mail_fetchfast mffast
#define mail_fetchflags mfflgs
#define mail_fetchenvelope mfenv
#define mail_fetchheader mfhdr
#define mail_fetchtext mftxt
#define mail_fetchbody mfbody
#define mail_fetchfrom mffrom
#define mail_fetchsubject mfsubj
#define mail_elt melt
#define mail_setflag msflag
#define mail_clearflag mcflag
#define mail_search msrch
#define mail_ping mping
#define mail_check mcheck
#define mail_expunge mexpng
#define mail_copy mcopy
#define mail_move mmove
#define mail_searched msrchd
#define mail_exists mexist
#define mail_recent mrcent
#define mail_expunged mepngd
#define mail_lock mlock
#define mail_unlock mulock
#define mail_checklock mclock
#define mail_debug mdbug
#define mail_nodebug mndbug
#define mail_newenvelope mnwenv
#define mail_newaddr mnwadr
#define mail_newbody mnwbod
#define mail_free_envelope mfrenv
#define mail_free_address mfradr
#define mail_free_body mfrbod
#define mail_free_messagearray mfrmar
#define mail_free_elt mfrelt
#endif

/* Function prototypes */

void mm_searched  ();
void mm_exists  ();
void mm_expunged  ();
void mm_notify  ();
void mm_log  ();
void mm_dlog  ();
void mm_login  ();

void mail_link  ();
MAILSTREAM *mail_open  ();
MAILSTREAM *mail_close  ();
MAILHANDLE *mail_makehandle  ();
void mail_free_handle  ();
MAILSTREAM *mail_stream  ();
void mail_fetchfast  ();
void mail_fetchflags  ();
ENVELOPE *mail_fetchenvelope  ();
char *mail_fetchheader  ();
char *mail_fetchtext  ();
char *mail_fetchbody  ();
char *mail_fetchfrom  ();
char *mail_fetchsubject  ();
MESSAGECACHE *mail_elt  ();
void mail_setflag  ();
void mail_clearflag  ();
void mail_search  ();
MAILSTREAM *mail_ping  ();
void mail_check  ();
void mail_expunge  ();
void mail_copy  ();
void mail_move  ();
void mail_searched  ();
void mail_exists  ();
void mail_recent  ();
void mail_expunged  ();
void mail_lock  ();
void mail_unlock  ();
short mail_checklock  ();
void mail_debug  ();
void mail_nodebug  ();
ENVELOPE *mail_newenvelope  ();
ADDRESS *mail_newaddr  ();
BODY *mail_newbody  ();
void mail_free_envelope  ();
void mail_free_address  ();
void mail_free_body  ();
void mail_free_messagearray  ();
void mail_free_elt  ();
