/*
 * Program:	Miscellaneous utility routines -- 32 bit version
 *
 * Author(s):	Mark Crispin
 *		Symbolic Systems Resources Group
 *		Knowledge Systems Laboratory
 *		Departments of Computer Science and Medicine
 *		Stanford University
 *		Stanford, CA  94305
 *
 * Date:	5 July 1988
 *
 * Sponsorship:	This work was supported by the Biomedical Research
 *		Technology Program of the National Institutes of Health
 *		under grant number RR-00785.
 *
 * Copyright 1988, 1989, 1990, 1991 by The Leland Stanford Junior University.
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license from the Stanford Office of
 * Technology Licensing.
 *
 */


#include <ctype.h>
#ifdef __STDC__
#include <stdlib.h>
#else
#include "ansi.h"
#endif
#include <sys/types.h>
#include "mail.h"
#include "misc.h"
#include "osdep.h"

/* Convert string to all uppercase
 * Accepts: string pointer
 * Returns: string pointer
 */

char *ucase (string)
	char *string;
{
  char *ret = string;
				/* if lowercase covert to upper */
  for (; *string; string++) if (islower (*string)) *string = toupper (*string);
  return (ret);			/* return string */
}


/* Convert string to all lowercase
 * Accepts: string pointer
 * Returns: string pointer
 */

char *lcase (string)
	char *string;
{
  char *ret = string;
				/* if uppercase covert to lower */
  for (; *string; string++) if (isupper (*string)) *string = tolower (*string);
  return (ret);			/* return string */
}


/* Copy string to free storage
 * Accepts: source string
 * Returns: free storage copy of string
 */

char *cpystr (string)
	char *string;
{
  if (string) {			/* make sure argument specified */
    char *dst = (char *) fs_get (1+strlen (string));
    strcpy (dst,string);
    return (dst);
  }
  else return NIL;
}

/* Returns index of rightmost bit in word
 * Accepts: pointer to a 32 bit value
 * Returns: -1 if word is 0, else index of rightmost bit
 *
 * Bit is cleared in the word
 */

int find_rightmost_bit (valptr)
	long *valptr;
{
  register long value= *valptr;
  register long clearbit;	/* bit to clear */
  register bitno;		/* bit number to return */
  if (value == 0) return (-1);	/* no bits are set */
  if (value & 0xFFFF) {		/* low order halfword has a bit? */
    bitno = 0;			/* yes, start with bit 0 */
    clearbit = 1;		/* which has value 1 */
  } else {			/* high order halfword has the bit */
    bitno = 16;			/* start with bit 16 */
    clearbit = 0x10000;		/* which has value 10000h */
    value >>= 16;		/* and slide the halfword down */
  }
  if (!(value & 0xFF)) {	/* low quarterword has a bit? */
    bitno += 8;			/* no, start 8 bits higher */
    clearbit <<= 8;		/* bit to clear is 2^8 higher */
    value >>= 8;		/* and slide the quarterword down */
  }
  while (T) {			/* search for bit in quarterword */
    if (value & 1) break;	/* found it? */
    value >>= 1;		/* now, slide the bit down */
    bitno += 1;			/* count one more bit */
    clearbit <<= 1;		/* bit to clear is 1 bit higher */
  }
  *valptr ^= clearbit;		/* clear the bit in the argument */
  return (bitno);		/* and return the bit number */
}

/* Return minimum of two integers
 * Accepts: integer 1
 *	    integer 2
 * Returns: minimum
 */

long min (i,j)
	long i;
	long j;
{
  return ((i < j) ? i : j);
}


/* Return maximum of two integers
 * Accepts: integer 1
 *	    integer 2
 * Returns: maximum
 */

long max (i,j)
	long i;
	long j;
{
  return ((i > j) ? i : j);
}

/* Case independent search
 * Accepts: base string
 *	    length of base string
 *	    pattern string
 *	    length of pattern string
 * Returns: T if pattern exists inside base, else NIL
 */

short search (s,c,pat,patc)
	char *s;
	int c;
	char *pat;
	int patc;
{
  register long m0,m1;
				/* validate arguments, c becomes # of tries */
  if (!(s && c > 0 && pat && patc > 0 && (c -= (patc - 1)) > 0)) return NIL;
				/* handle any chars not on word boundary */
  if (((unsigned) s & 3) && ssearch (&s,&c,pat)) return T;
  m0 = *pat * 0x01010101;	/* search mask */
  m1 = m0 ^ 0x20202020;		/* reverse case search mask */
  /*
   * Fast search algorithm XORs the mask with each word from the base string
   * and complements the result. This will give bytes of all ones where there
   * are matches.  We then mask out the high order bit in each byte and add 1
   * to all the bytes.  If we have a resulting carry, then we have a match.
   */
  do {				/* interesting word? */
    if ((0x80808080&(0x01010101+(0x7F7F7F7F&~(m0^*(unsigned long *) s)))) ||
	(0x80808080&(0x01010101+(0x7F7F7F7F&~(m1^*(unsigned long *) s))))) {
				/* yes, commence a slow search through it */
      if (ssearch (&s,&c,pat)) return T;
    }
    else {s += 4;c -= 4;}	/* try next word */
  } while (c >= 0);		/* continue until end of string */
  return NIL;			/* string not found */
}


/* Case independent slow search within a word
 * Accepts: base string
 *	    number of tries left
 *	    pattern string
 * Returns: T if pattern exists inside base, else NIL
 */

short ssearch (s,c,pat)
	char **s;
	int *c;
	char *pat;
{
  register char *p;
  register short ch;
  register char *ss = *s;	/* remember were we began */
  do {				/* search character at a time */
    p = pat;			/* start at beginning of pattern */
    do if (*p == '\0') return T;/* match case-independent until end */
    while (((ch = (short)(*p++ ^ *ss++)) == 0) || ch == 040);
    if ((--*c) <= 0) return NIL;/* punt if ran out of tries */
    ss = ++*s;			/* try next character */
  } while ((unsigned) *s & 3);	/* stop when we hit a word boundary */
  return NIL;			/* string not found */
}
