/*
 * Program:	Simple Mail Transfer Protocol (SMTP) routines
 *
 * Author(s):	Mark Crispin
 *		Symbolic Systems Resources Group
 *		Knowledge Systems Laboratory
 *		Departments of Computer Science and Medicine
 *		Stanford University
 *		Stanford, CA  94305
 *
 * Date:	27 July 1988
 *
 * Sponsorship:	This work was supported by the Biomedical Research
 *		Technology Program of the National Institutes of Health
 *		under grant number RR-00785.
 *
 * Copyright 1988, 1989, 1990, 1991 by The Leland Stanford Junior University.
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license from the Stanford Office of
 * Technology Licensing.
 *
 */


#include <ctype.h>
#ifdef __STDC__
#include <string.h>
#else
#include "ansi.h"
#endif
#include <sys/types.h>
#include "mail.h"
#include "smtp.h"
#include "rfc822.h"
#include "misc.h"
#include "osdep.h"
#include "tcpsio.h"



/* Mail Transfer Protocol routines
 *
 *  mtp_xxx routines are the interface between this module and the outside
 * world.  Only these routines should be referenced by external callers.
 * The idea is that if SMTP is ever replaced by some other mail transfer
 * protocol this discipline may reduce the amount of conversion effort.
 *
 */

/* Mail Transfer Protocol open connection
 * Accepts: service host list
 *	    initial debugging flag
 * Returns: T on success, NIL on failure
 */

SMTPSTREAM *mtp_open (hostlist,debug)
	char **hostlist;
	short debug;
{
  int reply;
  SMTPSTREAM *stream = NIL;
  void *tcpstream;
  char tmp[MAILTMPLEN];
  if (!(hostlist && *hostlist)) mm_log ("Missing MTP service host",ERROR);
  else do {			/* try to open connection */
    if (tcpstream = tcp_open (*hostlist,SMTPTCPPORT)) {
      stream = (SMTPSTREAM *) fs_get (sizeof (SMTPSTREAM));
      stream->tcpstream = tcpstream;
      stream->lock = NIL;
      stream->debug = debug;
      stream->reply = NIL;
				/* get SMTP greeting */
      if (smtp_reply (stream) == SMTPGREET &&
	  ((smtp_send (stream,"HELO",tcp_localhost (tcpstream)) == SMTPOK) ||
	   ((!strcmp ("localhost",lcase (strcpy (tmp,*hostlist)))) &&
	    (smtp_send (stream,"HELO","localhost") == SMTPOK))))
	return stream;		/* success return */
      mtp_close (stream);	/* otherwise punt stream */
    }
  } while (*++hostlist);	/* try next server */
  return NIL;
}


/* Mail Transfer Protocol close connection
 * Accepts: stream
 */

void mtp_close (stream)
	SMTPSTREAM *stream;
{
  if (stream) {			/* send "QUIT" */
    smtp_send (stream,"QUIT",NIL);
				/* close TCP connection */
    tcp_close (stream->tcpstream);
    if (stream->reply) fs_give ((void **) &stream->reply);
    fs_give ((void **) &stream);/* flush the stream */
  }
}

/* Mail Transfer Protocol deliver mail
 * Accepts: stream
 *	    delivery option (MAIL, SEND, SAML, SOML)
 *	    message envelope
 *	    message body
 * Returns: T on success, NIL on failure
 */

int mtp_mail (stream,type,env,body)
	SMTPSTREAM *stream;
	char *type;
	ENVELOPE *env;
	BODY *body;
{
  char tmp[8*MAILTMPLEN];
  int error = NIL;
  char *fake;
  if (!(env->to || env->cc || env->bcc)) {
  				/* no recipients in request */
    smtp_fake (stream,SMTPHARDERROR,"No recipients specified");
    return NIL;
  }
  				/* make sure stream is in good shape */
  smtp_send (stream,"RSET",NIL);
  strcpy (tmp,"FROM:<");	/* compose "MAIL FROM:<return-path>" */
  rfc822_address (tmp,env->return_path);
  strcat (tmp,">");
				/* send "MAIL FROM" command */
  if (!(smtp_send (stream,type,tmp) == SMTPOK)) return NIL;
				/* negotiate the recipients */
  if (env->to) smtp_rcpt (stream,env->to,&error);
  if (env->cc) smtp_rcpt (stream,env->cc,&error);
  if (env->bcc) smtp_rcpt (stream,env->bcc,&error);
  if (error) {			/* any recipients failed? */
      				/* reset the stream */
    smtp_send (stream,"RSET",NIL);
    smtp_fake (stream,SMTPHARDERROR,"One or more recipients failed");
    return NIL;
  }
				/* negotiate data command */
  if (!(smtp_send (stream,"DATA",NIL) == SMTPREADY)) return NIL;
				/* set up error in case failure */
  smtp_fake (stream,SMTPSOFTFATAL,"SMTP connection went away!");
  rfc822_header (tmp,env,body);	/* get RFC822 header */
				/* send header */
  if (!(tcp_soutr (stream->tcpstream,tmp))) return NIL;
				/* for each body section */
  if (body) do if (body->text && *body->text &&
		   !(tcp_soutr (stream->tcpstream,body->text) &&
		     tcp_soutr (stream->tcpstream,"\015\012"))) return NIL;
  while (body = body->next);	/* fetch next body section */
				/* end of mail data */
  return smtp_send (stream,".",NIL) == SMTPOK;
}

/* Mail Transfer Protocol parse address
 * Accepts: input string
 *	    default host name
 * Returns: address list
 */

ADDRESS *mtp_parse_address (string,defaulthost)
	char *string;
	char *defaulthost;
{
  ADDRESS *adr = NIL;
  rfc822_parse_adrlist (&adr,string,defaulthost);
  return adr;
}


/* Mail Transfer Protocol turn on debugging telemetry
 * Accepts: stream
 */

void mtp_debug (stream)
	SMTPSTREAM *stream;
{
  stream->debug = T;		/* turn on protocol telemetry */
}


/* Mail Transfer Protocol turn off debugging telemetry
 * Accepts: stream
 */

void mtp_nodebug (stream)
	SMTPSTREAM *stream;
{
  stream->debug = NIL;		/* turn off protocol telemetry */
}

/* Internal routines */


/* Simple Mail Transfer Protocol send recipient
 * Accepts: SMTP stream
 *	    address list
 *	    pointer to error flag
 */

void smtp_rcpt (stream,adr,error)
	SMTPSTREAM *stream;
	ADDRESS *adr;
	int *error;
{
  char tmp[MAILTMPLEN];
  while (adr) {
				/* clear any former error */
    if (adr->error) fs_give ((void **) &adr->error);
    strcpy (tmp,"TO:<");	/* compose "RCPT TO:<return-path>" */
    rfc822_address (tmp,adr);
    strcat (tmp,">");
				/* send "RCPT TO" command */
    if (!(smtp_send (stream,"RCPT",tmp) == SMTPOK)) {
      *error = T;		/* note that an error occurred */
      adr->error = cpystr (stream->reply);
    }
    adr = adr->next;		/* do any subsequent recipients */
  }
}


/* Simple Mail Transfer Protocol send command
 * Accepts: SMTP stream
 *	    text
 * Returns: reply code
 */

int smtp_send (stream,command,args)
	SMTPSTREAM *stream;
	char *command;
	char *args;
{
  char tmp[MAILTMPLEN];
				/* build the complete command */
  if (args) sprintf (tmp,"%s %s",command,args);
  else sprintf (tmp,command);
  if (stream->debug) mm_dlog (tmp);
  strcat (tmp,"\015\012");
				/* send the command */
  return tcp_soutr (stream->tcpstream,tmp) ? smtp_reply (stream) :
    smtp_fake (stream,SMTPSOFTFATAL,"SMTP connection went away!");
}

/* Simple Mail Transfer Protocol get reply
 * Accepts: SMTP stream
 * Returns: reply code
 */

int smtp_reply (stream)
	SMTPSTREAM *stream;
{
				/* flush old reply */
  if (stream->reply) fs_give ((void **) &stream->reply);
  				/* get reply */
  if (!(stream->reply = tcp_getline (stream->tcpstream)))
    return smtp_fake (stream,SMTPSOFTFATAL,"SMTP connection went away!");
  if (stream->debug) mm_dlog (stream->reply);
				/* handle continuation by recursion */
  return (stream->reply[3]=='-') ? smtp_reply (stream) : atoi (stream->reply);
}


/* Simply Mail Transfer Protocol set fake error
 * Accepts: SMTP stream
 *	    SMTP error code
 *	    error text
 * Returns: error code
 */

int smtp_fake (stream,code,text)
	SMTPSTREAM *stream;
	int code;
	char *text;
{
				/* flush any old reply */
  if (stream->reply ) fs_give ((void **) &stream->reply);
  				/* set up pseudo-reply string */
  stream->reply = (char *) fs_get (20+strlen (text));
  sprintf (stream->reply,"%d %s",code,text);
  return code;			/* return error code */
}
