/*
 * Program:	Tenex mail (TENEX) routines
 *
 * Author(s):	Mark Crispin
 *		Computing And Communications
 *		University of Washington
 *		Seattle, WA  98195
 *
 * Date:	22 May 1990
 *
 * Copyright 1990, 1991 by University of Washington.
 *
 * This program may be distributed without restriction for non-commercial use.
 * Any sale or use of this program or adaptations thereof for commercial
 * purposes is prohibited except under license.
 *
 */


#include <stdio.h>
#include <ctype.h>
#ifdef __STDC__
#include <string.h>
#include <stdlib.h>
#else
#include "ansi.h"
#endif
#include <pwd.h>
#include <netdb.h>
#include <errno.h>
extern int errno;		/* just in case */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/uio.h>
#include "mail.h"
#include "tenex.h"
#include "rfc822.h"
#include "misc.h"
#include "osdep.h"

/* Tenex mail routines */


/* Driver dispatch used by MAIL */

DRIVER tenexdriver = {
  (DRIVER *) NIL,		/* next driver */
  tenex_valid,			/* mailbox is valid for us */
  tenex_open,			/* open mailbox */
  tenex_close,			/* close mailbox */
  tenex_fetchfast,		/* fetch message "fast" attributes */
  tenex_fetchflags,		/* fetch message flags */
  tenex_fetchenvelope,		/* fetch message envelopes */
  tenex_fetchheader,		/* fetch message header only */
  tenex_fetchtext,		/* fetch message body only */
  tenex_fetchbody,		/* fetch message body section */
  tenex_setflag,		/* set message flag */
  tenex_clearflag,		/* clear message flag */
  tenex_search,			/* search for message based on criteria */
  tenex_ping,			/* ping mailbox to see if still alive */
  tenex_check,			/* check for new messages */
  tenex_expunge,		/* expunge deleted messages */
  tenex_copy,			/* copy messages to another mailbox */
  tenex_move			/* move messages to another mailbox */
};


/* Tenex mail validate mailbox
 * Accepts: mailbox name
 * Returns: our driver if name is valid, otherwise calls valid in next driver
 */

DRIVER *tenex_valid (name)
	char *name;
{
  int fd;
  char tmp[MAILTMPLEN];
  struct stat sbuf;
				/* if file, get its status */
  if (*name != '{' && (stat (tenex_file (tmp,name),&sbuf) == 0)) {
    if (sbuf.st_size != 0) {	/* if non-empty file */
      if ((fd = open (tmp,O_RDONLY,NIL)) >= 0 && read (fd,tmp,23) >= 0) {
	close (fd);		/* close the file */
				/* must begin with dd-mmm-yy hh:mm:ss" */
	if (tmp[2] == '-' && tmp[6] == '-' && tmp[9] == ' ' &&
	    tmp[12] == ':' && tmp[15] == ':') return &tenexdriver;
      }
    }
				/* allow empty if a ".txt" file */
    else if (strstr (name,".txt")) return &tenexdriver;
  }
				/* else let next driver have a try */
  return tenexdriver.next ? (*tenexdriver.next->valid) (name) : NIL;
}

/* Tenex mail open
 * Accepts: stream to open
 * Returns: stream on success, NIL on failure
 */

MAILSTREAM *tenex_open (stream)
	MAILSTREAM *stream;
{
  int i,fd;
  char *s,*t,*k;
  char tmp[MAILTMPLEN];
  struct stat sbuf;
  struct hostent *host_name;
  if (LOCAL) {			/* close old file if stream being recycled */
    tenex_close (stream);	/* dump and save the changes */
    stream->dtb = &tenexdriver;	/* reattach this driver */
				/* clean up messagearray (again) */
    mail_free_messagearray (stream->messagearray);
  }
  else {			/* flush old flagstring and flags if any */
    if (stream->flagstring) fs_give ((void **) &stream->flagstring);
    for (i = 0; i < NUSERFLAGS; ++i) stream->user_flags[i] = NIL;
				/* open .imapinit or .mminit file */
    if ((fd = open (strcat (strcpy (tmp,getpwuid (geteuid ())->pw_dir),
			    "/.imapinit"),O_RDONLY,NIL)) < 0)
      fd = open (strcat (strcpy (tmp,getpwuid (geteuid ())->pw_dir),
			 "/.mminit"),O_RDONLY,NIL);
    if (fd >= 0) {		/* got an init file? */
      fstat (fd,&sbuf);		/* yes, get size */
      read (fd,(s = stream->flagstring = (char *) fs_get (sbuf.st_size + 1)),
	    sbuf.st_size);	/* make readin area and read the file */
      close (fd);		/* don't need the file any more */
      s[sbuf.st_size] = '\0';	/* tie it off */
				/* parse init file */
      while (*s && (t = strchr (s,'\n'))) {
	*t = '\0';		/* tie off line, find second space */
	if ((k = strchr (s,' ')) && (k = strchr (++k,' '))) {
	  *k = '\0';		/* tie off two words, is it "set keywords"? */
	  if (!strcmp (lcase (s),"set keywords")) {
				/* yes, get first keyword */
	    k = strtok (++k,", ");
				/* until parsed all keywords or empty */
	    for (i = 0; k && i < NUSERFLAGS; ++i) {
				/* set keyword, get next one */
	      stream->user_flags[i] = k;
	      k = strtok (NIL,", ");
	    }
	    break;		/* don't need to look at rest of file */
	  }
	}
	s = ++t;		/* try next line */
      }
    }
  }

  if ((fd = open (tenex_file (tmp,stream->mailbox),O_RDWR,NIL)) < 0) {
    sprintf (tmp,"Can't open mailbox: %s",strerror (errno));
    mm_log (tmp,ERROR);
  }
  else {
    stream->local = fs_get (sizeof (TENEXLOCAL));
				/* note if an INBOX or not */
    LOCAL->inbox = !strcmp (ucase (stream->mailbox),"INBOX");
				/* canonicalize the stream mailbox name */
    fs_give ((void **) &stream->mailbox);
    stream->mailbox = cpystr (tmp);
    gethostname(tmp,MAILTMPLEN);/* get local host name */
    LOCAL->host = cpystr ((host_name = gethostbyname (tmp)) ?
			  host_name->h_name : tmp);
				/* bind and lock the file */
    flock (LOCAL->fd = fd,LOCK_SH);
    LOCAL->filesize = 0;	/* initialize parsed file size */
    for (i = 0; i < MAXMESSAGES; ++i) LOCAL->msgs[i] = NIL;
    LOCAL->text = NIL;		/* no text yet */
    LOCAL->textsize = LOCAL->textend = 0;
    LOCAL->buf = (char *) fs_get (MAXMESSAGESIZE + 1);
    LOCAL->buflen = MAXMESSAGESIZE;
    stream->sequence++;		/* bump sequence number */
				/* parse mailbox */
    stream->nmsgs = stream->recent = 0;
    if (tenex_ping (stream) && !stream->nmsgs)
      mm_log ("Mailbox is empty",WARN);
  }
  return LOCAL ? stream : NIL;	/* if stream is alive, return to caller */
}

/* Tenex mail close
 * Accepts: MAIL stream
 */

void tenex_close (stream)
	MAILSTREAM *stream;
{
  int i;
  if (stream && LOCAL) {	/* only if a file is open */
    if (LOCAL->host) fs_give ((void **) &LOCAL->host);
    flock (LOCAL->fd,LOCK_UN);	/* unlock local file */
    close (LOCAL->fd);		/* close the local file */
				/* free local cache */
    for (i = 0; i < MAXMESSAGES; ++i)
      if (LOCAL->msgs[i]) fs_give ((void **) &LOCAL->msgs[i]);
				/* free local text buffer */
    if (LOCAL->text) fs_give ((void **) &LOCAL->text);
    if (LOCAL->buf) fs_give ((void **) &LOCAL->buf);
				/* nuke the local data */
    fs_give ((void **) &stream->local);
    stream->dtb = NIL;		/* log out the DTB */
  }
}


/* Tenex mail fetch fast information
 * Accepts: MAIL stream
 *	    sequence
 */

void tenex_fetchfast (stream,sequence)
	MAILSTREAM *stream;
	char *sequence;
{
  return;			/* no-op for local mail */
}


/* Tenex mail fetch flags
 * Accepts: MAIL stream
 *	    sequence
 */

void tenex_fetchflags (stream,sequence)
	MAILSTREAM *stream;
	char *sequence;
{
  return;			/* no-op for local mail */
}

/* Tenex mail fetch envelope
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: envelope of this message
 *
 * Fetches the "fast" information as well
 */

ENVELOPE *tenex_fetchenvelope (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  int i = LOCAL->msgs[msgno-1]->headersize;
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  if (i > LOCAL->buflen)	/* make sure buffer can hold the text */
    fs_resize ((void **) &LOCAL->buf,(LOCAL->buflen = i) + 1);
  if (!elt->env)		/* make envelope now if don't have one */
    rfc822_parse_msg (elt,LOCAL->text + LOCAL->msgs[msgno-1]->header,i,
		      LOCAL->host,LOCAL->buf);
  return elt->env;		/* return the envelope */
}

/* Tenex mail fetch message header
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message header in RFC822 format
 */

char *tenex_fetchheader (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  FILECACHE *m = LOCAL->msgs[msgno-1];
				/* copy the string */
  return strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,
		     LOCAL->text + m->header,m->headersize);
}


/* Tenex mail fetch message text (only)
	body only;
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message text in RFC822 format
 */

char *tenex_fetchtext (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  FILECACHE *m = LOCAL->msgs[msgno-1];
  elt->seen = T;		/* mark message as seen */
				/* recalculate status */
  tenex_update_status (stream,msgno);
  fsync (LOCAL->fd);		/* make sure the update takes */
  return strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,
		     LOCAL->text + m->body,m->bodysize);
}

/* Tenex fetch message body as a structure
 * Accepts: Mail stream
 *	    message # to fetch
 *	    section #
 * Returns: pointer to section of message body
 */

char *tenex_fetchbody (stream,msgno,sec)
	MAILSTREAM *stream;
	int msgno;
	int sec;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  FILECACHE *m = LOCAL->msgs[msgno-1];
  BODY *body;
  int i;
  char *s,*t,*v;
  if (sec < 0) {		/* validate section */
    sprintf (LOCAL->buf,"Bogus section number: %d",sec);
    fatal (LOCAL->buf);
  }
				/* make sure header parsed */
  tenex_fetchenvelope (stream,msgno);
				/* treat as text if no structured body */
  if (!(body = elt->body)) return sec ? "" : tenex_fetchtext (stream,msgno);
  elt->seen = T;		/* yes, mark message as seen */
				/* recalculate status */
  tenex_update_status (stream,msgno);
  fsync (LOCAL->fd);		/* make sure the update takes */
  t = LOCAL->text + m->body;	/* start at body origin */
  v = t + m->bodysize;		/* absolute end of message */
  do {				/* count lines */
    s = t;			/* start of section */
				/* zero count means rest of message */
    if (!(i = body->count)) t = v;
				/* otherwise count lines */
    else while (i && (t < v)) if (*t++ == '\n') i--;
				/* until at the appropriate section */
  } while (sec-- && (!i) && (body = body->next) && (*t++ == '\n'));
				/* return the section requested if found */
  return (sec >= 0) ? "": strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,s,t - s);
}

/* Tenex mail set flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void tenex_setflag (stream,sequence,flag)
	MAILSTREAM *stream;
	char *sequence;
	char *flag;
{
  MESSAGECACHE *elt;
  unsigned char s[MAXMESSAGES];
  int i = stream->nmsgs;
  long uf;
  short f;
				/* no-op if no flags to modify */
  if (!((f = tenex_getflags (stream,flag,&uf)) || uf)) return;
				/* get sequence and loop on it */
  if (tenex_getsequence (stream,s,sequence)) while (i) if (s[--i]) {
				/* get cache entry */
    elt = tenex_elt (stream,i+1);
    if (f&fSEEN) elt->seen = T;	/* set all requested flags */
    if (f&fDELETED) elt->deleted = T;
    if (f&fFLAGGED) elt->flagged = T;
    if (f&fANSWERED) elt->answered = T;
    elt->user_flags |= uf;
				/* recalculate status */
    tenex_update_status (stream,i+1);
    fsync (LOCAL->fd);		/* make sure the update takes */
  }
}

/* Tenex mail clear flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void tenex_clearflag (stream,sequence,flag)
	MAILSTREAM *stream;
	char *sequence;
	char *flag;
{
  MESSAGECACHE *elt;
  unsigned char s[MAXMESSAGES];
  int i = stream->nmsgs;
  long uf;
  short f;
				/* no-op if no flags to modify */
  if (!((f = tenex_getflags (stream,flag,&uf)) || uf)) return;
				/* get sequence and loop on it */
  if (tenex_getsequence (stream,s,sequence)) while (i) if (s[--i]) {
				/* get cache entry */
    elt = tenex_elt (stream,i+1);
				/* clear all requested flags */
    if (f&fSEEN) elt->seen = NIL;
    if (f&fDELETED) elt->deleted = NIL;
    if (f&fFLAGGED) elt->flagged = NIL;
    if (f&fANSWERED) elt->answered = NIL;
    elt->user_flags &= ~uf;
				/* recalculate status */
    tenex_update_status (stream,i+1);
    fsync (LOCAL->fd);		/* make sure the update takes */
  }
}

/* Tenex mail search for messages
 * Accepts: MAIL stream
 *	    search criteria
 */

void tenex_search (stream,criteria)
	MAILSTREAM *stream;
	char *criteria;
{
  long i,n;
  char *d;
  unsigned char s[MAXMESSAGES];
  char (*f) ();
  memset (s,T,MAXMESSAGES);
				/* get first criterion */
  if (criteria && (criteria = strtok (criteria," "))) {
				/* for each criterion */
    for (; criteria; (criteria = strtok (NIL," "))) {
      f = NIL; d = NIL; n = 0;	/* init then scan the criterion */
      switch (*ucase (criteria)) {
      case 'A':			/* possible ALL, ANSWERED */
	if (!strcmp (criteria+1,"LL")) f = tenex_search_all;
	else if (!strcmp (criteria+1,"NSWERED")) f = tenex_search_answered;
	break;
      case 'B':			/* possible BCC, BEFORE, BODY */
	if (!strcmp (criteria+1,"CC"))
	  f = tenex_search_string (tenex_search_bcc,&d,&n);
	else if (!strcmp (criteria+1,"EFORE"))
	  f = tenex_search_date (tenex_search_before,&n);
	else if (!strcmp (criteria+1,"ODY"))
	  f = tenex_search_string (tenex_search_body,&d,&n);
	break;
      case 'C':			/* possible CC */
	if (!strcmp (criteria+1,"C"))
	  f = tenex_search_string (tenex_search_cc,&d,&n);
	break;
      case 'D':			/* possible DELETED */
	if (!strcmp (criteria+1,"ELETED")) f = tenex_search_deleted;
	break;
      case 'F':			/* possible FLAGGED, FROM */
	if (!strcmp (criteria+1,"LAGGED")) f = tenex_search_flagged;
	else if (!strcmp (criteria+1,"ROM"))
	  f = tenex_search_string (tenex_search_from,&d,&n);
	break;
      case 'K':			/* possible KEYWORD */
	if (!strcmp (criteria+1,"EYWORD"))
	  f = tenex_search_flag (tenex_search_keyword,&n,stream);
	break;
      case 'N':			/* possible NEW */
	if (!strcmp (criteria+1,"EW")) f = tenex_search_new;
	break;

      case 'O':			/* possible OLD, ON */
	if (!strcmp (criteria+1,"LD")) f = tenex_search_old;
	else if (!strcmp (criteria+1,"N"))
	  f = tenex_search_date (tenex_search_on,&n);
	break;
      case 'R':			/* possible RECENT */
	if (!strcmp (criteria+1,"ECENT")) f = tenex_search_recent;
	break;
      case 'S':			/* possible SEEN, SINCE, SUBJECT */
	if (!strcmp (criteria+1,"EEN")) f = tenex_search_seen;
	else if (!strcmp (criteria+1,"INCE"))
	  f = tenex_search_date (tenex_search_since,&n);
	else if (!strcmp (criteria+1,"UBJECT"))
	  f = tenex_search_string (tenex_search_subject,&d,&n);
	break;
      case 'T':			/* possible TEXT, TO */
	if (!strcmp (criteria+1,"EXT"))
	  f = tenex_search_string (tenex_search_text,&d,&n);
	else if (!strcmp (criteria+1,"O"))
	  f = tenex_search_string (tenex_search_to,&d,&n);
	break;
      case 'U':			/* possible UN* */
	if (criteria[1] == 'N') {
	  if (!strcmp (criteria+2,"ANSWERED")) f = tenex_search_unanswered;
	  else if (!strcmp (criteria+2,"DELETED")) f = tenex_search_undeleted;
	  else if (!strcmp (criteria+2,"FLAGGED")) f = tenex_search_unflagged;
	  else if (!strcmp (criteria+2,"KEYWORD"))
	    f = tenex_search_flag (tenex_search_unkeyword,&n,stream);
	  else if (!strcmp (criteria+2,"SEEN")) f = tenex_search_unseen;
	}
	break;
      default:			/* we will barf below */
	break;
      }
      if (!f) {			/* if can't determine any criteria */
	sprintf (LOCAL->buf,"Unknown search criterion: %s",criteria);
	mm_log (LOCAL->buf,ERROR);
	return;
      }
				/* run the search criterion */
      for (i = 0; i<stream->nmsgs; ++i) s[i] = (s[i] && (*f) (stream,i+1,d,n));
    }
				/* report search results to main program */
    for (i = 0; i < stream->nmsgs; ++i) if (s[i]) mail_searched (stream,i+1);
  }
}

/* Tenex mail ping mailbox
 * Accepts: MAIL stream
 * Returns: stream if still alive
 */

MAILSTREAM *tenex_ping (stream)
	MAILSTREAM *stream;
{
  int i = 0;
  short d = stream->debug;
  short debug = 36;
  short r;
  struct stat sbuf;
  struct iovec iov[2];
  MESSAGECACHE *elt;
  MAILSTREAM *bezerk = NIL;
				/* punt if stream no longer alive */
  if (!(stream && LOCAL)) return NIL;
#if unix
  if (LOCAL->inbox) {		/* only if this is an inbox */
    if (!LOCAL->filesize) {	/* do initial parse if necessary */
      stream->debug = debug;	/* don't notify upper level just yet */
      r = tenex_parse (stream);	/* initial silent part of stream */
      stream->debug = d;	/* restore former state */
      if (!r) return NIL;	/* punt if parse failed */
    }

				/* calculate name of bezerk file */
    sprintf (LOCAL->buf,"/usr/spool/mail/%s",getpwuid (geteuid ())->pw_name);
    stat (LOCAL->buf,&sbuf);	/* see if anything there */
				/* non-empty and we can lock our file? */
    if (sbuf.st_size && !flock (LOCAL->fd,LOCK_EX|LOCK_NB)) {
      fstat (LOCAL->fd,&sbuf);	/* yes, get current file size */
				/* sizes match and can get bezerk mailbox? */
      if ((sbuf.st_size == LOCAL->filesize) &&
	  (bezerk = mail_open (bezerk,LOCAL->buf,debug)) && bezerk->nmsgs) {
				/* yes, go to end of file in our mailbox */
	lseek (LOCAL->fd,sbuf.st_size,L_SET);
				/* for each message in bezerk mailbox */
	while (++i <= bezerk->nmsgs) {
				/* get its cache pointer, mark deleted */
	  (elt = mail_elt (bezerk,i))->deleted = T;
				/* snarf message from Berkeley mailbox */
	  iov[1].iov_base = bezerk_snarf (bezerk,i,&iov[1].iov_len);
				/* calculate header line */
	  sprintf ((iov[0].iov_base = LOCAL->buf),"%s,%d;000000000000\n",
		   elt->internal_date,iov[1].iov_len);
	  iov[0].iov_len = strlen (iov[0].iov_base);
				/* copy message to new mailbox */
	  if (writev (LOCAL->fd,iov,2) < 0) {
	    sprintf (LOCAL->buf,"Can't copy new mail: %s",strerror (errno));
	    fatal (LOCAL->buf);
	  }
	}
	mail_expunge (bezerk);	/* now expunge all those messages */
      }
      if (bezerk) mail_close (bezerk);
      flock (LOCAL->fd,LOCK_SH);/* back to shared access */
    }
  }
#endif
				/* parse mailbox, punt if parse dies */
  return (tenex_parse (stream)) ? stream : NIL;
}

/* Tenex mail check mailbox (too)
	reparses status too;
 * Accepts: MAIL stream
 */

void tenex_check (stream)
	MAILSTREAM *stream;
{
  int i = 1;
  if (tenex_ping (stream)) {	/* ping mailbox */
				/* get new message status */
    while (i <= stream->nmsgs) tenex_elt (stream,i++);
    mm_log ("Check completed",NIL);
  }
}


/* Tenex mail expunge mailbox
 * Accepts: MAIL stream
 */

void tenex_expunge (stream)
	MAILSTREAM *stream;
{
  int j;
  int i = 1;
  int n = 0;
  long delta = 0;
  MESSAGECACHE *elt;
  FILECACHE *m;
				/* do nothing if stream dead */
  if (!tenex_ping (stream)) return;
				/* get exclusive access */
  if (flock (LOCAL->fd,LOCK_EX|LOCK_NB)) {
    mm_log ("Expunge rejected: mailbox locked",ERROR);
    return;
  }

  while (i <= stream->nmsgs) {	/* for each message */
    m = LOCAL->msgs[i-1];	/* get file cache */
				/* if deleted */
    if ((elt = tenex_elt (stream,i))->deleted) {
				/* if recent, note one less recent message */
      if (elt->recent) --(stream->recent);
				/* one more message to flush */
      delta += m->headersize + m->bodysize + m->header - m->internal;
      fs_give ((void **) &m);	/* flush local cache entry */
      for (j = i; j < stream->nmsgs; j++) LOCAL->msgs[j-1] = LOCAL->msgs[j];
      LOCAL->msgs[stream->nmsgs-1] = NIL;
      mail_expunged (stream,i);	/* notify upper levels */
      n++;			/* count up one more expunged message */
    }
    else if (i++ && delta) {	/* if undeleted message needs moving */
				/* blat it down to new position */
      memmove (LOCAL->text + m->internal - delta,LOCAL->text + m->internal,
	       m->headersize + m->bodysize + m->header - m->internal);
      m->internal -= delta;	/* relocate pointers */
      m->header -= delta;
      m->body -= delta;
    }
  }
  if (n) {			/* expunged anything? */
				/* yes, calculate new file/text size */
    if ((LOCAL->textend -= delta) != (LOCAL->filesize -= delta))
      fatal ("File/text pointer screwup!");
    lseek (LOCAL->fd,0,L_SET);	/* seek to start of file */
				/* rewrite mailbox */
    if (write (LOCAL->fd,LOCAL->text,LOCAL->textend) < 0) {
      sprintf (LOCAL->buf,"Unable to rewrite mailbox: %s",strerror (errno));
      fatal (LOCAL->buf);
    }
				/* nuke any remaining cruft */
    ftruncate (LOCAL->fd,LOCAL->filesize);
    sprintf (LOCAL->buf,"Expunged %d messages",n);
    mm_log (LOCAL->buf,NIL);	/* output the news */
  }
  else mm_log ("No messages deleted, so no update needed",NIL);
				/* notify upper level of new mailbox size */
  mail_exists (stream,stream->nmsgs);
  mail_recent (stream,stream->recent);
  flock (LOCAL->fd,LOCK_SH);	/* allow sharers again */
}

/* Tenex mail copy message(s)
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 */

void tenex_copy (stream,sequence,mailbox)
	MAILSTREAM *stream;
	char *sequence;
	char *mailbox;
{
  unsigned char s[MAXMESSAGES];
				/* copy the messages */
  if (tenex_getsequence (stream,s,sequence))
    tenex_copy_messages (stream,s,mailbox);
}


/* Tenex mail move message(s)
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 */

void tenex_move (stream,sequence,mailbox)
	MAILSTREAM *stream;
	char *sequence;
	char *mailbox;
{
  int i = 0;
  unsigned char s[MAXMESSAGES];
  MESSAGECACHE *elt;
  if (tenex_getsequence (stream,s,sequence)) {
				/* copy the messages */
    tenex_copy_messages (stream,s,mailbox);
				/* delete all requested messages */
    while (i < stream->nmsgs) if (s[i++]) {
      (elt = tenex_elt (stream,i))->deleted = T;
				/* recalculate status */
      tenex_update_status (stream,i);
    }
    fsync (LOCAL->fd);		/* make sure the update takes */
  }
}

/* Internal routines */


/* Tenex mail generate file string
 * Accepts: temporary buffer to write into
 *	    mailbox name string
 * Returns: local file string
 */

char *tenex_file (dst,name)
	char *dst;
	char *name;
{
  char *home = getpwuid (geteuid ())->pw_dir;
  strcpy (dst,name);		/* copy the mailbox name */
  if (*dst != '/') {		/* pass absolute file paths */
    if (strcmp (ucase (dst),"INBOX")) sprintf (dst,"%s/%s",home,name);
				/* INBOX becomes ~USER/mail.txt file */
    else sprintf (dst,"%s/mail.txt",home);
  }
  return dst;
}


/* Parse flag list
 * Accepts: MAIL stream
 *	    flag list as a character string
 *	    pointer to user flags to return
 * Returns: system flags
 */

short tenex_getflags (stream,flag,uf)
	MAILSTREAM *stream;
	char *flag;
	long *uf;
{
  char key[MAILTMPLEN];
  char *t,*s;
  short f = 0;
  long i;
  short j;
  *uf = 0;			/* initially no user flags */
  if (flag && *flag) {		/* no-op if no flag string */
				/* check if a list and make sure valid */
    if ((i = (*flag == '(')) ^ (flag[strlen (flag)-1] == ')')) {
      mm_log ("Bad flag list",ERROR);
      return NIL;
    }
				/* copy the flag string w/o list construct */
    strncpy (LOCAL->buf,flag+i,(j = strlen (flag) - (2*i)));
    LOCAL->buf[j] = '\0';	/* tie off tail */

				/* make uppercase, find first, parse */
    if (t = strtok (ucase (LOCAL->buf)," ")) do {
      i = 0;			/* no flag yet */
				/* system flag, dispatch on first character */
      if (*t == '\\') switch (*++t) {
      case 'S':			/* possible \Seen flag */
	if (t[1] == 'E' && t[2] == 'E' && t[3] == 'N' && t[4] == '\0')
	  f |= i = fSEEN;
	break;
      case 'D':			/* possible \Deleted flag */
	if (t[1] == 'E' && t[2] == 'L' && t[3] == 'E' && t[4] == 'T' &&
	    t[5] == 'E' && t[6] == 'D' && t[7] == '\0') f |= i = fDELETED;
	break;
      case 'F':			/* possible \Flagged flag */
	if (t[1] == 'L' && t[2] == 'A' && t[3] == 'G' && t[4] == 'G' &&
	    t[5] == 'E' && t[6] == 'D' && t[7] == '\0') f |= i = fFLAGGED;
	break;
      case 'A':			/* possible \Answered flag */
	if (t[1] == 'N' && t[2] == 'S' && t[3] == 'W' && t[4] == 'E' &&
	    t[5] == 'R' && t[6] == 'E' && t[7] == 'D' && t[8] == '\0')
	  f |= i = fANSWERED;
	break;
      default:			/* unknown */
	break;
      }
      else {			/* user flag, search through table */
	for (j = 0; !i && j < NUSERFLAGS && (s = stream->user_flags[j]); ++j)
	  if (!strcmp (t,ucase (strcpy (key,s)))) *uf |= i = 1 << j;
      }
      if (!i) {			/* didn't find a matching flag? */
	sprintf (key,"Unknown flag: %s",t);
	mm_log (key,ERROR);
	*uf = NIL;		/* be sure no user flags returned */
	return NIL;		/* return no system flags */
      }
				/* parse next flag */
    } while (t = strtok (NIL," "));
  }
  return f;
}

/* Tenex mail parse mailbox
 * Accepts: MAIL stream
 * Returns: T if parse OK
 *	    NIL if failure, stream aborted
 */

#define PAD(s) LOCAL->textsize = (s | 0x7fff) + 0x8001

short tenex_parse (stream)
	MAILSTREAM *stream;
{
  struct stat sbuf;
  MESSAGECACHE *elt = NIL;
  FILECACHE *m = NIL;
  char c,*s,*t,*v;
  int i,delta;
  long j;
  int nmsgs = stream->nmsgs;
  int recent = stream->recent;
  fstat (LOCAL->fd,&sbuf);	/* get status */
				/* calculate change in size */
  if ((delta = sbuf.st_size - LOCAL->filesize) < 0) {
    sprintf (LOCAL->buf,"Mailbox shrank from %d to %d bytes, aborted",
	     LOCAL->filesize,sbuf.st_size);
    mm_log (LOCAL->buf,ERROR);	/* this is pretty bad */
    tenex_close (stream);
    return NIL;
  }
  if (delta > 0) {		/* apparent change in the file */
    if (!LOCAL->text)		/* create buffer if first time through */
      LOCAL->text = (char *) fs_get (PAD (sbuf.st_size));
				/* else resize if need more space */
    else if (LOCAL->textsize < (i = delta + LOCAL->textend))
      fs_resize ((void **) &LOCAL->text,PAD (i));
				/* get to that position in the file */
    lseek (LOCAL->fd,LOCAL->filesize,L_SET);
				/* read the new text */
    if (read (LOCAL->fd,(s = LOCAL->text+LOCAL->textend),delta) < 0) {
      sprintf (LOCAL->buf,"I/O error reading mail file: %s",strerror (errno));
      mm_log (LOCAL->buf,ERROR);
      tenex_close (stream);
      return NIL;
    }
				/* must begin with dd-mmm-yy hh:mm:ss" */
    if (s[2] != '-' || s[6] != '-' || s[9] != ' ' || s[12] != ':' ||
	s[15] != ':') {
      mm_log ("Mailbox invalidated by some other process, aborted",ERROR);
      tenex_close (stream);
      return NIL;
    }
				/* update parsed file size */
    LOCAL->filesize = sbuf.st_size;
				/* calculate and tie off end of text */
    *(LOCAL->text + (LOCAL->textend += delta)) = '\0';

				/* mailbox parse loop */
    while (s < (LOCAL->text + LOCAL->textend)) {
      v = s;			/* save pointer to internal header */
				/* parse header */
      if ((t = strchr (s,'\n')) && ((i = t - s) < LOCAL->buflen) &&
	  strncpy (LOCAL->buf,s,i) && (!(LOCAL->buf[i] = '\0')) &&
	  (s = t + 1) && strtok (LOCAL->buf,",") && (t = strtok (NIL,";")) &&
	  (i = strtol (t,&t,10)) && (!(t && *t)) && (t = strtok (NIL," ")) &&
	  (strlen (t) == 12)) {	/* create new cache entries if valid */
	m = LOCAL->msgs[nmsgs] = (FILECACHE *) fs_get (sizeof (FILECACHE));
	elt = mail_elt (stream,++nmsgs);
				/* note internal and message headers */
	m->internal = v - LOCAL->text;
	m->header = s - LOCAL->text;
	c = s[i];		/* save start of next message */
	s[i] = '\0';		/* tie off message temporarily */
				/* locate body */
	m->body = (v = strstr (s,"\n\n")) ? (v + 2) - LOCAL->text : NIL;
	s[i] = c;		/* restore next message character */
				/* compute header and body sizes */
	m->headersize = m->body ? m->body - m->header : i;
	m->bodysize = i - m->headersize;
	if (stream->nmsgs) {	/* if not first time through */
	  elt->recent = T;	/* this is a recent message */
	  recent++;		/* count up a new recent message */
	}
	elt->rfc822_size = i;	/* note message size */
	for (j = 0; j < i; j++) if (s[j] == '\n') elt->rfc822_size++;
				/* set internal date */
	elt->internal_date = cpystr (LOCAL->buf);
				/* as well as parsed value */
	m->date = tenex_date (LOCAL->buf);
	s += i;			/* advance to next message */
				/* calculate system flags */
	if ((i = ((t[10]-'0') * 8) + t[11]-'0') & fSEEN) elt->seen = T;
	if (i & fDELETED) elt->deleted = T;
	if (i & fFLAGGED) elt->flagged = T;
	if (i & fANSWERED) elt->answered = T;
	t[10] = '\0';		/* tie off flags */
	j = strtol (t,NIL,8);	/* get user flags value */
				/* set up all valid user flags (reversed!) */
	while (j) if (((i = 29 - find_rightmost_bit (&j)) < NUSERFLAGS) &&
		      stream->user_flags[i]) elt->user_flags |= 1 << i;
      }
      else {			/* header parse failed! */
	sprintf (LOCAL->buf,"Bad format in message %d",nmsgs);
	mm_log (LOCAL->buf,ERROR);
	tenex_close (stream);
	return NIL;
      }
    }
  }
  mail_exists (stream,nmsgs);	/* notify upper level of new mailbox size */
  mail_recent (stream,recent);	/* and of change in recent messages */
  return T;			/* return the winnage */
}

/* Tenex copy messages
 * Accepts: MAIL stream
 *	    mailbox copy vector
 *	    mailbox name
 */

void tenex_copy_messages (stream,s,mailbox)
	MAILSTREAM *stream;
	unsigned char s[];
	char *mailbox;
{
  int fd;
  struct iovec iov[2];
  int i = 0;
  FILECACHE *m;
				/* open file for create or append */
  if ((fd = open (tenex_file (LOCAL->buf,mailbox),O_WRONLY|O_APPEND|O_CREAT,
		  S_IREAD|S_IWRITE)) >= 0) {
    flock (fd,LOCK_EX);
				/* for each requested message */
    while (i < stream->nmsgs) if (s[i++]) {
      m = LOCAL->msgs[i-1];	/* internal message pointers */
				/* pointer to text of message */
      iov[1].iov_base = LOCAL->text + m->header;
      iov[1].iov_len = m->headersize + m->bodysize;
      sprintf ((iov[0].iov_base = LOCAL->buf),"%s,%d;000000000000\n",
	       mail_elt (stream,i)->internal_date,iov[1].iov_len);
				/* new internal header */
      iov[0].iov_len = strlen (iov[0].iov_base);
				/* write the message */
      if (writev (fd,iov,2) < 0) {
	sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
	fatal (LOCAL->buf);
      }
    }
    fsync (fd);			/* force out the update */
    flock (fd,LOCK_UN);		/* unlock mailbox */
    close (fd);			/* close the file */
  }
  else {
    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
    mm_log (LOCAL->buf,ERROR);
  }
}

/* Tenex get cache element with status updating from file
 * Accepts: MAIL stream
 *	    message number
 * Returns: cache element
 */

MESSAGECACHE *tenex_elt (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  long i,j;
  char c;
  char *s;
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  FILECACHE *m = LOCAL->msgs[msgno-1];
				/* locate flags in header */
  if (!(s = strchr (LOCAL->text + m->internal,';')) ||
      ((m->header - 1 - (i = ++s - LOCAL->text)) < 12))
    fatal ("Bad mailbox flag syntax!");
				/* set the seek pointer */
  lseek (LOCAL->fd,(off_t) i,L_SET);
				/* read the new flags */
  if (read (LOCAL->fd,s,12) < 0) {
    sprintf (LOCAL->buf,"Unable to read new status: %s",strerror (errno));
    fatal (LOCAL->buf);
  }
				/* calculate system flags */
  elt->seen = (i = ((s[10]-'0') * 8) + s[11]-'0') & fSEEN ? T : NIL;
  elt->deleted = i & fDELETED ? T : NIL;
  elt->flagged = i & fFLAGGED ? T : NIL;
  elt->answered = i & fANSWERED ? T : NIL;
  c = s[10];			/* remember first system flags byte */
  s[10] = '\0';			/* tie off flags */
  j = strtol (s,NIL,8);		/* get user flags value */
  s[10] = c;			/* restore first system flags byte */
				/* set up all valid user flags (reversed!) */
  while (j) if (((i = 29 - find_rightmost_bit (&j)) < NUSERFLAGS) &&
		stream->user_flags[i]) elt->user_flags |= 1 << i;
  return elt;
}

/* Tenex update status string
 * Accepts: MAIL stream
 *	    message number
 */

void tenex_update_status (stream,msgno)
	MAILSTREAM *stream;
	int msgno;
{
  long i,j;
  long k = 0;
  char *s;
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  FILECACHE *m = LOCAL->msgs[msgno-1];
				/* locate flags in header */
  if (!(s = strchr (LOCAL->text + m->internal,';')) ||
      ((m->header - 1 - (i = ++s - LOCAL->text)) < 12))
    fatal ("Bad mailbox flag syntax!");
  j = elt->user_flags;		/* get user flags */
				/* get the bits reversed */
  while (j) k |= 1 << 29 - find_rightmost_bit (&j);
				/* print new flag string */
  sprintf (LOCAL->buf,"%010lo%02o",k,
	   (fSEEN * elt->seen) + (fDELETED * elt->deleted) +
	   (fFLAGGED * elt->flagged) + (fANSWERED * elt->answered));
  memcpy (s,LOCAL->buf,12);	/* set the new flags w/o a NUL */
				/* set the seek pointer */
  lseek (LOCAL->fd,(off_t) i,L_SET);
				/* write the new flags */
  if (write (LOCAL->fd,s,12) < 0) {
    sprintf (LOCAL->buf,"Unable to write new status: %s",strerror (errno));
    fatal (LOCAL->buf);
  }
}

/* Tenex parse sequence
 * Accepts: MAIL stream
 *	    sequence vector
 *	    sequence to parse
 * Returns: T if parse successful, else NIL
 */

short tenex_getsequence (stream,s,sequence)
	MAILSTREAM *stream;
	unsigned char s[];
	char *sequence;
{
  int i,j,x;
  memset (s,NIL,MAXMESSAGES);	/* initialize sequence */
  while (*sequence) {		/* while there is something to parse */
				/* parse and validate message number */
    if (((i = (int) strtol ((const char *) sequence,&sequence,10)) < 1) ||
	(i > stream->nmsgs)) {
      mm_log ("Sequence invalid",ERROR);
      return NIL;
    }
    switch (*sequence) {	/* see what the delimiter is */
    case ':':			/* sequence range */
				/* parse end of range */
      if (((j = (int) strtol ((const char *) ++sequence,&sequence,10)) < 1) ||
	  (j > stream->nmsgs) || (*sequence && *sequence++ != ',')) {
	mm_log ("Sequence range invalid",ERROR);
	return NIL;
      }
      if (i > j) {		/* swap the range if backwards */
	x = i; i = j; j = x;
      }
      while (i <= j) s[--j] = T;/* mark each item in the sequence */
      break;
    case ',':			/* single message */
      ++sequence;		/* skip the delimiter, fall into end case */
    case '\0':			/* end of sequence */
      s[i-1] = T;		/* mark this message */
      break;
    default:			/* anything else is a syntax error! */
      mm_log ("Syntax error in sequence",ERROR);
      return NIL;
    }
  }
  return T;			/* successfully parsed sequence */
}

/* Search support routines
 * Accepts: MAIL stream
 *	    message number
 *	    pointer to additional data
 * Returns: T if search matches, else NIL
 */


char tenex_search_all (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return T;			/* ALL always succeeds */
}


char tenex_search_answered (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->answered ? T : NIL;
}


char tenex_search_deleted (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->deleted ? T : NIL;
}


char tenex_search_flagged (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->flagged ? T : NIL;
}


char tenex_search_keyword (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->user_flags & n ? T : NIL;
}


char tenex_search_new (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (elt->recent && !elt->seen) ? T : NIL;
}

char tenex_search_old (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->recent ? NIL : T;
}


char tenex_search_recent (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->recent ? T : NIL;
}


char tenex_search_seen (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->seen ? T : NIL;
}


char tenex_search_unanswered (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->answered ? NIL : T;
}


char tenex_search_undeleted (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->deleted ? NIL : T;
}


char tenex_search_unflagged (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->flagged ? NIL : T;
}


char tenex_search_unkeyword (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->user_flags & n ? NIL : T;
}


char tenex_search_unseen (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->seen ? NIL : T;
}

char tenex_search_before (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return (char) (LOCAL->msgs[msgno-1]->date < n);
}


char tenex_search_on (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  return (char) (LOCAL->msgs[msgno-1]->date == n);
}


char tenex_search_since (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
				/* everybody interprets "since" as .GE. */
  return (char) (LOCAL->msgs[msgno-1]->date >= n);
}


char tenex_search_body (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  FILECACHE *m = LOCAL->msgs[msgno-1];
  return search (LOCAL->text + m->body,m->bodysize,d,n);
}


char tenex_search_subject (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  char *s = tenex_fetchenvelope (stream,msgno)->subject;
  return s ? search (s,strlen (s),d,n) : NIL;
}


char tenex_search_text (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  FILECACHE *m = LOCAL->msgs[msgno-1];
  return search (LOCAL->text + m->header,m->headersize,d,n) ||
    tenex_search_body (stream,msgno,d,n);
}

char tenex_search_bcc (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,tenex_fetchenvelope (stream,msgno)->bcc);
  return search (LOCAL->buf,strlen (LOCAL->buf),d,n);
}


char tenex_search_cc (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,tenex_fetchenvelope (stream,msgno)->cc);
  return search (LOCAL->buf,strlen (LOCAL->buf),d,n);
}


char tenex_search_from (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,tenex_fetchenvelope (stream,msgno)->from);
  return search (LOCAL->buf,strlen (LOCAL->buf),d,n);
}


char tenex_search_to (stream,msgno,d,n)
	MAILSTREAM *stream;
	int msgno;
	char *d;
	long n;
{
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,tenex_fetchenvelope (stream,msgno)->to);
  return search (LOCAL->buf,strlen (LOCAL->buf),d,n);
}

/* Search parsers */


/* Parse a date
 * Accepts: function to return
 *	    pointer to date integer to return
 * Returns: function to return
 */

char (*tenex_search_date (f,n)) ()
	char (*f) ();
	long *n;
{
  long i;
  char *s;
				/* parse the date and return fn if OK */
  return (tenex_search_string (f,&s,&i) && (*n = tenex_date (s))) ? f : NIL;
}


/* Actual date parser (routine)
	minimal and stupid routine;
 * Accepts: date to parse
 * Returns: date integer
 */

int tenex_date (s)
	char *s;
{
  long ms;
  unsigned int d,m,y;
				/* parse first number (probable month) */
  if (!(s && (m = (int) strtol ((const char *) s,&s,10)))) return NIL;
  switch (*s) {			/* different parse based on delimiter */
  case '/':			/* mm/dd/yy format */
				/* parse remainder of date */
    if (!((d = (int) strtol ((const char *) ++s,&s,10)) && *s == '/' &&
	  (y = (int) strtol ((const char *) ++s,&s,10)) && *s == '\0'))
      return NIL;
    break;

  case '-':			/* dd-mmm-yy format */
    d = m;			/* so the number we got is a day */
				/* make sure string is UC and long enough! */
    if (strlen (ucase (s)) < 5) return NIL;
				/* slurp up the month string */
    ms = (s[1] << 16) + (s[2] << 8) + s[3];
    switch (ms) {		/* determine the month */
    case ('J' << 16) + ('A' << 8) + 'N':
      m = 1; break;
    case ('F' << 16) + ('E' << 8) + 'B':
      m = 2; break;
    case ('M' << 16) + ('A' << 8) + 'R':
      m = 3; break;
    case ('A' << 16) + ('P' << 8) + 'R':
      m = 4; break;
    case ('M' << 16) + ('A' << 8) + 'Y':
      m = 5; break;
    case ('J' << 16) + ('U' << 8) + 'N':
      m = 6; break;
    case ('J' << 16) + ('U' << 8) + 'L':
      m = 7; break;
    case ('A' << 16) + ('U' << 8) + 'G':
      m = 8; break;
    case ('S' << 16) + ('E' << 8) + 'P':
      m = 9; break;
    case ('O' << 16) + ('C' << 8) + 'T':
      m = 10; break;
    case ('N' << 16) + ('O' << 8) + 'V':
      m = 11; break;
    case ('D' << 16) + ('E' << 8) + 'C':
      m = 12; break;
    default:
      return NIL;
    }
				/* parse the year */
    if (s[4] == '-' && (y = (int) strtol ((const char *) s+5,&s,10)) &&
	(*s == '\0' || *s == ' ')) break;
  default:			/* unknown format */
    return NIL;
  }
  y += (y < 64) ? 2000 : 0;	/* the PDP-10 world began in 1964 */
  y -= (y >= 1964) ? 1964 : 64;
				/* minimal validity check of date */
  if (d < 1 || d > 31 || m < 1 || m > 12 || y < 0 || y >= 100) return NIL;
  return (y*12+m-1)*31+d-1;	/* calculate and return date value */
}

/* Parse a flag
 * Accepts: function to return
 *	    pointer to keyword integer to return
 *	    MAIL stream
 * Returns: function to return
 */

char (*tenex_search_flag (f,n,stream)) ()
	char (*f) ();
	long *n;
	MAILSTREAM *stream;
{
  short i;
  char *s,*t;
  if (t = strtok (NIL," ")) {	/* get a keyword */
    ucase (t);			/* get uppercase form of flag */
    for (i = 0; i < NUSERFLAGS && (s = stream->user_flags[i]); ++i)
      if (!strcmp (t,ucase (strcpy (LOCAL->buf,s))) && (*n = 1 << i)) return f;
  }
  return NIL;			/* couldn't find keyword */
}


/* Parse a string
 * Accepts: function to return
 *	    pointer to string to return
 *	    pointer to string length to return
 * Returns: function to return
 */


char (*tenex_search_string (f,d,n)) ()
	char (*f) ();
	char **d;
	long *n;
{
  char *c = strtok (NIL,"");	/* remainder of criteria */
  if (c) {			/* better be an argument */
    switch (*c) {		/* see what the argument is */
    case '\0':			/* catch bogons */
    case ' ':
      return NIL;
    case '"':			/* quoted string */
      if (!(strchr (c+1,'"') && (*d = strtok (c,"\"")) && (*n = strlen (*d))))
	return NIL;
      break;
    case '{':			/* literal string */
      *n = strtol (c+1,&c,10);	/* get its length */
      if (*c++ != '}' || *c++ != '\015' || *c++ != '\012' ||
	  *n > strlen (*d = c)) return NIL;
      c[*n] = '\255';		/* write new delimiter */
      strtok (c,"\255");	/* reset the strtok mechanism */
      break;
    default:			/* atomic string */
      *n = strlen (*d = strtok (c," "));
      break;
    }
    return f;
  }
  else return NIL;
}
